<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Julian Yerfino <jyerfino@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('lib/core/Project.php');
require_once('lib/core/ProjectPeer.php');
require_once('lib/GFConf.php');
require_once('lib/exceptions/InvalidUserException.php');

class GFExportImport {
	
	private $project;
	private $specData;
	private $phpSpecData;
	private $tableQueue;
	private $importedXMLData;
	private $filesArray;
	
	/* 
	
		$this->specData = array of XML Strings returned by the various plugins.
	
	*/
	
	public function  __construct(Project $prj = null) {
		$this->project = $prj;
		$this->specData = array();
		$this->phpSpecData = array();
		$this->tableQueue = array();
		$this->importedXMLData = null;
		$this->filesArray = array();
	}
	
	function scanDirectories($rootDir, $allData=array()) {
		// set filenames invisible if you want
		$invisibleFileNames = array(".", "..", ".htaccess", ".htpasswd");
		// run through content of root directory
		if(is_dir($rootDir)) {
			$dirContent = scandir($rootDir);
			foreach($dirContent as $content) {
				// filter all files not accessible
				$path = $rootDir.'/'.$content;
				if(!in_array($content, $invisibleFileNames)) {
					// if content is file & readable, add to array
					if(is_file($path) && is_readable($path)) {
						// save file name with path
						$allData[] = $path;
						// if content is a directory and readable, add path and name
					}elseif(is_dir($path) && is_readable($path)) {
						// recursive callback to open new directory
						$allData = $this->scanDirectories($path, $allData);
					}
				}
			}
		}
		return $allData;
	}
	
	private function processSpecData() {
		reset($this->specData);
		
		$this->phpSpecData = array();
		$this->tableQueue = array();
		
		
		foreach($this->specData as $xml) {
			// Convert the XML string to an usable array
			$dom = new DOMDocument();
			$dom->loadXML($xml);

			/** @var DOMNode $specNode */
			$specNode = $dom->getElementsByTagName('gforgeSpecData')->item(0);
			/** @var DOMNodeList $dbNodes */
			$dbNodes = $specNode->getElementsByTagName('db');
			
			foreach($dbNodes as $dbNode) {
				// Retrieve all the tables stated in the XML file.
				
				foreach($dbNode->childNodes as $table ) {
					
					if($table->nodeName=='table') {
						$name = $table->attributes->getNamedItem('name')->value;
						
						// Parse the links info (conditioned foreign keys)
						
						$keyArray = array();
						foreach($table->childNodes as $key)
						{
							if($key->nodeName=='key') {
								$types = $key->attributes->getNamedItem('type')->value;
								if(!is_null($types)) {
									$types = explode(',', $types);
								}
								else {
									$types = array();
								}
								$key_name = $key->getAttribute('name');
								$keyArray[] = array('name' => $key_name, 'types' => $types, 'value' => $key->getAttribute('value'));
							}
						}
						
						// Retrieve the table info from the DB to search for primary and foreign keys.
						$con = Propel::getConnection(ProjectPeer::DATABASE_NAME);
						$dbInfo = $con->getDatabaseInfo();
						$tableInfo = $dbInfo->getTable($name);
						$primaryKeyInfo = $tableInfo->getPrimaryKey();
						$foreignKeysInfo = $tableInfo->getForeignKeys();
						// We assume that all indexes that end with _key represent a unique key constraint.
						$indexesInfo = $tableInfo->getIndexes();

						// Add the info to $this->phpSpecData, if the table $name already exists there its info will be replaced.
						if(!isset($this->phpSpecData[$name])) {
							$this->phpSpecData[$name]=array();
						}
						$phpSpecDataIndex = count($this->phpSpecData[$name]);
						$this->phpSpecData[$name][$phpSpecDataIndex] = array('fk' => array(), 'keys' => $keyArray, 'unique' => array(), 'pk' => array());
						
						$this->dbData[$name] = array('loaded' => false);
						$this->tableQueue[]=$name;
						// Process Foreign Keys
						foreach($foreignKeysInfo as $fk) {
							$fkInfo = $fk->getReferences();
							$this->phpSpecData[$name][$phpSpecDataIndex]['fk'][$fkInfo[0][0]->getName()] = array('table' => $fkInfo[0][1]->getTable()->getName(), 'column' =>$fkInfo[0][1]->getName());
						}
						
						//Process Primary Keys
						if(is_object($primaryKeyInfo)) {
							$pkCols = $primaryKeyInfo->getColumns();
							if(is_array($pkCols)) {
								foreach($pkCols as $pkCol) {
									$this->phpSpecData[$name][$phpSpecDataIndex]['pk'][]=$pkCol->getName();
								}
							}
						}


						// Search for unique column value constraints
						foreach($indexesInfo as $indexInfo) {
							$indexName = $indexInfo->getName();
							if(strcasecmp('_key', substr($indexName, -4))==0) {
								// The index name ends with _key, we assume its a unique key constraint.
								$colname = preg_replace("/${name}_/", '', $indexName);
								$colname = preg_replace("/_key$/", '', $colname);
								//$colname = ereg_replace("${name}_", '', $indexName);
								//$colname = ereg_replace("_key$", '', $colname);
								if($colname!='') {
									$this->phpSpecData[$name][$phpSpecDataIndex]['unique'][]=$colname;
								}
							}
						}
					}					
				}
			}
			unset($dom);
		}
	}
	

	
	/* Process all of phpSpecData, retrieving the info from the DB.*/
	
	private function loadDBData() {
		$this->dbData = array();
		$tblQ = $this->tableQueue;

		reset($tblQ);
		for($tblQCounter=0;$tblQCounter < count($tblQ); $tblQCounter++) {
			$name = $tblQ[$tblQCounter];
			
			for($phpSpecDataIndex=0; $phpSpecDataIndex < count($this->phpSpecData[$name]); $phpSpecDataIndex++) {
				if(!isset($this->phpSpecData[$name][$phpSpecDataIndex]['loaded'])) {
					$this->phpSpecData[$name][$phpSpecDataIndex]['loaded'] = false;
				}
				if (!$this->phpSpecData[$name][$phpSpecDataIndex]['loaded']) {
					$tbl = $this->phpSpecData[$name][$phpSpecDataIndex];
					// Process all the table's keys as stated in the spec, adjust the SQL query accordingly.

					$sqlBase = "SELECT \"$name\".* FROM \"$name\"";
					$sqlJoin='';
					$sql='';
					reset($tbl['keys']);
					foreach($tbl['keys'] as $keys_array) {
						reset($keys_array);
						$key_name = $keys_array['name'];
						foreach ($keys_array['types'] as $key_type) {
							switch($key_type) {
								case 'project_id':
									// Filter key with the current project_id
									$sql .= "\"$name\".$key_name = " . $this->project->getProjectId() . " AND ";
									// If the column $key_name is listed for processing with foreign_keys, remove it.
									if(isset($tbl['fk'][$key_name])) {
										unset($tbl['fk'][$key_name]);
									}
									break;
								case 'mailing_list_name':
								case 'user_id':
								case 'plugin_id':
								case 'system':
									// Cross-System invariant ID, don't need to link to its information as it can be copied safely in the export and import system.
								case 'ignore':
								case 'exportIgnore':
								case 'map':
								case 'nullify':
									// Just ignore the key
									// If the column $key_name is listed for processing with foreign_keys, remove it.
									if(isset($tbl['fk'][$key_name])) {
										unset($tbl['fk'][$key_name]);
									}
									break;
								case 'fixed':
									// Filter key with a fixed value.
									$sql .= "\"$name\".$key_name = '" . $keys_array['value'] . "' AND ";
									if(isset($tbl['fk'][$key_name])) {
										unset($tbl['fk'][$key_name]);
									}
									break;
								case 'fixedJoin':
									// name has format table.column
									// value = fixed_value that table.column must have
									// table must be referenced by one of the current table foreign keys
										
										$names = explode('.', $key_name);
										if(count($names)==2) {
											$tn = $names[0];
											$cn = $names[1];
											$v =  $keys_array['value'];

											foreach($tbl['fk'] as $col => $fk) {
												if($fk['table']==$tn) {
													$sqlJoin .= "JOIN $tn ON \"$tn\".".$fk['column']."=\"$name\".$col";
													$sql.= "\"$tn\".$cn='$v' AND ";
													break;
												}
											}
										}
									break;
								case 'link':
								case 'exportLink':
									// Linked field, this applies a foreign-key like restriction to a key.
									$values = explode('.', $keys_array['value']);
									if (count($values)==2) {
										$tbl['fk'][$key_name]=array('table' => $values[0], 'column' => $values[1]);
									}
								case 'project_name':
								default:
									break;
							}
						}
					}

					reset($tbl['fk']);
					$skipSQL = false;

					if(!isset($this->dbData[$name])) {
						$this->dbData[$name]=array();
						$this->dbData[$name]['loaded'] = false;
					}

					foreach($tbl['fk'] as $fkOrig => $fkDest) {
						// Search previous data, if not all foreign tables are loaded, push the table to be loaded later.
						if (array_key_exists($fkDest['table'], $this->phpSpecData)) {
							if(array_key_exists($fkDest['table'], $this->dbData) && $this->dbData[$fkDest['table']]['loaded']) {
								if (count($this->dbData[$fkDest['table']]['data'])>0) {
									reset($this->dbData[$fkDest['table']]['data']);
									$values = '';
									$values_arr = array();

									$i = 0;
									foreach($this->dbData[$fkDest['table']]['data'] as $fkData) {


											$values .= "'" .$fkData[$fkDest['column']] . "',";

									}


									//preg_replace("/,\n/", "\n", $values);
									$values = substr($values, 0, strlen($values)-1);

									$sql .= "\"$name\".$fkOrig IN ($values) AND ";
								}
								else {
									// There will be no data for this entry.
									$skipSQL = true;
									//$this->dbData[$name]['loaded'] = true;
									if(!isset($this->dbData[$name]['data'])) {
										$this->dbData[$name]['data'] = array();
									}
									$this->phpSpecData[$name][$phpSpecDataIndex]['loaded'] = true;
								}
							}
							else {
								// One foreign table not yet loaded, push it for later.
								$tblQ[]=$name;
								$skipSQL = true;
							}
						}
						else {
							// Some foreign table not found, probably the plugin is disabled. Skip it.
							$skipSQL = true;
							if(!isset($this->dbData[$name]['data'])) {
								$this->dbData[$name]['data'] = array();
							}
							$this->phpSpecData[$name][$phpSpecDataIndex]['loaded'] = true;
						}
					}

					$sql = "$sqlBase $sqlJoin WHERE $sql";

					$sql = preg_replace('/AND $/', '', $sql);
					$sql = preg_replace('/ WHERE $/', '', $sql);
					//$sql = ereg_replace('AND $', '', $sql);
					//$sql = ereg_replace(' WHERE $', '', $sql);


					// Run SQL query and add the data to the dbData array.
					if (!$skipSQL) {
						$this->phpSpecData[$name][$phpSpecDataIndex]['loaded'] = true;
						if(!isset($this->dbData[$name]['data'])) {
							$this->dbData[$name]['data'] = array();
						}
						$con = Propel::getConnection(ProjectPeer::DATABASE_NAME);
						$stmt = $con->createStatement();
						$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_ASSOC);
						$rowCounter=count($this->dbData[$name]['data']);
						//$con->close();

						while($rs->next()) {
							$this->dbData[$name]['data'][$rowCounter] = $rs->getRow();


							/* Map bool fields to strings true and false */
							foreach($this->dbData[$name]['data'][$rowCounter] as $k => $f) {
								if(is_bool($f)) {
									$this->dbData[$name]['data'][$rowCounter][$k] = $f ? 'true':'false';
								}
							}

							/* 	Map the user_id and plugin_id fields from the user_id to the unix_name, from the plugin_id to the plugin_name
								Also remove from the mailing list names the project unix_name*/
							
							reset($tbl['keys']);
							
							foreach($tbl['keys'] as $keys_array) {
								reset($keys_array);
								$key_name = $keys_array['name'];
								foreach ($keys_array['types'] as $key_type) {
									switch($key_type) {
										case 'user_id':
											$mappedUser = UserPeer::retrieveByPk($this->dbData[$name]['data'][$rowCounter][$key_name]);
											if(is_object($mappedUser)) {												
												$this->dbData[$name]['data'][$rowCounter][$key_name] = $mappedUser->getUnixName();
											}
											else {
												$this->dbData[$name]['data'][$rowCounter][$key_name] = '';
											}
											break;
										case 'plugin_id':
											$mappedPlugin = PluginPeer::retrieveByPk($this->dbData[$name]['data'][$rowCounter][$key_name]);
											if(is_object($mappedPlugin)) {
												$this->dbData[$name]['data'][$rowCounter][$key_name] = $mappedPlugin->getPluginName();
											}
											else {
												$this->dbData[$name]['data'][$rowCounter][$key_name] = '';
											}
											break;
										case 'mailing_list_name':						
											$this->dbData[$name]['data'][$rowCounter][$key_name] = substr($this->dbData[$name]['data'][$rowCounter][$key_name], strlen($this->project->getUnixName())+1);
											break;
										case 'nullify':
											$this->dbData[$name]['data'][$rowCounter][$key_name] = '';
											break;
										default:
											break;
									}
								}
							}
							$rowCounter++;
						}
					}
				}
			}
			$all_loaded = true;
			$tbls = $this->phpSpecData[$name];
			foreach($tbls as $tbl){
				if(!isset($tbl['loaded']) || $tbl['loaded'] == false) {
					$all_loaded = false;
					break;
				}
			}
			if ($all_loaded) {
				$this->dbData[$name]['loaded'] = true;
			}

		}
	}
	
	/* Load all data XML files belonging to a exported project */
	private function loadExportedData($dataDir) {
		if(is_null($this->importedXMLData)) {
			$this->importedXMLData = array();
			
			if(!is_dir($dataDir)) {
				throw new InvalidDataException(array("Directory $dataDir not found"));
			}

			$this->importedXMLData['system'] = new DOMDocument();
			$xmlPath = SystemUtils::buildFilePath($dataDir, 'system.xml');
			if(!is_file($xmlPath) || !$this->importedXMLData['system']->load($xmlPath)) {
				throw new InvalidDataException(array("Can not load $xmlPath"));
			}

			$this->importedXMLData['filesystem'] = new DOMDocument();
			$xmlPath = SystemUtils::buildFilePath($dataDir, 'filesystem.xml');
			if (!is_file($xmlPath) || !$this->importedXMLData['filesystem']->load($xmlPath)) {
				throw new InvalidDataException(array("Can not load $xmlPath"));
			}

			$pluginsDataDir = SystemUtils::buildFilePath($dataDir, 'plugins');
			$pluginXMLFiles = scandir($pluginsDataDir);

			foreach($pluginXMLFiles as $pluginXMLFile) {
				$pluginXMLFilePath = SystemUtils::buildFilePath($pluginsDataDir, $pluginXMLFile);
				if(is_file($pluginXMLFilePath)) {
					$this->importedXMLData['plugins/'.$pluginXMLFile] = new DOMDocument();
					$this->importedXMLData['plugins/'.$pluginXMLFile]->load($pluginXMLFilePath);
				}
			}
		}
	}
	
	/* Check the exported plugin data to see which plugins were enabled */
	private function getEnabledPlugins($dataDir) {
		$this->loadExportedData($dataDir);
		$pluginList = array();
		$pf = PluginFactory::getFactory();
		foreach($this->importedXMLData as $domNode) {
			$tableNodes = $domNode->getElementsByTagName('table');
			foreach($tableNodes as $tableNode) {		
				if(strcasecmp($tableNode->getAttribute('name'), 'project_plugin')==0) {
					foreach($tableNode->childNodes as $row) {
						if($row->nodeName=='row') {
							foreach($row->childNodes as $column) {
								if($column->nodeName=='column' && strcasecmp($column->getAttribute('name'),'plugin_id')==0) {
									$pluginName = urldecode($column->textContent);
									$plugin = $pf->getPluginByName($pluginName);
									if(is_object($plugin)){
										$pluginList[$pluginName]=$plugin;
									}
								}
							}
						}		
					}
					break;
				}		
			}
		}
		return $pluginList;
	}
	
	private function findKeyInSpec($key_type_search) {
		$tblQ = $this->tableQueue;
		
		reset($tblQ);
		$foundProjectNameKey = false;
		foreach($tblQ as $name) {
			for($phpSpecDataIndex=0; $phpSpecDataIndex < count($this->phpSpecData[$name]); $phpSpecDataIndex++) {
				$tbl = $this->phpSpecData[$name][$phpSpecDataIndex];

				// Process all the table's keys, looking for a 'project_name' type of key.
				reset($tbl['keys']);
				foreach($tbl['keys'] as $keys_array) {
					reset($keys_array);
					$key_name = $keys_array['name'];
					foreach ($keys_array['types'] as $key_type) {
						if($key_type == $key_type_search) {
							$foundProjectNameKey = true;
							break;
						}
					}
					if($foundProjectNameKey) {
						break;
					}
				}
				if($foundProjectNameKey) {
					break;
				}
			}
		}

		if(! $foundProjectNameKey) {
			throw new Exception('Invalid Spec Data Provided. It must contain a project_name type key');
		}
		
		return array('table' => $name, 'column' => $key_name);
	}
	
	private function findFirstValueInExportedData($tableName, $columnName) {
		$projectNameByXML = '';
		$foundProjectNameKey = false;
		
		foreach($this->importedXMLData as $domNode) {
			$tableNodes = $domNode->getElementsByTagName('table');
			foreach($tableNodes as $tableNode) {		
				if(strcasecmp($tableNode->getAttribute('name'), $tableName)==0) {
					foreach($tableNode->childNodes as $row) {
						if($row->nodeName=='row') {
							foreach($row->childNodes as $column) {
								if($column->nodeName=='column' && strcasecmp($column->getAttribute('name'),$columnName)==0) {
									$projectNameByXML = urldecode($column->textContent);
									$foundProjectNameKey = true;
									break;	
								}
							}
						}
						if($foundProjectNameKey) {
							break;
						}
					}
				}
				if($foundProjectNameKey) {
					break;
				}
			}
			if($foundProjectNameKey) {
				break;
			}
		}
		
		if(! $foundProjectNameKey) {
			throw new Exception('Invalid XML Data Provided. It must contain a project_name type column');
		}
		
		return $projectNameByXML;
	}
	
	/* Roam through the already loaded dbData and delete EVERYTHING it points at */
	private function deleteAllProjectData() {
		/*$con = Propel::getConnection(ProjectPeer::DATABASE_NAME);
		$con->begin();
		
		$tblQ = $this->tableQueue;
		try {
			reset($tblQ);
			foreach($tblQ as $name) {
				$tbl = $this->phpSpecData[$name];

				// Process all the table's keys as stated in the spec, adjust the SQL query accordingly.

				$sql = "DELETE FROM \"$name\" WHERE ";
				reset($tbl['keys']);
				foreach($tbl['keys'] as $keys_array) {
					reset($keys_array);
					$key_name = $keys_array['name'];
					foreach ($keys_array['types'] as $key_type) {
						switch($key_type) {
							case 'project_id':
								// Filter key with the current project_id
								$sql .= "$key_name = " . $this->project->getProjectId() . " AND ";
								// If the column $key_name is listed for processing with foreign_keys, remove it.
								if(isset($tbl['fk'][$key_name])) {
									unset($tbl['fk'][$key_name]);
								}
								break;
							case 'user_id':
							case 'system':
								// Cross-System invariant ID, don't need to link to its information as it can be copied safely in the export and import system.
							case 'ignore':
							case 'exportIgnore':
								// Just ignore the key

								// If the column $key_name is listed for processing with foreign_keys, remove it.
								if(isset($tbl['fk'][$key_name])) {
									unset($tbl['fk'][$key_name]);
								}
								break;
							case 'fixed':
								// Filter key with a fixed value.
								$sql .= "$key_name = '" . $keys_array['value'] . "' AND ";
								if(isset($tbl['fk'][$key_name])) {
									unset($tbl['fk'][$key_name]);
								}
								break;
							case 'link':
							case 'exportLink':
								// Linked field, this applies a foreign-key like restriction to a key.
								$values = explode('.', $keys_array['value']);
								if (count($values)==2) {
									$tbl['fk'][$key_name]=array('table' => $values[0], 'column' => $values[1]);
								}
							case 'project_name':
							default:
								break;
						}
					}
				}

				reset($tbl['fk']);
				$skipSQL = false;

				foreach($tbl['fk'] as $fkOrig => $fkDest) {
					// Search previous data, if not all foreign tables are loaded, push the table to be loaded later.
					if (array_key_exists($fkDest['table'], $this->phpSpecData)) {
						if($this->dbData[$fkDest['table']]['loaded']) {
							if (count($this->dbData[$fkDest['table']]['data'])>0) {
								reset($this->dbData[$fkDest['table']]['data']);
								$values = '';

								foreach($this->dbData[$fkDest['table']]['data'] as $fkData) {
									$values .= "'" . $fkData[$fkDest['column']] . "',";
								}
								$values = substr($values, 0, strlen($values)-1);

								$sql .= "$fkOrig IN ($values) AND ";
							}
							else {
								$skipSQL = true;
							}
						}
						else {
							throw new Exception (__LINE__ . ':Invalid dbData. Table "' . $fkDest['table'] . '" referenced by table "' . $name . '" should be loaded but its not');
						}
					}
					else {
						throw new Exception (__LINE__ . ': Invalid specData. Table "' . $fkDest['table'] . '" referenced by table "' . $name . '" but it was not specified to be loaded in the spec');
					}
				}

				$sql = ereg_replace('AND $', '', $sql);
				$sql = ereg_replace(' WHERE $', '', $sql);

				// Run SQL query

				if (!$skipSQL) {
					$stmt = $con->createStatement();
					$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_ASSOC);

				}
			}

			// Commit deletions, if there are errors roll back and abort.
			$con->commit();
		} catch(SQLException $sqle) {
			$con->rollback();
			$session = Session::getSession();
			$Lang = $session->getLanguage();
			throw new Exception ($Lang->getText('ProjectTemplate.DeleteError'));
		}*/
	}
	
	private function doExport() {	
		
		$this->dateString = date('YmdHis');
		
		$this->processSpecData();
		
		$this->loadDBData();
		
		/* 
			Save all the info to various XML files.
			We will re-process all of the XML input spec Data and cross it against the data we fetched, to be able to output the XML files
		*/
		reset($this->specData);
		foreach($this->specData as $xml) {
			// Convert the XML string to an usable array
			
			$dom = new DOMDocument();
			$dom->loadXML($xml);
			
			$specNode = $dom->getElementsByTagName('gforgeSpecData')->item(0);
			$dbNodes = $specNode->getElementsByTagName('db');
			
			foreach($dbNodes as $dbNode) {
				// Retrieve all the tables stated in the XML file.
				$domOutput = new DOMDocument();
				$domOutput->formatOutput = true;
				$domOutput->encoding='UTF-8';
				//$domOutput->xmlEncoding='UTF-8';
				$mainNode = $domOutput->createElement('gforgeData');
				$mainNode->setAttribute('type', $dbNode->getAttribute('type'));
				foreach($dbNode->childNodes as $table ) {
					if($table->nodeName=='table' && strcasecmp($table->getAttribute('export'),'no')!=0) {
						$name = $table->getAttribute('name');
						$tableNode = $domOutput->createElement('table');
						$tableNode->setAttribute('name', $name);
						
						if(!isset($this->dbData[$name]['exported']) || !$this->dbData[$name]['exported']) {
							$this->dbData[$name]['exported']=true;
							reset($this->dbData[$name]);
							foreach($this->dbData[$name]['data'] as $rowNum => $row) {
								$rowNode = $domOutput->createElement('row');
								$rowNode->setAttribute('id', $rowNum);
								foreach($row as $colName => $value) {
									$colNode = $domOutput->createElement('column');
									$colNode->setAttribute('name', $colName);
									$cDATANode = $domOutput->createCDATASection(urlencode($value));
									$colNode->appendChild($cDATANode);
									$rowNode->appendChild($colNode);
								}
								$tableNode->appendChild($rowNode);
							}
							$mainNode->appendChild($tableNode);
						}
					}
				}
				//$domOutput->importNode($mainNode);
				$domOutput->appendChild($mainNode);
				$outputFile = SystemUtils::buildFilePath(GFConf::get('system.exportProjects'), $this->project->getUnixName());
				$outputFile = SystemUtils::buildFilePath($outputFile, $this->dateString);
				$outputFile = SystemUtils::buildFilePath($outputFile, $dbNode->getAttribute('outputFile'));
				if(!file_exists(dirname($outputFile))) {
					mkdir(dirname($outputFile), 0777, true);
				}
				$domOutput->save($outputFile);
			}
			
			/* Now fetch the files and output them */
			$fileNodes = $specNode->getElementsByTagName('files');
			foreach($fileNodes as $fileNode) {
				foreach($fileNode->childNodes as $filter ) {
					if($filter->nodeName=='filter') {
						$outputDirectory = SystemUtils::buildFilePath(GFConf::get('system.exportProjects'), $this->project->getUnixName());
						$outputDirectory = SystemUtils::buildFilePath($outputDirectory, $this->dateString);
						$outputDirectory = SystemUtils::buildFilePath($outputDirectory, $filter->getAttribute('outputDirectory'));
						
						$sql = "SELECT * FROM filesystem WHERE ";
						$skipSQL = false;
						
						foreach($filter->childNodes as $key) {
							if($key->nodeName=='field') {
								switch($key->getAttribute('type')) {
									case 'fixed':
										$sql .= $key->getAttribute('name') . "='" . $key->getAttribute('value') . "' AND ";
										break;
									case 'link':
									case 'exportLink':
										$values = explode('.', $key->getAttribute('value'));
										if (count($values)==2) {
											$linkDataArr = array();
											foreach($this->dbData[$values[0]]['data'] as $row) {
												$linkDataArr[] = $row[$values[1]];
											}
											if(count($linkDataArr)>0) {
												$sql .= $key->getAttribute('name') . " IN (" . implode(',',$linkDataArr) .") AND ";
											}
											else {
												$skipSQL = true;
											}
										}
										break;
									case 'project_id':
										$sql .= $key->getAttribute('name') . " =" . $this->project->getProjectId() ." AND ";
										break;
									default:
										break;
								}
							}
						}
						
						if(!$skipSQL) {
							$sql = preg_replace('/AND $/','', $sql);
							$sql = preg_replace('/ WHERE $/','',$sql);
							//$sql = ereg_replace('AND $', '', $sql);
							//$sql = ereg_replace(' WHERE $', '', $sql);

							$con = Propel::getConnection(ProjectPeer::DATABASE_NAME);
							$stmt = $con->createStatement();
							$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_ASSOC);
							
							while($rs->next()) {
								$fileEntry = $rs->getRow();
								//Map the posted_by entry
								$mappedUser = UserPeer::retrieveByPk($fileEntry['posted_by']);
								if(is_object($mappedUser)) {
									$fileEntry['posted_by'] = $mappedUser->getUnixName();
								} else {
									$fileEntry['posted_by'] = '';
								}

								$this->filesArray[] = $fileEntry;
								$filePath = '';
								$fileId = sprintf('%03u',$fileEntry['filesystem_id']);
								$idLength = strlen($fileId);
								for($fpc=0; $fpc < $idLength && $fpc < 3; $fpc++) {
									$filePath .= $fileId[$fpc] . '/';
								}
								$filePath .= $fileEntry['filesystem_id'] . '/' . $fileEntry['file_name_safe'];

								$outputFile = SystemUtils::buildFilePath($outputDirectory, $filePath);
								$inputFile = SystemUtils::buildFilePath(GFConf::get('system.filesPath'), $filter->getAttribute('plugin'));
								$inputFile = SystemUtils::buildFilePath($inputFile, $filePath);

								if(!file_exists(dirname($outputFile))) {
									mkdir(dirname($outputFile), 0777, true);
								}
								copy($inputFile, $outputFile);
							}
						}
					}
				}
			}
		}
		
		// Output the exported filesystem information to an XML file.
		// Retrieve all the tables stated in the XML file.
		$domOutput = new DOMDocument();
		$domOutput->formatOutput = true;
		$domOutput->encoding='UTF-8';
		//$domOutput->xmlEncoding='UTF-8';
		$mainNode = $domOutput->createElement('gforgeData');
		$mainNode->setAttribute('type', 'filesystem');
		$tableNode = $domOutput->createElement('table');
		$tableNode->setAttribute('name', 'filesystem');
		foreach($this->filesArray as $rowNum => $row) {
			$rowNode = $domOutput->createElement('row');
			$rowNode->setAttribute('id', $rowNum);
			foreach($row as $colName => $value) {
				$colNode = $domOutput->createElement('column');
				$colNode->setAttribute('name', $colName);
				$cDATANode = $domOutput->createCDATASection(urlencode($value));
				$colNode->appendChild($cDATANode);
				$rowNode->appendChild($colNode);
			}
			$tableNode->appendChild($rowNode);
		}
		$mainNode->appendChild($tableNode);
		$domOutput->appendChild($mainNode);
		
		$outputFile = SystemUtils::buildFilePath(GFConf::get('system.exportProjects'), $this->project->getUnixName());
		$outputFile = SystemUtils::buildFilePath($outputFile, $this->dateString);
		$outputFile = SystemUtils::buildFilePath($outputFile, 'filesystem.xml');
		if(!file_exists(dirname($outputFile))) {
			mkdir(dirname($outputFile), 0777, true);
		}
		$domOutput->save($outputFile);
		
		$session = Session::getSession();
		$Lang = $session->getLanguage();
		return $Lang->getText('ProjectTemplate.SuccessfulExport') . ' ' . $this->dateString . "\n";
	}
	
	public function doImport($dataDir, $overwrite=false, $merge=false, $projectName) {
		
		if($overwrite && $merge) {
			// You can't overwrite AND merge at the same time.
			throw new Exception('Cant overwrite and merge a project at the same time');
		}
		
		// Validate the project name (taken from Project.php)
		// Verify that the imported project name doesnt exist
		$project = ProjectPeer::retrieveByUnixName($projectName);
		if(is_object($project)) {
			throw new Exception(Language::getSessText('ProjectTemplate.ProjectExists'));
		}
		if (!ereg(Project::UNIX_NAME_REGEXP, $projectName)) {
			throw new Exception(Language::getSessText('Project.InvalidUnixName'));
			return;
		}
		if (is_numeric($projectName)) {
			throw new Exception(Language::getSessText('Project.NumericUnixName'));
			return;
		}
				
		$arrTrackerHierarchy = array();

		
		$con = Propel::getConnection(ProjectPeer::DATABASE_NAME);
		$con->begin();
		try {
			$dbInfo = $con->getDatabaseInfo();

			$this->processSpecData();

			//$projectNameByDir = array();
			//ereg("/[^/]*/[^/]*$", $dataDir, $projectNameByDir);
			//ereg("^/[^/]*", $projectNameByDir[0], $projectNameByDir);
			//$projectNameByDir = Validator::sanitizeString(substr($projectNameByDir[0], 1));

			// Check if the project already exists. Check the name by directory, and by the 'project_name' key.
			/*$prjNameArray = $this->findKeyInSpec('project_name');
			$tableName = Validator::sanitizeString($prjNameArray['table']);
			$columnName = Validator::sanitizeString($prjNameArray['column']);*/

			$this->loadExportedData($dataDir);

			//$projectNameByXML = Validator::sanitizeString($this->findFirstValueInExportedData($tableName, $columnName));

			// The project name obtained via directory and via XML data must match.

			//if ($projectNameByXML != $projectNameByDir) {
			//throw new Exception('Invalid XML Data Provided. Directory Name and Project Name inside XML data dont match');
			//}

			// Check if the project already exists.

			/*$c = new Criteria();
			$c->add("$tableName.$columnName", $projectNameByXML);
			$this->project = ProjectPeer::doSelectOne($c);*/


			// Determine the import mode
			$importMode = 'standard';
			/*
			if(is_object($this->project)) {
			// The project already exists
			if($overwrite) {
			$importMode = 'overwrite';
			}
			elseif ($merge) {
			$importMode = 'merge';
			}
			else {
			throw new Exception('Cant do standard import, project already exists.');
			}
			}*/


			switch($importMode) {
				/*case 'overwrite':
				// If we are overwriting, erase all project data, then go to the standard import mode.
				$this->loadDBData();
				$this->deleteAllProjectData();
				$importMode='standard';
				// Now we go into standard import mode.*/
				case 'standard':
					// We need to bootstrap the import process by creating a new project.
					/*$con = Propel::getConnection(ProjectPeer::DATABASE_NAME);
					$idGen = $con->getIdGenerator();
					$projectNewKey = $idGen->getId($tableName. "_" . $tableName . "_id_seq");

					$sql = "INSERT INTO \"$tableName\" ($columnName) VALUES ('$projectName')";
					$stmt = $con->createStatement();
					$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_ASSOC);
					$c->add("$tableName.$columnName", $projectNameByXML);
					$this->project = ProjectPeer::doSelectOne($c);*/


					break;
				default:
					break;
			}


			// This array will hold the translated ids.
			$translateArray = array();

			// Insert a fake spec for the filesystem, so we can import the filesystem info.
			/*$this->insertFakeSpec('filesystem');*/

			// Process the spec data, fetch the exported data from the XML and send it to the DB
			$tblQ = $this->tableQueue;

			$idGen = $con->getIdGenerator();
			$dbMap = Propel::getDatabaseMap(ProjectPeer::DATABASE_NAME);

			for($tblQCounter=0;$tblQCounter < count($tblQ); $tblQCounter++) {
				$name = $tblQ[$tblQCounter];

				try {
					$tableMap = $dbMap->getTable($name);
					$keyInfo = $tableMap->getPrimaryKeyMethodInfo();
				}
				catch (Exception $e) {
					$tableMap=null;
					$keyInfo=null;
				}




				for($phpSpecDataIndex=0; $phpSpecDataIndex < count($this->phpSpecData[$name]); $phpSpecDataIndex++) {
					if(!isset($this->phpSpecData[$name][$phpSpecDataIndex]['imported'])) {
						$this->phpSpecData[$name][$phpSpecDataIndex]['imported'] = false;
					}
					if (!$this->phpSpecData[$name][$phpSpecDataIndex]['imported']) {

						$tbl = $this->phpSpecData[$name][$phpSpecDataIndex];

						// Process all the table's keys as stated in the spec, adjust the SQL query accordingly.
						// Add the links to the foreign keys.
						reset($tbl['keys']);
						foreach($tbl['keys'] as $keys_array) {
							reset($keys_array);
							$key_name = $keys_array['name'];
							foreach ($keys_array['types'] as $key_type) {
								switch($key_type) {
									case 'project_id':
										$tbl['fk'][$key_name]=array('table' => 'project', 'column' => 'project_id');
										break;
									case 'link':
										// Linked field, this applies a foreign-key like restriction to a key.
										$values = explode('.', $keys_array['value']);
										if (count($values)==2) {
											$tbl['fk'][$key_name]=array('table' => $values[0], 'column' => $values[1]);
										}
									case 'importPK':
										if(array_search($key_name, $tbl['pk'])===false) {
											$tbl['pk'][]=$key_name;
										}
									default:
										break;
								}
							}
						}

						// Determine if all needed foreign data is present, if its not push the table for later processing.
						reset($tbl['fk']);
						$skipTable = false;

						foreach($tbl['fk'] as $fkOrig => $fkDest) {
							// Check if the FK is not of a special type...
							reset($tbl['keys']);
							$skipFK = false;
							foreach($tbl['keys'] as $keys_array) {
								reset($keys_array);
								$key_name = $keys_array['name'];
								if($key_name == $fkOrig) {
									foreach ($keys_array['types'] as $key_type) {
										switch($key_type) {
											case 'importPK':
											case 'link':
											case 'exportLink':
											case 'project_id':
											case 'exportIgnore':
												break;
											default:
												$skipFK=true;
												break;
										}
									}
								}
							}

							/*print ("Table $name -> FK: ${fkDest['table']}...");
							print $skipFK ? 'skip':'';
							print "<br>";*/
							if(!$skipFK) {
								if (array_key_exists($fkDest['table'], $this->phpSpecData)) {
									if( $name != $fkDest['table']) {
										foreach($this->phpSpecData[$fkDest['table']] as $fkd)
										{
											if(!isset($fkd['imported']) || !$fkd['imported']) {
												$skipTable = true;
												$tblQ[]=$name;
												break;
											}
										}
										if($skipTable) break;
									}
								}
								else {
									// The foreign table is not present, probably the plugin is disabled so just ignore the entry.
									$skipTable = true;
									$this->phpSpecData[$name][$phpSpecDataIndex]['imported']=true;
								}
							}
							if ($skipTable) {
								break;
							}
						}


						if(!$skipTable) {

							$tableInfo = $dbInfo->getTable($name);

							if(!isset($translateArray[$name])) {
								$translateArray[$name] = array();
							}
							$this->phpSpecData[$name][$phpSpecDataIndex]['imported']=true;

							// All data is there, roam through the imported data and process it...
							foreach($this->importedXMLData as $domDoc) {
								$domTables = $domDoc->getElementsByTagName('table');
								foreach($domTables as $domTable) {
									if(strcasecmp($domTable->getAttribute('name'), $name)==0) {
										//$domRows = $domTable->childNodes;
										for($domRowCounter=0;$domRowCounter < $domTable->childNodes->length; $domRowCounter++) {
											$domRow = $domTable->childNodes->item($domRowCounter);
											if($domRow->nodeName == 'row') {
												$columns=array();
												$skipRow=false;
												$domColumns = $domRow->childNodes;
												for($domColCounter=0;$domColCounter < $domRow->childNodes->length; $domColCounter++) {
													$domColumn = $domRow->childNodes->item($domColCounter);
													if($domColumn->nodeName=='column') {
														$columnName = $domColumn->getAttribute('name');
														$columnValue = urldecode($domColumn->textContent);

														// Update Project ID and map old IDs to new IDs.
														$skipTranslation = false;
														reset($tbl['keys']);

														foreach($tbl['keys'] as $keys_array) {
															reset($keys_array);
															$key_name = $keys_array['name'];
															if($key_name == $columnName) {
																foreach ($keys_array['types'] as $key_type) {
																	switch($key_type) {
																		case 'system':
																		case 'ignore':
																		case 'importIgnore':
																			$skipTranslation = true;
																			break;

																		case 'nullify':
																			// Might have been already nulled, but just the same...
																			$columnValue='';
																			$skipTranslation = true;
																			break;

																		case 'map':
																			// Do a "in-house" translation of the value
																			// Instead of translating from other tables, this value is translated from the current table
																			if($columnValue!=0){
																				if(isset($translateArray[$name][$keys_array['value']][$columnValue])) {
																					$columnValue = $translateArray[$name][$keys_array['value']][$columnValue];
																				}
																				else {
																					// Translated value not found yet.
																					// Push the row to the end.
																					$domTable->appendChild($domRow);
																					$skipRow = true;
																				}
																			}
																			$skipTranslation = true;
																			break;

																		case 'project_id':
																			if( $name==ProjectPeer::TABLE_NAME ) {	// Small hack so we can start the importing process somewhere
																				$skipTranslation = true;
																			}
																			break;

																		case 'user_id':
																			// User ID Field, map the unix_name stored in the export against an id.
																			$mappedUser = UserPeer::retrieveByUnixName($columnValue);
																			if(is_object($mappedUser)) {
																				$columnValue = $mappedUser->getUserId();
																			//} else $columnValue = 100;	// User not found, so set id to 100,nobody user.
																			} else {
																				throw new InvalidUserException("Error importing. User $columnValue is not present in the system.");
																			}

																			$skipTranslation = true;
																			break;

																		case 'plugin_id':
																			// Plugin ID Field, map the plugin_name stored in the export against an id.
																			$plugCrit = new Criteria();
																			$plugCrit->add(PluginPeer::PLUGIN_NAME, $columnValue);
																			$mappedPlugin = PluginPeer::doSelectOne($plugCrit);

																			if(is_object($mappedPlugin)) {
																				$columnValue = $mappedPlugin->getPluginId();
																			} else {
																				throw new InvalidPluginException("Plugin $columnValue not found.");
																			}

																			$skipTranslation = true;
																			break;

																		case 'mailing_list_name':
																			// Add the unix_name of the project to the mailing list name
																			$columnValue = "$projectName-$columnValue";
																			$skipTranslation = true;
																			break;

																		case 'fixed':
																			// If the column is fixed and the content doesnt match the fixed value, skip this row (its for another set of results of this same table).
																			if($keys_array['value']!=$columnValue) {
																				$skipRow=true;
																			}
																			$skipTranslation = true;
																			break;

																		default:
																			break;
																	}
																}
																if($skipRow) break;
															}
														}

														if($skipRow) break;

														if(!$skipTranslation) {
															// Map old IDs to new IDs.
															if(array_key_exists($columnName, $tbl['fk']))
															{
																// If its set, translate it, if not  leave it alone (it can happen for certain user ids).
																if(isset($translateArray[$tbl['fk'][$columnName]['table']][$tbl['fk'][$columnName]['column']][$columnValue])) {
																	$columnValue = $translateArray[$tbl['fk'][$columnName]['table']][$tbl['fk'][$columnName]['column']][$columnValue];
																}
															}
														}

														/* Ugly hack to implant the new project name */
														if(strcasecmp($name,ProjectPeer::TABLE_NAME)==0 && strcasecmp(ProjectPeer::TABLE_NAME . ".$columnName",ProjectPeer::UNIX_NAME)==0) {
															$columnValue = $projectName;
														}

														$columns[$columnName] = $columnValue;

													}
												}

												foreach($tbl['keys'] as $keys_array) {
													reset($keys_array);
													$key_name = $keys_array['name'];
													if($keys_array['types'][0]=='fixedJoin') {
														$names = explode('.', $key_name);
														if(count($names)==2) {
															$tn = $names[0];
															$cn = $names[1];
															$v =  $keys_array['value'];

															foreach($tbl['fk'] as $col => $fk) {
																if($fk['table']==$tn) {
																	if(array_key_exists($col, $columns))
																	{
																		$sql2 = "SELECT COUNT(*) AS count FROM \"$tn\" WHERE $cn='$v' AND " . $fk['column'] . '=' . $columns[$col];
																		$con2 = Propel::getConnection(ProjectPeer::DATABASE_NAME);
																		$stmt2 = $con2->createStatement();
																		$rs2 = $stmt2->executeQuery($sql2, ResultSet::FETCHMODE_ASSOC );
																		$rs2->next();
																		$count = $rs2->getInt('count');
																		if($count==0) {
																			$skipRow=true;
																		}

																	}

																	break;
																}
															}
														}
													}
												}


												// Send everything to the DB
												if(!$skipRow) {


													switch($importMode) {
														/*case 'merge':

														$sql = "UPDATE \"$name\" SET ";
														$sqlWhere = ' WHERE ';
														$whereOffset=0;

														foreach($columns as $columnName => $columnValue) {
														// Skip primary and unique keys...
														//$columnInfo = $tableInfo->getColumn($columnName);
														if(array_search($columnName, $tbl['pk'])===false && array_search($columnName, $tbl['unique'])===false) {
														$sql .= "$columnName=?,";
														$whereOffset++;
														}
														else {
														$sqlWhere .= "$columnName=? AND ";
														$translateArray[$name][$columnName][$columnValue]=$columnValue;
														}
														}

														$sqlWhere = ereg_replace('AND $', '', $sqlWhere);
														$sqlWhere = ereg_replace(' WHERE $', '', $sqlWhere);
														$sql = substr($sql, 0, strlen($sql)-1) . $sqlWhere;
														$stmt = $con->createStatement();
														$stmt2  = $con->createStatement();
														$stmt = $con->prepareStatement($sql);
														$stmt2 = $con->prepareStatement("SELECT COUNT(*) FROM \"$name\" $sqlWhere");

														$columnCounter=1;
														$whereCounter=1;

														foreach($columns as $columnName => $columnValue) {
														// Skip primary keys...
														$columnInfo = $tableInfo->getColumn($columnName);

														if(array_search($columnName, $tbl['pk'])===false && array_search($columnName, $tbl['unique'])===false) {
														switch($columnInfo->getNativeType()) {
														case 'int4':
														$stmt->setInt($columnCounter, intval($columnValue));
														break;
														case 'timestamptz':
														if($columnValue==''){
														$stmt->setNull($columnCounter);
														}
														else {
														$stmt->setTimestamp($columnCounter, $columnValue);
														}
														break;
														case 'boolean':
														$stmt->setBoolean($columnCounter, Validator::sanitizeBool($columnValue));
														break;
														default:
														$stmt->set($columnCounter, $columnValue);
														break;
														}

														$columnCounter++;
														}
														else {
														switch($columnInfo->getNativeType()) {
														case 'int4':
														$stmt->setInt($whereCounter+$whereOffset, $columnValue);
														$stmt2->setInt($whereCounter, $columnValue);
														break;
														case 'timestamptz':
														if($columnValue==''){
														$stmt->setNull($whereCounter+$whereOffset);
														$stmt2->setNull($whereCounter);
														}
														else {
														$stmt->setTimestamp($whereCounter+$whereOffset, $columnValue);
														$stmt2->setTimestamp($whereCounter, $columnValue);
														}
														break;
														case 'boolean':
														$stmt->setBoolean($whereCounter+$whereOffset, Validator::sanitizeBool($columnValue));
														$stmt2->setBoolean($whereCounter, Validator::sanitizeBool($columnValue));
														break;
														default:
														$stmt->set($whereCounter+$whereOffset, $columnValue);
														$stmt2->set($whereCounter, $columnValue);
														break;
														}

														$translateArray[$name][$columnName][$columnValue] = $columnValue;
														$whereCounter++;
														}

														}


														$rs = $stmt2->executeQuery();
														$rs->next();
														$count = $rs->getInt('count');

														if($count>1) {
														throw new Exception("$sql<br>Invalid spec or DB data. Row in table $name cant be uniquely merged");
														break;
														}
														elseif($count == 1) {
														// Merge
														//$columnCounter starts at 1, if its still at 1 we won't update as there is nothing to update.
														if($columnCounter>1) {
														$stmt->executeUpdate();
														}
														break;
														}*/

														// We need to insert, just let it slide into the next case

														case 'standard':
															//$stmt = $con->createStatement();



															$sql = "INSERT INTO \"$name\" ";
															$sqlCols = '(';
															$sqlValues = '(';
															foreach($columns as $columnName => $columnValue) {
																$sqlCols .= "$columnName,";
																$sqlValues .= "?,";
															}
															$sqlCols = substr($sqlCols, 0, strlen($sqlCols)-1) . ')';
															$sqlValues = substr($sqlValues, 0, strlen($sqlValues)-1) . ')';

															$sql .= "$sqlCols VALUES $sqlValues";

															$stmt = $con->prepareStatement($sql);
															$columnCounter=1;
															foreach($columns as $columnName => $columnValue) {
																// Skip primary keys...
																$columnInfo = $tableInfo->getColumn($columnName);
																if(array_search($columnName, $tbl['pk'])===false || !$columnInfo->isAutoIncrement()) {
																	switch($columnInfo->getNativeType()) {
																		case 'int4':
																		case 'int8':
																			try {
																				$stmt->setInt($columnCounter, intval($columnValue));
																			}
																			catch (Exception $e){
																				$stmt->setInt($columnCounter, 0);
																			}
																			break;
																		case 'timestamptz':
																			if($columnValue==''){
																				$stmt->setNull($columnCounter);
																			}
																			else {
																				$stmt->setTimestamp($columnCounter, $columnValue);
																			}
																			break;
																		case 'bool':
																			try {
																				$stmt->setBoolean($columnCounter, Validator::sanitizeBool($columnValue));
																			}
																			catch (Exception $e){
																				$stmt->setBoolean($columnCounter, false);
																			}
																			break;
																		default:
																			try {
																				$stmt->set($columnCounter, $columnValue);
																			}
																			catch (Exception $e){
																				$stmt->set($columnCounter, '');
																			}
																			break;
																	}

																}
																else {
																	$translateArray[$name][$columnName][0] = 0;	// This is a workaround for some ref_id columns that may contain a zero value.
																	try {
																		$columnInfo = $tableInfo->getColumn($columnName);
																		if($keyInfo!=null) {
																			$newPK = $idGen->getId($keyInfo);
																		}
																		else {
																			$newPK = $idGen->getId($name. "_" . $columnName . "_seq");
																		}
																		$stmt->setInt($columnCounter, $newPK);
																		$translateArray[$name][$columnName][$columnValue] = $newPK;
																		
																		if ($name == "tracker_item") {
																			$arrTrackerHierarchy[$columnValue] = array();
																			$arrTrackerHierarchy[$columnValue]['newPK'] = $newPK;
																			$arrTrackerHierarchy[$columnValue]['parent_id'] = $columns['parent_id'];
																		}
																		
																	}
																	catch (Exception $e) {
																		$newPK = $columnValue;
																		$stmt->setInt($columnCounter, $columnValue);
																		$translateArray[$name][$columnName][$columnValue] = $columnValue;
																	}
																}
																$columnCounter++;
															}

															$rs = $stmt->executeUpdate();

															break;


														default:
															break;
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}


			//Update parent_ids
			foreach($arrTrackerHierarchy as $tracker_id => $arr_tracker) {
				$tracker_item = TrackerItemPeer::retrieveByPk($arr_tracker['newPK']);
				if (is_object($tracker_item) && $arr_tracker['parent_id'] != 0) {
					$tracker_item->setParentId($arrTrackerHierarchy[$arr_tracker['parent_id']]['newPK']);
					$tracker_item->save();
				}
			}


			// Import files. Files are imported directly from the exported filesystem (in this case filters are ignored).
			// This means that you better not change the paths specified in each filter or something might go wrong! (ie 'filesystem' path is hardcoded below)
			// Filesystem ID translation is performed.

			$exportedFiles = $this->scanDirectories(SystemUtils::buildFilePath($dataDir, 'filesystem'));

			foreach($exportedFiles as $exportedFile) {

				// Get the original ID from the path...
				ereg("filesystem/plugins/([^/]*)/([0-9]/)*([^/]*)/([^/]*)$",$exportedFile, $regs);
				$plugin_name = $regs[1];
				$old_id = $regs[3];
				$file_name = $regs[4];

				if(array_key_exists('filesystem', $translateArray) && array_key_exists('filesystem_id', $translateArray['filesystem']) && array_key_exists($old_id, $translateArray['filesystem']['filesystem_id'])) {
					$new_id = $translateArray['filesystem']['filesystem_id'][$old_id];

					$outputFile = SystemUtils::buildFilePath(GFConf::get('system.filesPath'), $plugin_name);

					/*$idLength = strlen($new_id.'');
					for($fpc=0; $fpc < $idLength && $fpc < 3; $fpc++) {
					$outputFile = SystemUtils::buildFilePath($outputFile, substr($new_id,$fpc,1));
					}*/
					$padded_id = str_pad($new_id, 3, '0', STR_PAD_LEFT);
					$outputFile = SystemUtils::buildFilePath($outputFile, '/'.$padded_id[0].'/'.$padded_id[1].'/'.$padded_id[2]);

					$outputFile = SystemUtils::buildFilePath($outputFile, $new_id);
					$outputFile = SystemUtils::buildFilePath($outputFile, $file_name);

					if(!file_exists(dirname($outputFile))) {
						//mkdir(dirname($outputFile), 0777, true);
						$output = '';
						SystemUtils::execute('mkdir -m 0777 -p ' . dirname($outputFile), $output);
					}
					copy($exportedFile, $outputFile);
				}
			}


			// Small hack to fix the open_count field in each tracker.
			$project = ProjectPeer::retrieveByUnixName($projectName);
			if(is_object($project)) {
				$crit = new Criteria();
				$crit->add(TrackerPeer::PROJECT_ID, $project->getProjectId());
				$trackers = TrackerPeer::doSelect($crit);
				foreach($trackers as $tracker) {
					$crit = new Criteria();
					$crit->add(TrackerExtraFieldElementPeer::STATUS_ID, Tracker::STATUS_CLOSED);
					$crit->addJoin(TrackerExtraFieldDataPeer::FIELD_DATA, TrackerExtraFieldElementPeer::ELEMENT_ID);
					$crit->addJoin(TrackerExtraFieldPeer::TRACKER_EXTRA_FIELD_ID, TrackerExtraFieldDataPeer::TRACKER_EXTRA_FIELD_ID);
					$crit->add(TrackerExtraFieldPeer::FIELD_TYPE, TrackerExtraField::TYPE_STATUS );
					$crit->add(TrackerExtraFieldPeer::TRACKER_ID, $tracker->getTrackerId());
					$close_count = TrackerExtraFieldElementPeer::doCount($crit);
					$tracker->setOpenCount($tracker->getItemTotal()-$close_count);
					$tracker->save();
				}
			}
		
			// Fix the duplicate news forums created by the SQL INSERT RULE
			// Basically deletes the first forum for each of the projects news (the one that would've been auto inserted)
			$project_id = $project->getProjectId();
			$sql = "SELECT forum_id, ref_id FROM forum WHERE section='news' AND ref_id IN (SELECT news_id FROM news WHERE project_id=$project_id) ORDER BY forum_id ASC";						
			$stmt = $con->prepareStatement($sql);
			$rs = $stmt->executeQuery(ResultSet::FETCHMODE_NUM );
			$forums = array();
			
			while($rs->next()) {
				$forum_id = $rs->getInt(1);
				$ref_id = $rs->getInt(2);
				
				if(!isset($forums[$ref_id])) {
					$forums[$ref_id] = $forum_id;
				}				
			}
			
			if(count($forums)>0) {
				$forum_ids = implode(',', $forums);
				$sql = "DELETE FROM forum WHERE forum_id IN ($forum_ids)";
				$stmt = $con->prepareStatement($sql);
				$rs = $stmt->executeQuery();
			}
			
		} catch (SQLException $pe) {
			$con->rollback();
			throw new InvalidDataException(array($pe->getMessage()));
		} catch (Exception $e) {
			// If there was any error importing, roll back everything.
			$con->rollback();
			throw $e;
		}
		$con->commit();		
	}
	
	public function export($msg="PROJECT_FULLEXPORT", $systemconf='lib/systemSpecDataFull.xml') {
		if (is_null($this->project)) {
			throw new Exception('GFExportImport->export error. No project set.');
		}
		
		$this->dbData = array();
		$results = ActionManager::notify($msg, $this->project,$this->project->getPlugins());
		$this->specData = array();
		foreach ($results as $r) {
			$xmlString = $r->getResult();
			// Convert the xml string
			$this->specData[] = $xmlString;
		}
		
		// Add the system wide spec data
		$this->specData[] = file_get_contents(SystemUtils::buildFilePath(GFConf::get('system.libDir'), $systemconf));

		return $this->doExport();
	}
	
	public function import($dataDir, $overwrite=false, $merge=false, $msg="PROJECT_FULLIMPORT", $projectName, $systemconf='lib/systemSpecDataFull.xml')
	{
		$this->dbData = array();
		
		$this->specData = array();
		// Add the system wide spec data
		$this->specData[] = file_get_contents(SystemUtils::buildFilePath(GFConf::get('system.libDir'), $systemconf));
		
		$pluginList = $this->getEnabledPlugins($dataDir);
		
		$results = ActionManager::notify($msg, null, $pluginList);
		foreach ($results as $r) {
			$xmlString = $r->getResult();
			// Convert the xml string
			$this->specData[] = $xmlString;
		}
		
		
		
		return $this->doImport($dataDir, $overwrite, $merge, $projectName);
	}
	
	public function zipExport() {
		if (is_null($this->project)) {
			throw new Exception('GFExportImport->export error. No project set.');
		}
		
		$outputFile = SystemUtils::buildFilePath(GFConf::get('system.exportProjects'), $this->project->getUnixName());
		$inputDirectory = SystemUtils::buildFilePath($this->project->getUnixName() , $this->dateString);
		$outputFile = SystemUtils::buildFilePath($outputFile, $this->project->getUnixName() . '.template.tar.bz2');
		$output=array();
		SystemUtils::execute("tar -C " .GFConf::get('system.exportProjects')." -cjf $outputFile $inputDirectory", $output);
		
		$deleteDirectory = SystemUtils::buildFilePath(GFConf::get('system.exportProjects'), $inputDirectory);
		if(strlen($deleteDirectory)>10) {  //Security measure to avoid accidentally deleting a major folder...
			SystemUtils::execute("rm -rf $deleteDirectory", $output);
		}
		return $outputFile;
	}
	
	public function getExportList() {
		if (is_null($this->project)) {
			throw new Exception('GFExportImport->getExportList error. No project set.');
		}
		
		$exportsDir = SystemUtils::buildFilePath(GFConf::get('system.exportProjects'), $this->project->getUnixName());
		
		$directories = scandir($exportsDir);
		$outputArray = array();
		foreach($directories as $dir) {
			if($dir != '.' && $dir != '..' && is_dir(SystemUtils::buildFilePath($exportsDir, $dir))) {
				$outputArray[] = $dir;
			}
		}
		return $outputArray;
	}
	
	public static function getAllExports() {
		$exportsDir = GFConf::get('system.exportProjects');
		$outputArray = array();
		if(is_dir($exportsDir)){
			$directories = scandir($exportsDir);

			foreach($directories as $dir) {
				if($dir != '.' && $dir != '..' && is_dir(SystemUtils::buildFilePath($exportsDir, $dir))) {
					$outputArray[$dir] = array();
					$projectExportDir = SystemUtils::buildFilePath($exportsDir, $dir);
					$projectExports = scandir($projectExportDir);
					foreach($projectExports as $projectExport) {
						if($projectExport != '.' && $projectExport != '..' && is_dir(SystemUtils::buildFilePath($projectExportDir, $projectExport))) {
							$outputArray[$dir][] = $projectExport;
						}
					}
				}
			}
		}
		
		return $outputArray;
	}
}

?>
