<?php

/**
 * Copyright (c) 2005-2012 GForge, LLC
 *
 * @version
 * @author Mario A. Papayannis	mpapayannis@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
 
/**
 * A collection of methods that deal with url filtering and rendering
 */
class LinkRenderer {

	/**
	 * Renders a tracker item link
	 * Used by renderTrackerItemRefs method
	 * @param	string
	 * @return	string
	 */
	static function renderTrackerItem($tracker_item, $summary) {
		$title = $tracker_item->getTrackerItemId().': '.$tracker_item->getSummary();
		$description = $tracker_item->getTracker()->getProject()->getProjectName().'/'.$tracker_item->getTracker()->getTrackerName().'<br />'.str_replace("\n", '<br />', $tracker_item->getDetails());
		
    if ($summary) {
      return self::renderTrackerItemSummaryTemplate($tracker_item->getTrackerItemUrl(false), $tracker_item->getTrackerItemId(), $description, $title, '', false);
    } else {
      return self::renderTrackerItemLinkTemplate($tracker_item->getTrackerItemUrl(false), $tracker_item->getTrackerItemId(), $tracker_item->getSummary());
    }
	}


	/**
	 * Given a string, detects and renders tracker items in the format "[#xxxx]"
	 * Returns the original text with the tracker items rendered
	 * @param	string
	 * @return	string
	 */
	static function renderTrackerItemRefs($text) {
    //match all [#1111] and #1111 as long as they are surrounded by spaces/edge of lines
    //the <br /> exists because the chat replaces \n with a <br /> so that it will display properly later
		preg_match_all('/((?<=<br \/>)|(?<=\s)|(?<=^))\[?(#[0-9]+)(\]?)(.*?)((?=\s)|(?=$)|(?=<br \/>))/', $text, $matches, PREG_OFFSET_CAPTURE);

		//Calculate for the added characters (generated html) because we are using an index based replace
		if (is_array($matches[2]) && count($matches[2]) > 0) {
			foreach($matches[0] as $match) {
				preg_match('/[0-9]+/', $match[0], $id_match);

				if (count($id_match > 0)) {
					$tid = $id_match[0];
				} else {
					continue;
				}

				//save any text after #n to reapply as suffix
				$wordSuffix = preg_replace('/^.*?[0-9]+\]?(.*?)/', '$1', $match[0]);
				$hasBrackets = false;
				if (preg_match('/\[#[0-9]+\]/', $match[0])) {
					$hasBrackets = true;
				}

				//Get Tracker Item from Id
				$tracker_item = TrackerItemPeer::retrieveByPk($tid);
				if (is_object($tracker_item) && $tracker_item->userCanRead()) {
					//dont include summary if it doesnt have brackets
					$str_rendered_tracker_item = LinkRenderer::renderTrackerItem($tracker_item, $hasBrackets);
					$text = str_replace($match[0], $str_rendered_tracker_item . $wordSuffix, $text);
				} else {
				}
			}
		}
		return $text;
	}


	/**
	 * Verifies if the passed url is a full GFAS url (includes the hostname)
	 * @param	string
	 * @return	boolean
	 */
	static function isFullGFUrl($url) {
		$str_hostname = GFConf::get('system.hostName');
		
		return preg_match('/'.$str_hostname.'\/gf\//i', $url);
	}


	/**
	 * Parses the url and retrieves a GFUrl object
	 * @param	string
	 * @return	object
	 */
	static function parseGFUrl($url) {
		$gf_url = new GFUrl();
		$url = LinkRenderer::convertGFUrlToRelative($url);

		$gf_url->parseURI($url);
		return $gf_url; //Returns a GFUrl with the parsed url
	}


	/**
	 * Checks if the url is a GFAS url
	 * Receives a GFUrl object with the parsed url
	 * @param	object
	 * @return	boolean
	 */
	static function isGFUrl($gf_url, $url) {
		$str_hostname = GFConf::get('system.hostName') . '/gf/';

		if ($gf_url->getProject() != null || $gf_url->getUser() != null || ($gf_url->getPlugin() && $gf_url->getSubdir() != '') || strpos($url, $str_hostname) != -1) {
			return true;
		} else {
			return false;
		}
	}


	/**
	 * Converts a Full GFAS url to a relative url (removes the domain)
	 * @param	string
	 * @return	string
	 */
	static function convertGFUrlToRelative($url) {
		$str_hostname = GFConf::get('system.hostName');
		$url = preg_replace('/(https?|ftp):\\/\\/'.$str_hostname.'\//i', '', $url);
		$url = rtrim($url, '/');
		return $url;
	} 
	
	static function preprocessGFUrls($text) {
		$str_hostname = GFConf::get('system.hostName');

		preg_match_all('/(https?):\\/\\/'.$str_hostname.'\/gf\//i', $text, $matches);
			
		if (is_array($matches[0]) && count($matches[0]) > 0) {
			$added_links = array();			// this will avoid nesting links
			foreach ($matches[0] as $link) {			
				if (array_search($link, $added_links) === false) {		// link was not replaced

					$link_edited = preg_replace('/(https?):\\/\\/'.$str_hostname.'\/gf\//i', '/gf/' , $link);
					$text = str_replace($link, $link_edited, $text);
					$added_links[] = $link;
				}
			}
		}

	}
	

	/**
	 * Renders a generic internal GFAS url
	 * @param	string
	 * @return	string
	 */	
	static function renderGFUrlHtml($message, $url) {
		return '<a href="'.$url.'">'.$message.'</a>';
	}


	/**
	 * Given a GFAS url, it detects the section (user url, project url, etc) and plugin and renders the url accordingly
	 * For plugin urls, it calls the appropiate plugin to deal with the url rendering
	 * @param	string
	 * @return	string
	 */
	static function renderGFUrl($url, $gf_url) {
		$parsed_url = parse_url(html_entity_decode($url));
	 	if (isset($parsed_url['query'])) {
	 		parse_str($parsed_url['query'], $arr_query);
	 	}

	 	$str_rendered_link = '';

		if ($gf_url->getPlugin() != null) {
			$params = array();
			$params['url'] = $url;
			$params['gf_url'] = $gf_url;
			$params['query_string'] = $arr_query;
			$plugin_results = ActionManager::notify('RENDER_PLUGIN_LINK', $params);
			foreach ($plugin_results as $result) {
				$str_rendered_link = $result->getResult();
				if ($str_rendered_link != '') {
					break;
				}
			}
		} else if ($gf_url->getUser() != null) {
			//It's a user url
		} else if ($gf_url->getPlugin() && $gf_url->getSubdir() != '') {
			//It's some other url. We dont' provide a custom renderer in this case
		} else if (strpos($url, '/gf/download/') != -1) {
			$components = $gf_url->getRelativeURI();
			$arr = explode('/',$components);
			if (count($arr)==5 || (count($arr)==6 && $arr[5]='open')) {
				$section = $arr[1];
				$ref_id = $arr[2];
				$file_id = $arr[3];	
				$file_name = urldecode($arr[4]);

				$params = array();
				$params['url'] = $url;
				$params['section'] = $section;
				$params['ref_id'] = $ref_id;
				$params['file_id'] = $file_id;
				$params['file_name'] = $file_name;
				$plugin_results = ActionManager::notify('RENDER_DOWNLOAD_URL', $params);
				
				foreach ($plugin_results as $result) {
					$str_rendered_link = $result->getResult();
					if ($str_rendered_link != '') {
						break;
					}
				}
			}
		}

		if ($str_rendered_link == '') {
			$str_rendered_link = '
				<div class="link-summary">
					<p class="title">
						GForge url
					</p>
					<p class="url">
						<a href="'.$url.'" target="_blank">'.$url .'</a>
					</p>
				</div>';
		}

		return $str_rendered_link;
	}
	


	/**
	 * Given an external (non-GFAS) url, it performs a request to google using the safe browsing API and attempts to identify
	 * if it's a safe url
	 * @param	string
	 * @return	array
	 */
	static function validateSafeUrl($link) {
		//If outbound connections are not allowed, this feature is diabled
		try {
			$outboundConnections = GFConf::get('system.outboundConnections');
		} catch(Exception $e) {
			$outboundConnections = false;
		}
		if (!$outboundConnections) {
			$res['is_safe'] = true;
			$res['result'] = '';
		}

		//If the safe browsing key does not exist, this feature is disabled
		try {
			$apiKey = GFConf::get('system.safeBrowsingKey');
		} catch(Exception $e) {
			$apiKey = false;
		}
		if ($apiKey == '') {
			$res['is_safe'] = true;
			$res['result'] = '';
		}


		$url = 'https://sb-ssl.google.com/safebrowsing/api/lookup?client=api&apikey=' . $apiKey . '&appver=1.0&pver=3.0&url=' . urlencode($link);

		$ch = curl_init();
		curl_setopt($ch, CURLOPT_URL, $url);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		curl_setopt($ch, CURLOPT_HTTPGET, true);
		curl_setopt($ch, CURLOPT_FAILONERROR, true);

		$response = curl_exec($ch);
		$err = curl_error($ch);
		curl_close($ch);
		
		$res = array();
		
		if ($response != '') {
			$res['is_safe'] = false;
			$res['result'] = $response;
			return $res;
		} else {
			$res['is_safe'] = true;
			$res['result'] = '';
			return $res;
		}
	}


	/**
	 * Parses a url, determines if it's a GFAS filesystem url pointing to an image, and renders the corresponding HTML
	 * @param	string
	 * @return	string
	 */
	static function renderGFASImageUrl($link) {
		$str_rendered_link = '';
		$pos = strpos($link, "/gf/download/");
		if ($pos >= 0) {
			$link = substr($link, $pos, strlen($link));
		}
		$path_components = explode('/', $link);
		if  ($path_components[1] == 'gf' && $path_components[2] == 'download') {
			$filename = $path_components[6];
			$filename_components = explode('.', $filename);
			$file_extension = strtolower($filename_components[sizeof($filename_components)-1]);
			$arr_types = array("png", "jpg", "jpeg", "gif", "png");
			if (in_array($file_extension, $arr_types)) {
				$linkaux = str_replace("/open", "", $link);
				$str_rendered_link = LinkRenderer::renderImage($link, $linkaux);	
			}
		}
		return $str_rendered_link;
	}


	/**
	 * Given a text, renders the links in it
	 * @param	string
	 * @return	string
	 */
	static function renderHtmlLinks($text) {
		
		preg_match_all('/[^(href=")](https?|ftp):\\/\\/[^[:blank:][:cntrl:]<>"\']*/i', $text, $matches);	
		
		$text = preg_replace("/<a[^>]+>/i", "", $text);
		
		if (is_array($matches[0]) && count($matches[0]) > 0) {
			$added_links = array();			// this will avoid nesting links
			foreach ($matches[0] as $link) {
				if (preg_match('/\\.\\.\\.$/i', $link)) {
					// This ensures that link wasn't wrapped. For example, $text may end
					// with "see http://www.yahoo..."
					continue;
				}
				// remove last dot, if any (may be confused with part of the URL)
				$link = preg_replace('/\\.$/', '', $link);
				$link = preg_replace('/>/', '', $link);
				if (array_search($link, $added_links) === false) {		// link was not replaced
					//$text = str_replace($link, '<a href="'.$link.'">'.$link.'</a>', $text);
					$added_links[] = $link;

					$str_rendered_link = $link;

					//Check if the link is a GForge link	
					$gf_url = '';
					try {				
						$gf_url = LinkRenderer::parseGFUrl($link);
					} catch(Exception $err) {
						$gf_url = null;
					}

					if (is_object($gf_url) && LinkRenderer::isGFUrl($gf_url, $link)) {
						$str_rendered_link = LinkRenderer::renderGFUrl($link, $gf_url);
						$str_escaped_link = preg_quote($link);
						$str_escaped_link = str_replace('/', '\\/', $str_escaped_link);
						$text = str_replace($link, $str_rendered_link, $text);
						continue;
					}

					$path_parts = pathinfo($link);
					$filetype = '';

					$str_rendered_link = LinkRenderer::renderGFASImageUrl($link);
					if ($str_rendered_link != '') {
						$text = str_replace($link, $str_rendered_link, $text);
						continue;
					}

					if (array_key_exists('extension',$path_parts)) {
						$filetype = strtolower($path_parts['extension']);
						switch($filetype) {							
							case "png":
							case "jpg":
							case "jpeg":
							case "gif":
							case "png":
								$str_rendered_link = LinkRenderer::renderImage($link);	
							break;
							
							default:						
								$res = LinkRenderer::validateSafeUrl($link);

								if ($res['is_safe']) {
									$str_rendered_link = LinkRenderer::renderHtmlLinkSummary($link);
								} else {
									$str_rendered_link = LinkRenderer::renderUnsafeHtmlLink($link, $res['result']);
								}
							break; 
						}
					} else if (preg_match('/youtube.com\/watch/i', $link) || (preg_match('/youtu.be/i', $link))) {
						$str_rendered_link = LinkRenderer::renderYoutubeVideo($link);
					} else {
						$res = LinkRenderer::validateSafeUrl($link);

						if ($res['is_safe']) {
							$str_rendered_link = LinkRenderer::renderHtmlLinkSummary($link);
						} else {
							$str_rendered_link = LinkRenderer::renderUnsafeHtmlLink($link, $res['result']);
						}
					}

					$text = str_replace($link, $str_rendered_link, $text);

				}
			}
		}
		
		//Render tracker item references in the format: [#xxxx]
		$text = LinkRenderer::renderTrackerItemRefs($text);
		return $text;
	}


	/**
	 * Renders an image url
	 * @param	string
	 * @return	string
	 */
	static function renderImage($link, $link2='') {
		if ($link2 == '') {
			$link2 = $link;	
		}
		return '<br /><a target="_blank" href="'.$link.'"><img class="rendered_image_link" src="' . $link2 . '" /></a><br />';
	}


	/**
	 * Renders a url that has been fed to Google's Safe Browsing and informed as dangerous
	 * @param	string	url to show
	 * @param string	message that describes the type of danger the url imposes
	 * @return	string
	 */
	static function renderUnsafeHtmlLink($link, $message) {
		return '<span style="color: red;">This url is unsafe (' . $message . '):</span>&nbsp' . $link;
	}

	/**
	 * Renders a youtube video url
	 * @param	string
	 * @return	string
	 */
	static function renderYoutubeVideo($link) {
		$link = LinkRenderer::getFinalUrl($link);
	
		$oembed_url = "http://www.youtube.com/oembed?url=" . $link . "&format=json";
		$ch = curl_init();
		$timeout = 10;
		curl_setopt($ch, CURLOPT_URL, $oembed_url);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
		curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, $timeout);
		curl_setopt($ch,CURLOPT_USERAGENT,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.13) Gecko/20080311 Firefox/2.0.0.13');
			
		$string = curl_exec($ch);
		curl_close($ch); 	

		if ($string == '401 Unauthorized') {
			return LinkRenderer::renderHtmlLinkSummary($link);
		}

		//Detect if the current URL is http or https. This prevents browsers from blocking content when GFAS is running under SSL and the content is HTTP
		if (isset($_SERVER['HTTPS']) && $_SEREVER['HTTPS'] != 'off') {
			$proto = 'https';
		} else {
			$proto = 'http';
		}

		preg_match("/v=(\w+)/", $link, $matches);
		return '<br /><iframe src="'.$proto.'://www.youtube.com/embed/'.$matches[1].'?html5=1" allowfullscreen></iframe><br />';
	}


	/**
	 * Renders external urls when there are no outbound connections and there's no way to render the summary-style url
	 * @param	string	url to show
	 * @return	string
	 */
	static function renderBasicHtmlLink($link) {
		return '<a href="' . $link . '" target="_blank">' . $link . '</a>';
	}


	/**
	 * Renders a an external url
	 * @param	string
	 * @return	string
	 */
	static function renderHtmlLinkSummary($link) {
		//If outbound connections are not allowed, this feature is diabled
		if (!GFConf::get('system.outboundConnections')) {
			return LinkRenderer::renderBasicHtmlLink($link);
		}
		
		$value = trim($link);
		if (get_magic_quotes_gpc()) {
			$value = stripslashes($value);
		}
		$value = strtr($value, array_flip(get_html_translation_table(HTML_ENTITIES)));
		$value = strip_tags($value);
		$url = htmlspecialchars($value);
			
		$link = str_replace('>', '', $link);
		$link = htmlspecialchars_decode($link);
		
		$link = strip_tags($link);

		$string = '';
		$link_final = LinkRenderer::getFinalUrl($link);

		$ch = curl_init();
		$timeout = 10;
		curl_setopt($ch, CURLOPT_URL, $link_final);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
		curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, $timeout);
		curl_setopt($ch,CURLOPT_USERAGENT,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.13) Gecko/20080311 Firefox/2.0.0.13');
			
		$string = curl_exec($ch);
		curl_close($ch); 	
		/// fecth title
		$title_regex = "/<title>(.+)<\/title>/i";
		preg_match($title_regex,$string,$matches);
		$str_title = isset($matches[1]) ? trim($matches[1]) : '';

		/// fecth decription
		$tags = get_meta_tags($url);
		 
		// fetch images
		$image_regex = '/<img[^>]*'.'src=[\"|\'](.*)[\"|\']/Ui';
		preg_match_all($image_regex, $string, $img, PREG_PATTERN_ORDER);
		$images_array = $img[1];

		$k=1;
		$str_image = '';
		for ($i=0;$i<=1;$i++) {
			if(@$images_array[$i]) {
				if(@getimagesize(@$images_array[$i]))	{
					list($width, $height, $type, $attr) = getimagesize(@$images_array[$i]);
					if($width >= 200 || $height >= 200 ){
						$str_image = "<img src='".@$images_array[$i]."'>";
						$k++;
					}
				}
			}
		}

		$str_description = isset($tags['description']) ? trim($tags['description']) : '';

		return self::renderHtmlLinkSummaryTemplate($url, $str_description, $str_title, $str_image);
	}

	static function renderHtmlLinkSummaryTemplate($url, $description = '', $title = '', $imageTag = '', $show_url=true) {
		$ret = '<div class="link-summary">';

		if (strlen($imageTag) > 0) {
			$ret .='
				<div class="images"><a href="' . $url . '" target="_blank">
				'.$imageTag.'
					</a>
				</div>';
		}
		if (strlen($title) > 0) {
			$title = StringUtils::getPlainUTF8FromHtml($title);
			$ret .= '
			<p class="title">
				<a href="' . $url . '" target="_blank">'.$title.'</a>
			</p>';
		}
		if (strlen($url) > 0 && $show_url) {
			$ret .='
				<p class="url">
					<a href="'.$url.'" target="_blank">'.$url .'</a>
				</p>';
		}
		if (strlen($description) > 0) {
			$description = StringUtils::getPlainUTF8FromHtml($description);
			$ret .='
			<p class="desc chat-autolink-description">
				'.$description.'
			</p>';
		}
		$ret .='
			</div>';

		return $ret;
	}



	static function renderHtmlDocumentSummaryTemplate($url, $description, $title, $download_url, $modification_date, $version_number) {
		$ret = '<div class="link-summary">';

		if (strlen($title) > 0) {
			$ret .= '
			<p class="title">
				<a href="'.$url.'" target="_blank">'.$title.'</a>
			</p>';
		}

		if (strlen($description) > 0) {
			$ret .='
			<p class="desc">
				download/open: <a href="'.$download_url.'" target="_blank">'.$download_url.'</a><br />'.'
				modification date: '.$modification_date.'<br />'.'
				version_number: '.$modification_date.'<br />'.'
				description: '.$description.'
			</p>';
		}
		$ret .='
			</div>';

		return $ret;
	}


	//static function renderDocumentDetailsTemplate($docman_url, $download_url, $docman_file_id, $description = '', $filename = '', $show_url=true) {
	static function renderDocumentVersionDetailsTemplate($params) {
		$ret = '<div class="link-summary">';

		if (strlen($params['filename']) > 0) {
			$ret .= '
				<p class="title">
				';
			if (Session::getSession()->isLoggedIn()) {
				$ret .= '
					Document: <a style="float:left;" href="javascript:;" class="discuss" data-section="docmanfile" data-ref-id="' . $params['docman_file_id'] . '" data-forum-name="' . $params['filename'] . '-v' . $params['version_number'] . '" onMouseOver="tooltip(\'Discuss this document\')" onmouseout="tooltipexit()">
						<img src="/images/chat_icon_24.png" alt="Discuss this tracker item" />
					</a>
				';				
			}
			$ret .= '<a href="'.$params['url'].'" target="_blank" style="left: 2px;" class="url">'.$params['filename'].'</a>
				</p>';
		}
	
		$ret .='
			<br />
			<p class="desc">
				<strong>description</strong>: '.$params['description'].'<br />
				<strong>version_number</strong>: '.$params['version_number'].'<br />'.'
				<strong>modification date</strong>: '.$params['modification_date'].'<br />'.'
				<strong>download/open</strong>: <a href="'.$params['download_url'].'" target="_blank">[click here]</a>'.'
			</p>';

		$ret .='
			</div>';

		return $ret;
	}


	static function renderTrackerItemSummaryTemplate($url, $tracker_item_id, $description = '', $title = '', $imageTag = '', $show_url=true) {
		$ret = '<div class="link-summary">';

		if (strlen($imageTag) > 0) {
			$ret .='
				<div class="images"><a href="' . $url . '" target="_blank">
				'.$imageTag.'
					</a>
				</div>';
		}
		if (strlen($title) > 0) {
			$ret .= '
				<p class="title">
				';
			if (Session::getSession()->isLoggedIn()) {
				$ret .= '
					<a style="float:left;" href="javascript:;" class="discuss" data-section="trackeritem" data-ref-id="' . $tracker_item_id . '" data-forum-name="' . $title . '" onMouseOver="tooltip(\'Discuss this tracker item\')" onmouseout="tooltipexit()">
						<img src="/images/chat_icon_24.png" alt="Discuss this tracker item" />
					</a>
				';				
			}
			$ret .= '<a href="' . $url . '" target="_blank" style="left: 5px;">'.$title.'</a>
				</p>';
		}

		if (strlen($url) > 0 && $show_url) {
			$ret .='
				<p class="url">
					<a href="'.$url.'" target="_blank">'.$url .'</a>
				</p>';
		}
		if (strlen($description) > 0) {
			$ret .='
			<div class="chat-autolink-description" readmore-url="'.$url.'">
				'.$description.'
			</div>
			<div class="chat-autolink-description-bkp" style="display: none;">
			</div>
			';
		}

		$ret .='
			</div>';

		return $ret;
	}

  static function renderTrackerItemLinkTemplate($url, $id, $title) {
    $title = htmlspecialchars($title, ENT_QUOTES);
    $ret = "<b><a title='$title' href='$url'>#$id</a></b>";
    return $ret;
  }



	/**
	 * Gets the final url (follows redirects)
	 * @param	string
	 * @param int
	 * @return	string
	 */
	static function getFinalUrl( $url, $timeout = 10 ) {
		$url = str_replace( "&amp;", "&", urldecode(trim($url)) );

		$cookie = tempnam ("/tmp", "CURLCOOKIE");
		$ch = curl_init();
		curl_setopt( $ch, CURLOPT_USERAGENT, "Mozilla/5.0 (Windows; U; Windows NT 5.1; rv:1.7.3) Gecko/20041001 Firefox/0.10.1" );
		curl_setopt( $ch, CURLOPT_URL, $url );
		curl_setopt( $ch, CURLOPT_COOKIEJAR, $cookie );
		curl_setopt( $ch, CURLOPT_FOLLOWLOCATION, true );
		curl_setopt( $ch, CURLOPT_ENCODING, "" );
		curl_setopt( $ch, CURLOPT_RETURNTRANSFER, true );
		curl_setopt( $ch, CURLOPT_AUTOREFERER, true );
		curl_setopt( $ch, CURLOPT_CONNECTTIMEOUT, $timeout );
		curl_setopt( $ch, CURLOPT_TIMEOUT, $timeout );
		curl_setopt( $ch, CURLOPT_MAXREDIRS, 10 );
		$content = curl_exec($ch);
		$response = curl_getinfo($ch);
		curl_close ($ch);

		if ($response['http_code'] == 301 || $response['http_code'] == 302) {
			ini_set("user_agent", "Mozilla/5.0 (Windows; U; Windows NT 5.1; rv:1.7.3) Gecko/20041001 Firefox/0.10.1");
			$headers = get_headers($response['url']);

			$location = "";
			foreach( $headers as $value ) {
				if ( substr( strtolower($value), 0, 9 ) == "location:" )
					return get_final_url( trim( substr( $value, 9, strlen($value) ) ) );
			}
		}

		if (preg_match("/window\.location\.replace\('(.*)'\)/i", $content, $value) || preg_match("/window\.location\=\"(.*)\"/i", $content, $value)) {
			return get_final_url ( $value[1] );
		} else {
			return $response['url'];
		}
	}

}
