<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author	Marcelo Mottalli <marcelo@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('GFUrl.php');

/**
 * This class displays a paginator
 */
class Paginator {
	const SEPARATOR = '';
	private $name;
	private $items_per_page;
	private $total_items;
	private $page_count;
	private $offset;
	
	/** These two variables are used for naming GET variables */
	private $get_page_name;
	private $get_items_name;
	
	private $current_page;
	private $lang;
	
	private $default_by;
	private $headers;
	
	private $forbiddenVars;

	/**
	 * @param	string $name	Name that will be given to this paginator. Should be a simple string (no spaces and weird stuff) because
	 *							it will be named to pass some variables by GET.
	 * @param	int $items_per_page	Number of items that are displayed per page
	 * @param	int $total_items	Total of items in the list
	 * @param	int $page_count	Number of page numbers to be displayed (0=all)
	 */
	public function __construct($name, $items_per_page, $total_items, $default_by, $columns=array(), $page_count=10) {
		if ($items_per_page == 0) {
			throw new Exception('Tried to build a paginator with 0 items per page');
		}
		
		if ($page_count == 0) {
			throw new Exception('Tried to build a paginator with a limit of 0 pages');
		}

		$this->name = $name;
		$this->default_by = $default_by;
		if ($items_per_page==-1 || $items_per_page==null) {
			$this->items_per_page = $this->getLimit();
		} else {
			$this->items_per_page = $items_per_page;
		}
		$this->total_items = $total_items;
		$this->page_count = $page_count;
		$this->headers = $columns;

		$this->get_page_name = '_'.$this->name.'_page';
		$this->get_items_name = '_'.$this->name.'_total';
		
		if (isset($_GET[$this->get_page_name]) && isset($_GET[$this->get_items_name]) &&
				$_GET[$this->get_items_name] == $this->items_per_page) {
			$this->current_page = Validator::sanitizeString($_GET[$this->get_page_name]);
		} else {
			// first page by default
			$this->current_page = 1;
		}
		
		$this->offset = ($this->current_page-1) * $this->items_per_page;
		$this->lang = Session::getSession()->getLanguage();
		
		$this->forbiddenVars = array('sessionHash', 'change_view', 'submit');
	}
	
	/**
	 * Returns the point where we should be getting the items
	 */
	public function getOffset() {
		return $this->offset;
	}
	
	/**
	 * Returns the paginator, ready for displaying
	 * @return string
	 */
	public function getPaginator($forbidden_vars=array()) {
		if ($this->total_items == 0) {	// there are no items -- don't display anything
			return '';
		}
		
		$url = GFUrl::getUrl();

	//	$res = '<span style="paginator">';
    $res =  "<strong>" . $this->lang->getText("Paginator.Page") . ": </strong>";

		
		// variables to pass
		//$vars = array();
		$vars = $_REQUEST;
		$fvs = array_merge($forbidden_vars, $this->forbiddenVars);
		
		foreach($fvs as $fv) {
			if (isset($vars[$fv])) {
				unset($vars[$fv]);
			}
		}
		
		
		
		$vars[$this->get_items_name] = $this->items_per_page;
		
		$pages = ceil($this->total_items / $this->items_per_page);
		if ($pages == 1) return '';		// no need to create a paginator

		$firstpage = 1;
		$lastpage = $pages;
		$skipped_pages = false;
		
		if ($this->current_page != 1) {
			$vars[$this->get_page_name] = $this->current_page - 1;
			$res .= '<a href="'.$url->currentUrl($vars).'">';
			$res .= '&lt; '.$this->lang->getText('Paginator.Previous');
			$res .= '</a> '.self::SEPARATOR.' ';		
		}

		for ($pagenum = $firstpage; $pagenum <= $lastpage; $pagenum++) {
			$vars[$this->get_page_name] = $pagenum;
			
			if ($pages > 20) {
				if ( ($pagenum > 3 && $pagenum < ($this->current_page-4)) ||
						($pagenum > $this->current_page+4 && $pagenum < ($pages-4)) ) {
					if (!$skipped_pages) {
						$skipped_pages = true;
						$res .= ' ... '.self::SEPARATOR.' ';
					}
					continue;
				} else {
					$skipped_pages = false;
				}
			}

			if ($pagenum == $this->current_page) {
				$res .= '<strong>'.$pagenum.'</strong>';
			} else {
				$res .= '<a href="'.$url->currentUrl($vars).'">';
				$res .= $pagenum;
				$res .= '</a>';		
			}
			if ($pagenum < $pages) $res .= ' '.self::SEPARATOR.' ';
		}
		
		if ($this->current_page != $lastpage) {
			$vars[$this->get_page_name] = $this->current_page + 1;
			$res .= ' '.self::SEPARATOR.' <a href="'.$url->currentUrl($vars).'">';
			$res .= $this->lang->getText('Paginator.Next').' &gt;';
			$res .= '</a>';
		}

    $res .= '<strong> (' . $this->total_items . ' ' .
       $this->lang->getText("Paginator.TotalRecords") . ')</strong>';

	//	$res .= '</span>';
		return $res;
	}
	
	public function addPageClauses(Criteria $c) {
		$c->setLimit($this->items_per_page);
		$c->setOffset($this->offset);
		
		$sort = $this->getSortingOptions();
		$sort_by = $sort['sort_by'];
		$sort_order = $sort['order'];
		
		$sort_order == 'asc' ? $c->addAscendingOrderByColumn($sort_by) : $c->addDescendingOrderByColumn($sort_by);
	}
	
	public function getSQL() {
		$sort = $this->getSortingOptions();
		$sort_by = $sort['sort_by'];
		$sort_order = $sort['order'];
		
		$sql = " ORDER BY $sort_by $sort_order LIMIT {$this->items_per_page} OFFSET {$this->offset} ";
		return $sql;
	}
	
	
	public function getSortingOptions() {
		$default_by = $this->default_by;
		$user = Session::getSession()->getUser();
		//$default_by = $this->pref_default_by;
		$default_order = 'desc';
		
		if (isset($_REQUEST[$this->name . 'sort_by']) || isset($_REQUEST[$this->name . 'sort_order'])) {
			$sort_by = (isset($_REQUEST[$this->name . 'sort_by'])) ? Validator::sanitizeString($_REQUEST[$this->name . 'sort_by']) : $default_by;
			$order = (isset($_REQUEST[$this->name . 'sort_order'])) ? Validator::sanitizeString($_REQUEST[$this->name . 'sort_order']) : $default_order;

			if (!is_null($user)) {
				// Saves the new style as a preference for the user
				$user->setPreference($this->name . '_order', $order);
				$user->setPreference($this->name . '_by', $sort_by);
			}

		} else if (!is_null($user)) {
			$sort_by = $user->getPreference($this->name . '_by');
			$order = $user->getPreference($this->name . '_order');
			
			if (!$sort_by) $sort_by = $default_by;
			if (!$order) $order = $default_order;
		} else {
			$sort_by = $default_by;
			$order = $default_order;
		}
		
		//Validate sort by
		if (is_numeric($sort_by) || ! array_key_exists($sort_by,$this->headers)) {
			$_REQUEST[$this->name . 'sort_by'] = $sort_by = $this->default_by;
		}
		
		$ret = array();				
		$ret['sort_by'] = $sort_by;
		$ret['order'] = ($order == 'asc' ? 'asc':'desc');
		return $ret;
	}
	
	public function getLimit() {
		$user = Session::getSession()->getUser();
		$default_limit = 50;

		if (isset($_REQUEST[$this->name .'_limit'])) {
			$limit = Validator::sanitizeInt($_REQUEST[$this->name . '_limit']);
			if (!is_null($user) && ($user->getPreference($this->name . '_limit') != $limit) ) {
				// Saves the new style as a preference for the user
				$user->setPreference($this->name . '_limit', $limit);
			}
			return $limit;
		} elseif (!is_null($user)) {
			$limit = $user->getPreference($this->name . '_limit');
			if (!$limit) {
				return $default_limit;
			} else {
				return $limit;
			}
		} else {
			return $default_limit;
		}
	}
	
	public function getSortableTableHeaders($theme, $dis_vars=array(), $add_limit_selector=true) {
		$columns = $this->headers;
		$disabled_vars = array_merge($this->forbiddenVars, array($this->name . '_limit', $this->name . '_total', $this->name . '_page'), $dis_vars);
		$sort = $this->getSortingOptions();
		$sort_by = $sort['sort_by'];
		$sort_order = $sort['order'];
		$new_order = ($sort_order == 'asc') ? 'desc' : 'asc';
		$img_name = ($sort_order == 'asc') ? 'arrow_up.gif' : 'arrow_down.gif';
		$url = GFUrl::getUrl();
		
		$theader = array();
		$counter = 0;
		
		foreach ($columns as $name => $show) {
			
			if (is_integer($name)) {
				$theader[] = $show;				
			} else {
				
				$mapname = str_replace("'",'_',str_replace('"','_',str_replace(',', '_', str_replace('.','_',$this->name . $name))));
				$img_up = $theme->getImgRoot() . '/arrow_up.gif';
				$img_down = $theme->getImgRoot() . '/arrow_down.gif';									
				if ($sort_by == $name) {
					$link = $url->currentUrl(array($this->name . 'sort_by' => $name,$this->name . 'sort_order' => $new_order), $disabled_vars);
					//$theader[] = $show.'&nbsp;<a href="'.$link.'">'.$theme->image($img, 0, 0, array('border' => 0)).'</a>';
					$img_normal = $theme->getImgRoot() .'/' . $img_name;
					if ($sort_order == 'asc') {
						$map = "<map id='{$mapname}_map' name='{$mapname}_map'><area shape='rect' coords='0,6,7,11' onMouseOver='document[\"{$mapname}_img\"].src=\"$img_down\"' onMouseOut='document[\"{$mapname}_img\"].src=\"$img_normal\"' href='$link' /></map> ";
					} else {
						$map = "<map id='{$mapname}_map' name='{$mapname}_map'><area shape='rect' coords='0,1,7,6' onMouseOver='document[\"{$mapname}_img\"].src=\"$img_up\"' onMouseOut='document[\"{$mapname}_img\"].src=\"$img_normal\"' href='$link' /></map>";
					}
					$img = $theme->image($img_name, 0, 0, array('border' => '0', 'usemap'=>'#'.$mapname.'_map', 'name'=>$mapname.'_img'));
										
					
				} else {
					$link1 = $url->currentUrl(array($this->name . 'sort_by' => $name, $this->name . 'sort_order' => 'asc'), $disabled_vars);
					$link2 = $url->currentUrl(array($this->name . 'sort_by' => $name, $this->name . 'sort_order' => 'desc'), $disabled_vars);
					$img_normal = $theme->getImgRoot() .'/arrow.gif';	
					$img = $theme->image('arrow.gif', 0, 0, array('border' => '0', 'usemap'=>'#'.$mapname.'_map', 'name'=>$mapname.'_img'));
					$map = "<map id='{$mapname}_map' name='{$mapname}_map' style='border: none;'><area shape='rect' coords='0,1,7,6' onMouseOver='document[\"{$mapname}_img\"].src=\"$img_up\"' onMouseOut='document[\"{$mapname}_img\"].src=\"$img_normal\"' href='$link1' /><area shape='rect' coords='0,5,7,10' onMouseOver='document[\"{$mapname}_img\"].src=\"$img_down\"' onMouseOut='document[\"{$mapname}_img\"].src=\"$img_normal\"' href='$link2' /></map>";
				}
				
				$theader[] = "$map<table cellspacing='0' cellpadding='0' style='padding:0px;margin:0px;height:14px;max-height:14px;line-height:14px;' ><th style='padding-right:3px;margin:0px;font-size: 1em; border: none;' nowrap='nowrap' >$show</th><th style='padding:0px;margin:0px; border: none;' nowrap='nowrap'><div style='height:14px;padding-top:2px;'>$img</div></th></table>";
			}
			$counter++;
		}
		
		if($add_limit_selector && $counter>0) 
		{		
				$theader[$counter-1] = $this->insertLimitOptions($theader[$counter-1]);
		}
		
		return $theader;
	}
	
	public function showLimitOptions($res_vars=array()) {
		$reserved_vars = array_merge($this->forbiddenVars, $res_vars);
		$default_limit = $this->getLimit();
		$lang = Session::getSession()->getLanguage();
		$output = '<table class="quickJumpMenu" cellpadding="0" cellspacing="0">';
		$vars = array();
		foreach ($_REQUEST as $var => $val) {
			if (in_array($var,$reserved_vars) || $var == $this->name . '_limit') continue;
			$vars[$var] = $val;
		}

		$limits = array(5, 25, 50, 75, 100);				
		foreach ($limits as $limit) {
			$newvars = $vars;
			$newvars[$this->name . '_limit'] = $limit;	
			if ($limit == $default_limit) {
				$selected = '(*) ';
			} else {
				$selected = '';
			}
			$output .= '<tr><td class="quickJumpMenu" style="text-align:right;line-height:100%;padding:1px;white-space:nowrap;"><a class="quickJumpMenu" href="' . GFUrl::getUrl()->currentUrl($newvars) . "\">$selected$limit</a></td></tr>";			
		}
		$output .= '</table>';
		return $output;
	}

	public function insertLimitOptions($original_header) {
		$lang = Session::getSession()->getLanguage();
		$theme = Session::getSession()->getTheme();
		$row_img = $theme->getImgRoot() .'/row_selector.gif';
		$id = str_replace("'",'_',str_replace('"','_',str_replace(',', '_', str_replace('.','_',$this->name))));

		$return = '<table class="tabular2" width="100%" cellpadding="0" cellspacing="0"><tr class="th"><th style="text-align:left;" valign="center" class="tabular2">'
				. $original_header
				. '</th><th valign="center" style="text-align:right;" class="tabular2">'
				. "<div style='height:14px;padding-bottom:1px;'><img id='{$id}Icon' src='$row_img' onclick=\"expandFloatingMenu('{$id}Div','{$id}Icon');\" /></div>"
				. '</th></tr></table>';

		/*$return = /*"<div style='float:left;position:relative;top:0px;'>$original_header</div>"
				"$original_header<img class='quickJumpMenu' id='{$id}Icon' src='$row_img' style='float:right;top:4px;padding-right:0px; position:relative;' onclick=\"expandFloatingMenu('{$id}Div','{$id}Icon');\" onmouseover='this.src=\"$row_img_over\";' onmouseout='this.src=\"$row_img\";'/>";*/
				
				
		$return .= "<div id='{$id}Div' class=\"quickJumpMenu\" style=\"visibility:hidden; position:absolute; z-index:2; max-height: 60%; overflow:auto;\">" . $this->showLimitOptions() . '</div>';
		return $return;
	}
}
?>
