<?php

/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Tim Perdue tim@gforge.org
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
require_once('core/PluginPeer.php');
require_once('plugin/GFPlugin.php');

/**
 * This class builds GFPlugin objects. It is accessed as a Singleton.
 */
class PluginFactory {

	/**
	 * This variable will hold a list of the plugins that have been
	 * initialized.
	 */
	private $initializedPlugins = null;

	/**
	 * This variable holds the global instance of the plugin factory. Only one instance
	 * is needed, we don't want to create the instances of the plugins more than once.
	 */
	private static $instance = null;

	private function __construct() {

	}

	/**
	 * Return the only (global) instance of the factory.
	 * @return PluginFactory
	 */
	public static function getFactory() {
		if (is_null(self::$instance)) {
			self::$instance = new PluginFactory();
		}

		return self::$instance;
	}

	/**
	 * @return	array	Associative array with all the available plugins
	 */
	public function getPlugins() {
		if (!is_array($this->initializedPlugins)) {
			$this->initializePlugins();
		}
		return $this->initializedPlugins;
	}

	/**
	 * Retrieve a GFPlugin by its name
	 * @param	string $name	Name of the plugin
	 * @return	GFPlugin
	 */
	public function getPluginByName($name) {
		$plugins = $this->getPlugins();
		if (!array_key_exists($name, $plugins)) {
			return null;
		}
		return $plugins[$name];
	}

	/**
	 * This methods makes sure that all the plugins are initialized properly.
	 * @param bool $failOnMissing If TRUE, then throw an exception if a given plugin record has no corresponding implementation class
	 * @return void
	 * @throws Exception If a plugin class is not found, and the fail parameter is TRUE (default), then an exception is thrown.
	 */
	public function initializePlugins($failOnMissing = true) {
		if (!is_array($this->initializedPlugins)) {
			$this->initializedPlugins = array();

			$c = new Criteria();
			//		$c->add(PluginPeer::PLUGIN_TYPE, GFPlugin::SITE_PLUGIN);
			$c->addAscendingOrderByColumn(PluginPeer::PLUGIN_ORDER);
			$plugins = PluginPeer::doSelect($c);

			$pluginsPath = GFConf::get('system.pluginsPath');

			foreach ($plugins as $plugin) {
				$name = $plugin->getPluginName();
				$classname = ucfirst($name) . 'Plugin';

				//include and instantiate each plugin
				$plugin_php = SystemUtils::buildFilePath($pluginsPath, '/' . $name . '/lib/' . $classname . '.php');
				if (!is_file($plugin_php)) {
					$msg = "Couldn't locate file $plugin_php, implementation class $classname for plugin $name";
					if ($failOnMissing) {
						throw new Exception($msg);
					} else {
						error_log("PluginFactory->initializePlugins $msg");
						continue;
					}
				}

				// Plugin file was found... instantiate it
				require_once($plugin_php);
				$this->initializedPlugins[$name] = new $classname();

				//  Initialize each plugin
				$this->initializedPlugins[$name]->initialize();
			}
		}

		return;
	}

}

?>
