<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Tim Perdue tim@gforge.org
 * @date 2006-02-27
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
require_once('initialize.php');
require_once('GFConf.php');
require_once('external/propel/Propel.php');
require_once('external/creole/Creole.php');
require_once('external/creole/common/ResultSetCommon.php');
require_once('core/Language.php');
require_once('core/User.php');
require_once('core/Theme.php');
require_once('core/User.php');
require_once('core/UserSession.php');
require_once('core/SiteAdmin.php');
require_once('core/SecureSession.php');
require_once('exceptions/SessionException.php');
require_once('html/ThemeFactory.php');
require_once('Validator.php');
require_once('external/Browser.php');

/**
 * Represents an user session
 */
class Session {

	const DEFAULT_LANGUAGE = 'en';

	private static $instance = null;
	private $isLoggedIn = false;
	/** @var User $user */
	private $user = null;
	private $language = null;
	private $theme = null;
	private $isSiteAdmin = false;
	private $conn_host = '';
	private $conn_port = '';
	private $conn_db = '';
	private $conn_username = '';
	private $conn_password = '';


	/**
	 * Holds now long a session is active (in seconds)
	 * @var integer
	 */
	private $ttlInSeconds = null;

	/**
	 * @var boolean
	 */
	private $expireOnLoginTime = null;

    /**
     * On catastrofic failure we still need a valid TTL, otherwise all sessions
     * will be deleted in gc()
     */
    const DEFAULT_TTL_SEC = 86400;

	/**
	 * Private constructor -- do not construct this object directly.
	 *
	 * @see Session::getSession()
	 */
	private function __construct($connection_data) {
		//Gets DB connection data
		$this->connection_data = include($connection_data);
        if (isset($this->connection_data['propel']['datasources']['gf5']['connection'])) {
            $conn_array = $this->connection_data['propel']['datasources']['gf5']['connection'];
            $this->conn_host = $this->getArrayElementValue($conn_array, 'hostspec');
            $this->conn_port = $this->getArrayElementValue($conn_array, 'port');
            $this->conn_db = $this->getArrayElementValue($conn_array, 'database');
            $this->conn_username = $this->getArrayElementValue($conn_array, 'username');
            $this->conn_password = $this->getArrayElementValue($conn_array, 'password');
        }
		try {
			//Get TTL
			$this->ttlInSeconds = GFConf::get('system.sessionTimeout');
			$this->expireOnLoginTime = Validator::sanitizeBool(GFConf::get('system.expireOnLoginTime'));
		} catch (Exception $ex) {// Exception will be managed in gf.php
			error_log("Exception in Session ctor: " . $ex->getMessage());
			$this->ttlInSeconds = self::DEFAULT_TTL_SEC;
			$this->expireOnLoginTime = false;
		}
		//Sets custom session management for PHP
		session_set_save_handler(
			array($this, 'open'), // open
			array($this, 'close'), // close
			array($this, 'read'),
			array($this, 'write'),
			array($this, 'destroy'),
			array($this, 'gc')
		);

		//Starts or resumes session
		$this->start_session();
	}

    private function getArrayElementValue($array, $index, $default = '') {
        return isset($array[$index]) ? $array[$index] : $default;
    }


	/**
	 * Returns the connection string for raw postgres access (no Propel)
	 *
	 * @return String DB connection string
	 */
	public function getConnectionString() {
		//Format: "host=$host dbname=$db user=$user password=$pass"
		$arr_con = array();
		if ($this->conn_host != '') {
			$arr_con[] = 'host=' . $this->conn_host;
		}
		if ($this->conn_db != '') {
			$arr_con[] = 'dbname=' . $this->conn_db;
		}
		if ($this->conn_username != '') {
			$arr_con[] = 'user=' . $this->conn_username;
		}
		if ($this->conn_password != '') {
			$arr_con[] = 'password=' . $this->conn_password;
		}
		if ($this->conn_port != '') {
			$arr_con[] = 'port=' . $this->conn_port;
		}

		return implode(" ", $arr_con);
	}

	/**
	 * Method that returns the singleton session object
	 * @param string $config_file The string path and file name of the config file to be used for DB connection data.
	 * @return Session
	 */
	public static function getSession($config_file = '/etc/gforge/gforge5-db-conf.php') {
		if (is_null(self::$instance)) {
			self::$instance = new Session($config_file);
		}
		register_shutdown_function('session_write_close');
		return self::$instance;
	}



	/**
	 * Starts a PHP session, or resumes an existing one
	 *
	 */
	function start_session($logout=false) {
		//Set session cookie not accessible from javascript
		$httponly = true;

		$secureCookies = Validator::sanitizeBool(GFConf::get("system.secureCookies"));

		//Sets hash algorithm to use for the sessionid
		$session_hash = 'sha512';

		//Checks if hash is available
		if (in_array($session_hash, hash_algos())) {
			//Sets the has function.
			ini_set('session.hash_function', $session_hash);
		}
		//How many bits per character of the hash.
		//The possible values are '4' (0-9, a-f), '5' (0-9, a-v), and '6' (0-9, a-z, A-Z, "-", ",").
		ini_set('session.hash_bits_per_character', 5);

		//Forces the session to only use cookies, not URL variables.
		ini_set('session.use_only_cookies', 1);

		//Gets session cookie parameters
		$cookieParams = session_get_cookie_params();

		//Sets the parameters
		//error_log("Session > start_settion, cook path: " . $cookieParams["path"]);
		//error_log("Session > start_settion, cook domain: " . $cookieParams["domain"]);
		session_set_cookie_params($this->ttlInSeconds, $cookieParams["path"], $cookieParams["domain"], $secureCookies, $httponly);

		//Start or resume the session
		if (session_id() == "") {
			session_start();
		}
		$current_session_hash = session_id();
		if ($logout) {
			$this->destroy($current_session_hash);
			$this->deleteUserSession($current_session_hash);
			session_regenerate_id();
		}

		//error_log("Session > start_session > existing session_hash: " . session_id());
	}


	/**
	 * Verifies if there's a user_session object associated to the current session id
	 *
	 */
	public function checkUserSession() {
		try {
			$con = Propel::getConnection(UserPeer::DATABASE_NAME);
			$stmt = $con->createStatement();
			$sql = "SELECT * FROM user_session WHERE session_hash = '".session_id()."'";
			//error_log("Session > checkUserSession, sessid: " . session_id());
			$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_ASSOC);
			if ($rs->next() ) {
				//error_log("Session > checkUserSession, exists");
				return true;
			} else {
				//error_log("Session > checkUserSession, does not exist");
				return false;
			}
		} catch(Exception $e) {
			error_log("Session > checkUserSession, error: " . $e->getMessage());
			return false;
		}
	}


	/**
	 * Retrieves user id associated to the current session id
	 * @return int
	 * @throws Exception
	 */
	public function getSessionUser() {
		try {
			$con = Propel::getConnection(UserPeer::DATABASE_NAME);
			$stmt = $con->createStatement();
			$sql = "SELECT * FROM user_session WHERE session_hash = '".session_id()."'";
			$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_ASSOC);
			if ($rs->next()) {
				return $rs->getInt("user_id");
			} else {
				return User::UNDEFINED_USER_ID;
			}
		} catch(Exception $e) {
			error_log("Session > checkUserSession, error: " . $e->getMessage());
			return 0;
		}
	}


	/**
	 * Sets up (continues) an old session given by $sessionHash
	 * This method will set timezone, theme and language if the user logs in.
	 * If this is an anonymous session, it's the caller task to set language, theme and zone.
	 * @todo We may need to revisit this in the future, seems like centralizing everything here
	 *
	 * @param	string $ip			IP address of the remote host
	 * @param	string $session_id	Session hash string
	 * @throws Exception
	 * @return bool
	 *
	 */
	public function setup($ip = "127.0.0.1", $session_id='') {
		//error_log("Session > setup, ip: " . $ip);
		if ($session_id != '') {
			//Set the received session_id
			session_id($session_id);
		}

		if (ActionManager::notify('CHECK_INTEGRITY', array()) != utf8_encode('6.0')) {
			throw new Exception('Invalid version number detected [1]. Please contact GForge Group.');
		}

		$user_id = $this->getSessionUser();
		$user_id = Validator::sanitizeInt($user_id);  // Prevent malicious strings being injected in the server

		/** @var User $user */
		$user = UserPeer::retrieveByPk($user_id);
		$sessionHash = session_id();
		if (is_null($user) || $user->getUserId() == User::UNDEFINED_USER_ID  || $user->getStatus() != User::STATUS_ACTIVE) {
			//throw new SessionException("Invalid user ID for the session");
			//Start session as 'none' user
			try {
				$c = new Criteria();
				$c->add(UserSessionPeer::SESSION_HASH, $sessionHash);
				$us = UserSessionPeer::doSelectOne($c);

				if (!$us) {
					$us = new UserSession();
				}

				$us->setUserId(User::UNDEFINED_USER_ID);
				$us->setIpAddr($ip);
				$us->setSessionHash($sessionHash);
				$us->setLoginTime(null);
				$us->save();

			} catch(Exception $e) {
				error_log("Session > write, error: " . $e->getMessage());
			}
			$this->isLoggedIn = false;
			return $this->isLoggedIn;
		}

		putenv('TZ='. $user->getTimezone());
		$this->language = $user->getLanguage();
		$this->user = $user;
		$this->isLoggedIn = true;

		$c = new Criteria();
		$c->add(SiteAdminPeer::USER_ID, $user_id);
		$isSiteAdmin = SiteAdminPeer::doSelect($c);
		if ($isSiteAdmin) {
			$this->isSiteAdmin = true;
		} else {
			$this->isSiteAdmin = false;
		}

		$theme = $user->getTheme();

		if (is_object($theme)) {
			$theme = ThemeFactory::getTheme($theme->getDirName());
		} else {
			$default = GFConf::get('system.defaultTheme');
			$theme = ThemeFactory::getTheme($default);
		}

		$this->theme = $theme;

		return $this->isLoggedIn;

	}

	/**
	 * @return bool
	 */
	public function isLoggedIn() {
		return $this->isLoggedIn;
	}

	/**
	 * @return Language
	 * @throws SessionException
	 */
	public function getLanguage() {
		if (is_null($this->language)) {
			if ($this->isLoggedIn && $this->user) {
				$this->language = $this->user->getLanguage();
			} else {
				// Get the default language if the user isn't logged in
				$defaultLangCode = GFConf::get("system.defaultLanguage", self::DEFAULT_LANGUAGE);

				// retrieve the default language object from the DB
				$c = new Criteria();
				$c->add(LanguagePeer::LANGUAGE_CODE, $defaultLangCode);
				$this->language = LanguagePeer::doSelectOne($c);

				if (is_null($this->language)) {
		            throw new SessionException("Couldn't retrieve language object: " . $defaultLangCode);
		        }
			}
		}
		return $this->language;
	}

	public function setLanguage($langobj) {
		$this->language = $langobj;
	}

	/**
	 * @return bool
	 */
	public function isSiteAdmin() {
		return $this->isSiteAdmin;
	}

	/**
	 * @return User
	 */
	public function getUser() {
		return $this->user;
	}

	/**
	 * @return Theme
	 */
	public function getTheme() {
		if (is_null($this->theme)) {
			if (!is_null($this->user)) {
				//should be done already in session::setup
				/* 	$theme = $this->user->getTheme()->getDirname();
				$this->theme = ThemeFactory::getTheme($theme);
				*/
			} else {
			// User is not logged in, get the default theme
				$defaultThemeName = GFConf::get('system.defaultTheme');

				$this->theme = ThemeFactory::getTheme($defaultThemeName);
			}
		}

        if ($this->theme == false) {
			die("<h3>No valid theme found.</h3>");
		}

		return $this->theme;
	}

	/**
	 * @return void
	 */
	public function terminate() {
		$this->isLoggedIn = false;
		$this->user = null;
		$this->language = null;
		//We start a new session, detached from the user
		$this->start_session(true);
	}

	/**
	 * @return string
	 */
	public function getSessionHash() {
		return session_id();
	}

	/**
	 * Attaches this (anonymous) session to a user session.
	 * This is usually done right after login:
	 * <code>
	 * 	$login = LoginFactory::getLogin();
	 * 	$session = Session::getSession();
	 * 	$ip = $_SERVER["REMOTE_ADDR"];
	 * 	$userSession = $login->doLogin("username", "password");
	 * 	$session->setUserSession($userSession, $ip);
	 * 	$hash = $session->getSessionHash();
	 * 	// Store $hash somewhere in the client (usually a cookie)
	 * </code>
	 * @param	UserSession $userSession
	 * @param string $ip
	 */
	public function setUserSession(UserSession $userSession, $ip = "127.0.0.1") {
		if ($this->isLoggedIn) {
			return;
			//throw new SessionException("Tried to assign a user to an already existing session");
		}
		ActionManager::notify('SESSION_SET_USER_SESSION', array('userSession'=>$userSession));

		$this->user = $userSession->getUser();
		$this->language = $this->user->getLanguage();
		$this->user->setLastLogin(time());

		// update the session with the new info
		$userSession->setIpAddr($ip);
		$userSession->setLoginTime(strtotime(date(DATE_RFC2822)));
		$userSession->setSessionHash(session_id());
		$userSession->save();

		// We should do this at the last step, otherwise a method call above could throw an exception
		// and it'd think we're logged in
		$this->isLoggedIn = true;
	}


	public function deleteUserSession($session_hash='') {
		$c = new Criteria();
		$c->add(UserSessionPeer::SESSION_HASH, $session_hash);
		$user_session = UserSessionPeer::doSelectOne($c);
		if (is_object($user_session)) {
			$user_session->delete();
		}
	}


	/**
	 * Attaches this session to a User.
	 * This is done when the "session" is actually a small request
	 * that needs to be attached to a user. For example, we use this when a user
	 * commits a file to SCM. It doesn't store any information in the database, nor
	 * generates a session hash.
	 * @param User $user
	 * @param bool $force
	 * @throws SessionException
	 */
	public function setUser(User $user, $force=false) {
		if ($this->isLoggedIn && !$force) {
			throw new SessionException("Tried to assign a user to an already existing session");
		}

		$this->isSiteAdmin = $user->isSiteAdmin();
		$this->isLoggedIn = true;
		$this->user = $user;
		$this->language = $this->user->getLanguage();
	}



	/**
	 * Informs we must run this session as a cronjob. A cronjob has admin privileges,
	 * but no user associated.
	 */
	public function beginCronjobSession() {
	// Use the default language as the session language
		$this->language = LanguagePeer::retrieveByPK(Language::BASE_LANGUAGE_ID);
		$this->isLoggedIn = true;
		$this->isSiteAdmin = true;
		$this->user = null;
	}

	/**
	 * Disable gz buffering. Call this method BEFORE doing any output to the browser
	 */
	public function disableGZHandler() {
		$handlers = ob_list_handlers();
		if (count($handlers) > 0 && strtolower($handlers[0] == 'ob_gzhandler')) {
			ob_end_clean();
		}
	}


	/**
	 * We don't do anything special here.
	 * @param  string  $savePath    Not used
	 * @param  string  $sessionName Not used
	 * @return boolean Always true
	 */
	public function open($savePath, $sessionName) {
		return true;
	}


	/**
	 * We don't do anything special here.
	 *
	 * @return boolean Always true
	 */
	public function close()
	{
		return true;
	}

	private function getDBConnection() {
		$con = pg_connect($this->getConnectionString())
			or die ("Could not connect to server\n");
		return $con;
	}

	/**
	 * @param resource $con pgsql connection
	 */
	private function closeDBConnection($con) {
		try {
			pg_close($con);
		} catch(Exception $e) {}
	}


	/**
	 * Executes a raw postgres sql that returns a single value
	 *
	 * @param  string $sql to be executed
	 * @param  array $params for the sql query
	 * @return string single value returned from the query, or false if no result
	 *
	 */
	private function execSingleValue($sql, $params) {
		try {
			$con = $this->getDBConnection();

			$rs = pg_query_params($con, $sql, $params) or die("Cannot execute query: $sql\n");
			if ($row = pg_fetch_row($rs)) {
				$this->closeDBConnection($con);
				return $row[0];
			} else {
				$this->closeDBConnection($con);
				return false;
			}

		} catch(Exception $e) {
			error_log("Session > read, error: " . $e->getMessage());
			return '';
		}
	}


	/**
	 * Executes a raw postgres sql that doesn't return a value
	 *
	 * @param  string $sql to be executed
	 * @param  array $params for the sql query
	 * @return bool true: if success, false: if failure
	 *
	 */
	private function execUpdate($sql, $params) {
		try {
			$con = $this->getDBConnection();

			pg_query($con, "BEGIN WORK");

			$res = pg_query_params($con, $sql, $params) or die("Cannot execute update: $sql\n");

			if (!$res) {
				pg_query($con, "ROLLBACK");
				$this->closeDBConnection($con);
				return false;
			} else {
				pg_query($con, "COMMIT");
				$this->closeDBConnection($con);
				return true;
			}
		} catch(Exception $e) {
			error_log("Session > read, error: " . $e->getMessage());
			return false;
		}
	}


	/**
	 * Sanitizes the session hash
	 *
	 * @param  string $id Session ID (as assigned by PHP)
	 * @return string sanitized Session ID
	 *
	 */
	private function sanitizeSessionId($id) {
		return htmlspecialchars($id, ENT_COMPAT);
	}


	/**
	 * Gets the session data tied to the given session ID
	 *
	 * @param  string $id Session ID (as assigned by PHP)
	 * @return string Serialized data (PHP will deserialize it)
	 *
	 */
	public function read($id) {
		//error_log('Session > read');
		$id = $this->sanitizeSessionId($id); //sanitize id


		if (GFConf::get('system.expireOnLoginTime') == true) {
			$ttl_param = " login_time > now() - INTERVAL '".$this->ttlInSeconds." seconds'";
		} else {
			$ttl_param = " modified > now() - INTERVAL '".$this->ttlInSeconds ." seconds'";
		}
		$sql = "SELECT data FROM user_session WHERE session_hash = $1 AND $ttl_param ";
		try {
			$res = $this->execSingleValue($sql, array($id));
			if ($res === false) {// session expired, terminate
				$this->terminate();
				return '';
			} else {
				//Delete cookie first, if exists
				setcookie('PHPSESSID', session_id(), time() - 1, '/');

				$sql = "UPDATE user_session SET modified=now() WHERE session_hash=$1";
				setcookie('PHPSESSID', session_id(), time() + $this->ttlInSeconds, '/');


				$this->execUpdate($sql, array($id));

				return $res;
			}
		} catch(Exception $e) {
			error_log("Session > read, error: " . $e->getMessage());
			$res = array();
			return serialize($res);
		}
	}

	/**
	 * Writes session data
	 *
	 * @param  string  $id   PHP Session ID
	 * @param  string  $data =Data to persist
	 * @return boolean True if write worked, otherwise false
	 *
	 */
	public function write($id, $data) {
		//error_log('Session > write, data: ' . $data);
		$id = $this->sanitizeSessionId($id); //sanitize id

		try {
			$sql = "UPDATE user_session SET data=$2, modified=now() WHERE session_hash=$1";
			$res = $this->execUpdate($sql, array($id, $data));
			return $res;
		} catch(Exception $e) {
			error_log("Session > write, error: " . $e->getMessage());
			return false;
		}
	}

	/**
	 * Destroys the current user session
	 *
	 * @param  string  $id
	 * @return boolean True if session was destroyed otherwise false
	 *
	 */
	public function destroy($id) {
		//error_log('Session > destroy');
		$id = $this->sanitizeSessionId($id); //sanitize id

		try {
			$sql = "DELETE FROM user_session WHERE session_hash = $1";
			$res = $this->execUpdate($sql, array($id));
			return $res;
		} catch(Exception $e) {
			error_log("Session > write, error: " . $e->getMessage());
			return false;
		}
	}

	/**
	 * Garbage collects (expires) old sessions
	 *
	 * NOTE: we don't use the php.ini maxlifetime setting
	 *
	 * @param  integer $max NOTE: not used
	 * @return boolean True if GC works, otherwise false
	 *
	 */
	public function gc($max) {
		try {
			syslog(LOG_ERR, 'running gc');
			$ttl_param = " INTERVAL '".$this->ttlInSeconds ." seconds'";
			$sql = "DELETE FROM user_session WHERE modified < now() - $ttl_param";
			$res = $this->execUpdate($sql, array());
			return $res;
		} catch(Exception $e) {
			error_log("Session > write, error: " . $e->getMessage());
			return false;
		}
	}

}
