<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Marcelo Mottalli	<marcelo@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
 
require_once('system/CronManager.php');
require_once('plugin/ActionManager.php');
require_once('StringUtils.php');
require_once('core/PendingAction.php');
 
/**
 * GForge daemon. 
 * The daemon is started by the "gforge" script, and is accessed by the individual static methods.
 */
class GFDaemon {
	/** PID file */
	const PID_FILE = '/var/run/gforge.pid';
	const MAX_MESSAGE_SIZE = 4096;
	
	protected static $queue = null;
	
	/** Do not instantiate directly */
	private function __construct() {
	}

	public static function start() {
		if (!function_exists("posix_setgid")) {
			throw new Exception("POSIX functions not present in PHP");
		}
		if (!function_exists("pcntl_fork")) {
			throw new Exception("pcntl_fork function not present in PHP. Please recompile with --enable-pcntl.");
		}
		if (self::isRunning()) {
			return;
		}

		$userData = posix_getpwuid(posix_getuid());
		if ($userData["name"] != "root") {
			throw new Exception("Must be root for running the daemon");
		}
		
		if (!self::fork()) {
			throw new Exception("Couldn't fork process");
		}
		
	}
	
	public static function stop() {
		if (!self::isRunning()) {
			return;
		}
		
		$pid = self::getDaemonPid();
		if (!posix_kill($pid, SIGTERM)) {
			throw new Exception("Could not kill PID $pid");
		}
		
		// Give it a second to cleanup
		sleep(1);
	}
	
	public static function isRunning() {
		if (file_exists(self::PID_FILE) && ($pid = @file_get_contents(self::PID_FILE))) {
			// this will tell if process is running without actually killing it
			return file_exists("/proc/$pid");
		} else {
			return false;
		}
	}
	
	/**
	 * Returns the PID of the daemon
	 */
	public static function getDaemonPid() {
		if (!file_exists(self::PID_FILE)) {
			return 0;
		}
		
		$pid = file_get_contents(self::PID_FILE);
		return intval($pid);
	}
	
	/**
	 * Sends a message to the running daemon.
	 * NOTE: this function runs from the "user" side, *not* from the daemon side,
	 * so self::$queue will be null.
	 */
	public static function notify($action, $params) {
		if (!self::isRunning()) {
			//Insert a register in the database
			$pa = new PendingAction();
			$pa->setActionName($action);
			$pa->setParameter($params);
			$pa->save();
			return false;
		}

		
		// Pushes the message in the system IPC queue
		$queue = msg_get_queue(self::getQueueKey());
		$message = array('a' => $action, 'p' => $params);		// Try to keep the message small
		msg_send($queue, 1, $message);
		
		return true;
	}
	
	/**
	 * Begins the daemon process
	 */
	private static function fork() {
		$pid = pcntl_fork();
		if ($pid == -1) {
			return false;
		} else if ($pid != 0) {
			// We're on parent process
			return true;
		}
		
		// Here we are on the forked process
		self::handleDaemon();
	}
	
	public static function handleSignal($signal) {
		// Clean up
		unlink(self::PID_FILE);
		
		if (self::$queue) {
			msg_remove_queue(self::$queue);
		}
		
		exit(0);
	}
	
	/**
	 * Handler for the daemon process
	 */
	private static function handleDaemon() {
		// Install the signal handler
		declare(ticks = 1);		// Somehow this is needed... else it won't work
		$signals = array(SIGTERM, SIGHUP, SIGQUIT);
		foreach ($signals as $signal) {
			if (!pcntl_signal($signal, array("self", "handleSignal"))) {
				echo "Could not install signal handler. Aborting.\n";
				exit(1);
			}
		}

		// get the daemon's PID and write it to the PID file
		$pid = posix_getpid();
		$fh = fopen(self::PID_FILE, "w");
		if (!$fh) {
			echo "Could not open ".self::PID_FILE." for writing. Aborting.\n";
			exit(1);
		}

		fwrite($fh, $pid);
		fclose($fh);

		//Process pending actions before entering the main loop
		$pendingActions = self::getPendingActions();
		foreach ($pendingActions as $action) {
			ActionManager::notify($action->getActionName(), $action->getParameter());
		}
		self::clearPendingActions();
		self::daemonLoop();
	}
	
	static function clearPendingActions() {
		$con = Propel::getConnection();
		$sql = "DELETE FROM pending_action";
		$stmt = $con->createStatement();
		$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_NUM);
	}

	static function getPendingActions() {
		$con = Propel::getConnection();
		$sql = "SELECT DISTINCT ON(action_name,parameter) * FROM pending_action";
		$stmt = $con->createStatement();
		$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_NUM);

		return PendingActionPeer::populateObjects($rs);
	}
	
	private static function getQueueKey() {
		return ftok(self::PID_FILE, 'g');
	}
	
	/**
	 * Main daemon loop
	 */
	private static function daemonLoop() {
		// Create the IPC queue
		self::$queue = msg_get_queue(self::getQueueKey(), 0666);
		
		// Open the system logger
		openlog('GForge daemon', LOG_CONS | LOG_PID, LOG_DAEMON);
		syslog(LOG_DEBUG, 'Daemon initialized, running main loop...');
		
		while (true) {
			msg_receive(self::$queue, 0, $msgtype, self::MAX_MESSAGE_SIZE, $message);
			$action = $message['a'];
			$params = $message['p'];
			self::logQueueMessage($action, $params);
			
			try {
				// Dispatch the message to the GForge system using the ActionManager
				ActionManager::notify($action, $params);
				syslog(LOG_DEBUG, 'Finished processing message');
			} catch (Exception $ex) {
				syslog(LOG_ERR, 'Error: '.$ex->getMessage());
			}
		}
	}
	
	private static function logQueueMessage($action, $params) {
		// Log to the system log about the message
		$paramsInfo = var_export($params, true);
		$paramsInfo = StringUtils::trimString($paramsInfo, 50);
		$queueInfo = msg_stat_queue(self::$queue);
		
		syslog(LOG_DEBUG, "Received message $action (sender PID: {$queueInfo['msg_lspid']})");
		syslog(LOG_DEBUG, "Parameters: $paramsInfo");
	}
	
	public static function logDebugMessage($message) {
		syslog(LOG_DEBUG, $message);
	}
}
?>
