<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id: cleanup.php 489 2006-02-26 20:04:55Z tperdue $
 * @author	Marcelo Mottalli <mottalli@gforgegroup.com>
 * @date 2006-02-27
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
/**
 * This class has several functions for managing /etc/[passwd|group|shadow] files
 */
class SystemConfigManager {
	const GFORGE_MARKER_BEGIN = '### GFORGEBEGIN';
	const GFORGE_MARKER_END = '### GFORGEEND';
	
	private $global_users = array();
	private $gforge_users = array();
	private $users = array();
	
	private $global_groups = array();
	private $gforge_groups = array();
	private $groups = array();
	
	private $testing;
	private $changed = true;
	
	public function __construct($testing = false) {
		$this->testing = $testing;
		$this->changed = false;
		$this->parseGroups();
		$this->parsePasswd();
	}
	
	public function purgeGForgeGroups() {
		foreach ($this->gforge_groups as $groupname => $foo) {
			unset($this->groups[$groupname]);
		}
		$this->gforge_groups = array();
	}
	
	public function purgeGForgeUsers() {
		foreach ($this->gforge_users as $username => $foo) {
			unset($this->users[$username]);
		}
		$this->gforge_users = array();
	}
	
	public function createGForgeGroup($groupname, $gid) {
		if ($this->groupExists($groupname)) return true;
		$group = array('name' => $groupname, 'password' => 'x', "gid" => $gid, "users" => array());
		$this->gforge_groups[$groupname] = $group;
		$this->groups[$groupname] = $group;
		$this->changed = true;
	}
	
	public function deleteGForgeGroup($groupname) {
		if (!array_key_exists($groupname, $this->gforge_groups)) {
			return false;
		}
		
		unset($this->gforge_groups[$groupname]);
		unset($this->groups[$groupname]);
		$this->changed = true;
		
		return true;
	}
	
	/**
	 * Creates/modifies information of a user. If the user already exists, it overwrites the
	 * information. The user is added inside the GForge-specific markers in /etc/passwd
	 */
	public function createGForgeUser($username, $uid, $unix_password, $gid, $home, $shell, $gcos) {
		// Note: if the user already exists, the user's information is overwritten
		if ($username == 'root') return false;		// small security check
		
		$user = array('name' => $username, 'unix_password' => $unix_password, 'uid' => $uid,
			'gid' => $gid, 'gcos' => $gcos, 'home' => $home, 'shell' => $shell);
		
		if (array_key_exists($username, $this->gforge_users)) {
			unset($this->gforge_users[$username]);
		} elseif (array_key_exists($username, $this->users)) {
			unset($this->global_users[$username]);
		}
		
		$this->gforge_users[$username] = $user;
		$this->users[$username] = $user;
		
		$this->changed = true;
		return true;
	}
	
	public function userExists($username) {
		return array_key_exists($username, $this->users);
	}
	
	public function groupExists($groupname) {
		return array_key_exists($groupname, $this->groups);
	}
	
	public function addUserToGroup($username, $groupname) {
		if (!$this->groupExists($groupname)) return false;
		elseif ($this->isUserInGroup($username, $groupname)) return true;	// user already in group
		
		if (array_key_exists($groupname, $this->gforge_groups)) {
			$this->gforge_groups[$groupname]['users'][] = $username;
		} elseif (array_key_exists($groupname, $this->global_groups)) {
			$this->global_groups[$groupname]['users'][] = $username;
		}
		
		$this->groups[$groupname]['users'][] = $username;
		$this->changed = true;

		return true;
	}
	
	public function deleteGForgeUser($username) {
		if (!array_key_exists($username, $this->gforge_users)) {
			return false;
		}
		
		foreach ($this->gforge_groups as $groupname => $foo) {
			$this->deleteUserFromGroup($username, $groupname);
		}
		
		unset($this->gforge_users[$username]);
		unset($this->users[$username]);
		$this->changed = true;
		
		return true;
	}
	
	public function deleteUserFromGroup($username, $groupname) {
		if (!$this->isUserInGroup($username, $groupname)) {
			return false;
		}
		
		unset($this->groups[$groupname]['users'][$username]);
		if (array_key_exists($username, $this->gforge_groups[$groupname]['users'])) {
			unset($this->gforge_groups[$groupname]['users'][$username]);
		}
		
		$this->changed = true;
		return true;
	}
	
	public function isUserInGroup($username, $groupname) {
		if (!$this->groupExists($groupname)) return false;
		return (array_search($username, $this->groups[$groupname]['users']) !== false);
	}
	
	/**
	 * Writes all the changes to the disk. This should be called when you finish making ALL your
	 * changes, and not after individual changes.
	 */
	public function writeChanges() {
		if (!$this->changed) return true;
		
		// First, write /etc/group
		if ($this->testing) $file = '/etc/group.gforge-debug';
		else $file = '/etc/group';
		
		$fh = fopen($file, 'w');
		if (!$fh) {
			throw new Exception('Couldn\'t open '.$file.' for writing');
		} else {
			$this->writeGroups($fh, $this->global_groups);
			
			fputs($fh, self::GFORGE_MARKER_BEGIN."\n");
			$this->writeGroups($fh, $this->gforge_groups);
			fputs($fh, self::GFORGE_MARKER_END."\n");
		}
		fclose ($fh);
		
		// write /etc/passwd
		if ($this->testing) $file = '/etc/passwd.gforge-debug';
		else $file = '/etc/passwd';
		$fh = fopen($file, 'w');
		if (!$fh) {
			throw new Exception('Couldn\'t open '.$file.' for writing');
		} else {
			$this->writePasswd($fh, $this->global_users);
			
			fputs($fh, self::GFORGE_MARKER_BEGIN."\n");
			$this->writePasswd($fh, $this->gforge_users);
			fputs($fh, self::GFORGE_MARKER_END."\n");
		}
		fclose ($fh);
		
		$this->changed = false;
	}
	
	private function writeGroups($fh, $groups) {
		foreach ($groups as $groupname => $group) {
			$line = $group['name'].':'.$group['password'].':'.$group['gid'].':'.implode(',', $group['users']);
			fputs($fh, $line."\n");
		}
	}
	
	private function writePasswd($fh, $users) {
		foreach ($users as $useranme => $user) {
			$line = $user['name'].':'.$user['unix_password'].':'.$user['uid'].':'.$user['gid'].':'.$user['gcos'].':'.
				$user['home'].':'.$user['shell'];
			fputs($fh, $line."\n");
		}
	}
	
	private function parseGroups() {
		$groups = @file('/etc/group');
		if ($groups !== false) {
			$n = count($groups);
			$checking_gforge_groups = false;
			for ($i=0; $i < $n; $i++) {
				$group_line = $groups[$i];
				
				$group_line = trim($group_line);
				if (empty($group_line)) continue;
				
				if (preg_match('/^[[:blank:]]*'.self::GFORGE_MARKER_BEGIN.'/', $group_line)) {		// start of GForge groups
					$checking_gforge_groups = true;
					continue;
				} elseif (preg_match('/^[[:blank:]]*'.self::GFORGE_MARKER_END.'/', $group_line)) {		// end of GForge groups
					$checking_gforge_groups = false;
					continue;
				} elseif (preg_match('/^[[:blank:]]*#/', $group_line)) {		// regular comment?
					continue;
				}
				
				list($name, $password, $gid, $users) = explode(':', $group_line);
				
				if (empty($users)) {
					$users = array();
				} else {
					$users = explode(', *', $users);
				}
				$group = array("name" => $name, "password" => $password, "gid" => $gid, "users" => $users);
				
				if ($checking_gforge_groups) {
					$this->gforge_groups[$name] = $group;
				} else {
					$this->global_groups[$name] = $group;
				}
				
				$this->groups[$name] = $group;
			}
		}
	}
	
	private function parsePasswd() {
		$users = @file('/etc/passwd');
		if ($users !== false) {
			$n = count($users);
			$checking_gforge_users = false;
			for ($i=0; $i < $n; $i++) {
				$user_line = $users[$i];
				
				$user_line = trim($user_line);
				if (empty($user_line)) continue;
				
				if (preg_match('/^[[:blank:]]*'.self::GFORGE_MARKER_BEGIN.'/', $user_line)) {		// start of GForge users
					$checking_gforge_users = true;
					continue;
				} elseif (preg_match('/^[[:blank:]]*'.self::GFORGE_MARKER_END.'/', $user_line)) {		// end of GForge users
					$checking_gforge_users = false;
					continue;
				} elseif (preg_match('/^[[:blank:]]*#/', $user_line)) {		// regular comment?
					continue;
				}
				
				list($name, $password, $uid, $gid, $gcos, $home, $shell) = explode(':', $user_line);
				
				$user = array("name" => $name, "unix_password" => $password, "uid" => $uid, 
						"gid" => $gid, "gcos" => $gcos, "home" => $home, "shell" => $shell);
				
				if ($checking_gforge_users) {
					$this->gforge_users[$name] = $user;
				} else {
					$this->global_users[$name] = $user;
				}
				
				$this->users[$name] = $user;
			}
		}
	}

	public static function updateSshKey($userUnixName) {
		$c = new Criteria();
		//$c->add(UserUnixPeer::AUTHORIZED_KEYS, null, Criteria::NOT_EQUAL);
		$c->add(UserPeer::STATUS, 1);
		$c->add(UserPeer::UNIX_NAME,$userUnixName);
		$Users = UserUnixPeer::doSelectJoinUser($c);
		$user = $Users[0];

		$user_dir = SystemUtils::getHomedir($userUnixName);
		if (!is_dir($user_dir)) {
			SystemUtils::initializeHomedir($userUnixName);
		}
		if (!is_null($user->getAuthorizedKeys())) {

			$ssh_dir = SystemUtils::buildFilePath($user_dir, '/.ssh');
			if (!is_dir($ssh_dir)) mkdir($ssh_dir, 0755);
			chown($ssh_dir, $userUnixName);
			chgrp($ssh_dir, 'users');

			$ssh = "# This file is automatically generated from your account settings.\n".$user->getAuthorizedKeys();
			file_put_contents($ssh_dir.'/authorized_keys', $ssh);
			
			chmod ($ssh_dir.'/authorized_keys', 0644);
			chown ($ssh_dir.'/authorized_keys', $userUnixName);
			chgrp ($ssh_dir.'/authorized_keys', 'users');
		}
	}
}
?>
