<?php
/**
 * Manages Two Factor Authentication. This file defines the
 *  TwoFactorAuthenticator class and two custom Exception classes,
 *  TFAConfigurationException and TFACellException.
 *
 * @category Class
 * @package  Default
 * @author   Hector Pizarro <hpizarro@gforgegroup.com>
 * @license  [url] [description]
 * @link     [url] [description]
 */

require_once 'lib/Twilio.php';

/**
 * Manages Two Factor Authentication.
 *
 * @category Class
 * @package  Default
 * @author   Hector Pizarro <hpizarro@gforgegroup.com>
 * @license  [url] [description]
 * @link     (target, link)
 */
class TwoFactorAuthenticator
{
    /**
     * Min limit to choose a random numeric token for authentication
     *
     * @var integer
     */
    private $_minToken = 100000;

     /**
     * Max limit to choose a random numeric token for authentication
     *
     * @var integer
     */
    private $_maxToken = 999999;

    /**
     * Max number of consecutive times that user can fail token validation
     *
     * @var integer
     */
    private $_maxConsecTokenFailures = 3;

    /**
     * If there were too many consecutive failures, number of seconds before
     * unblocking TFA.
     *
     * @var integer
     */
    private $_failedBlockTime = 60;

    /**
     * Validates if received $cell is a valid cell number. Allowed cell should
     * contain 10 digits or more, optionally can start with a '+' sign.
     *
     * @param string $cell Cell number
     *
     * @return boolean
     */
    public function isValidCell($cell)
    {
        return preg_match('/^\+?[\d]{10,}$/', $cell) ? true : false;
    }

    /**
     * Creates a new authentication token, sends to defined $cell using SMS
     * API service and stores auth data in session.
     *
     * @param string $cell   Cell number
     * @param string $action Prefix used to set TFA data in session
     *
     * @return void
     * @throws Exception from Twilio API service
     */
    public function sendToken($cell, $action)
    {
        $from = $this->getConf('system.tfaTwilioCellNumber');
        $twilioSid = $this->getConf('system.tfaTwilioSID');
        $twilioToken = $this->getConf('system.tfaTwilioToken');
        $smsToken = mt_rand($this->_minToken, $this->_maxToken);
        $message = "GFORGE ACCESS TOKEN: $smsToken";
        try {
            $client = new Services_Twilio($twilioSid, $twilioToken);
            $client->account->messages->sendMessage($from, $cell, $message);
        } catch (Exception $e) {
            $code = $e->getCode();
            error_log("Twilio: Exception code $code");
            switch ($code) {
            case 20003: // API access denied

            case 21611: // 'From' number errors
            case 21212:
            case 21603:
            case 21606:
                throw new TFAConfigurationException($code);
                break;

            case 21211: // 'To' number errors
            case 21612:
            case 21408:
            case 21610:
            case 21614:
                throw new TFACellException($code);
                break;

            default:
                throw new Exception("Exception $code, please contact GForge Group.");
            }
        }
        //TODO remove token from error log when debug is done
        error_log("*** TFA token for action $action, cell $cell: $smsToken ***");
        $_SESSION[$action . '_tfa'] = array(
            'cell'         => $cell,
            'token'        => $smsToken,
            'lastTryTime'  => time(),
            'failsAllowed' => $this->_maxConsecTokenFailures
        );
    }

    /**
     * Validates if received $smsToken is equal to the one stored in session.
     * On success deletes TFA data from session and returns true.
     *
     * @param string $smsToken Numeric string
     * @param string $action   Prefix for TFA action: register, login, etc
     *
     * @return string          Error message, empty on success
     */
    public function verifyToken($smsToken, $action='')
    {
        $tfa = $_SESSION[$action . '_tfa'];
        $now = time();
        if (empty($action)) { // method requires $action parameter
            return "Service configuration error, please contact the GForge team.";
        }
        if (empty($tfa)) {//session has no TFA data stored, expired?
            return "Your session has expired.";
        }
        if ($tfa['failsAllowed'] <= 0) {// failed too many times
            $secBlocked = $now - $tfa['lastTryTime'];//seconds blocked
            if ($secBlocked < $this->_failedBlockTime) {
                $seconds = $this->_failedBlockTime - $secBlocked;
                return "You have exceeded the failure limit, please wait {$seconds} seconds and try again.";
            } else {// enough time passed, unblock
                $tfa['failsAllowed'] = $this->_maxConsecTokenFailures;
            }
        }
        $tfa['lastTryTime'] = $now;
        if ($smsToken != $tfa['token']) {
            $tfa['failsAllowed']--;
            $_SESSION[$action . '_tfa'] = $tfa;
            return "The token is not valid.";
        } else {
            $_SESSION[$action . '_tfa'] = array();
            return '';
        }
    }

    /**
     * Return cell stored in session validation array, empty if not found.
     *
     * @return string
     */
    public function getSessionCell($action)
    {
        if (!empty($_SESSION[$action . '_tfa']['cell'])) {
            return $_SESSION[$action . '_tfa']['cell'];
        }
        return '';
    }

    /**
     * Getter
     *
     * @codeCoverageIgnore
     * @return int
     */
    public function getMaxConsecTokenFailures()
    {
        return $this->_maxConsecTokenFailures;
    }

    /**
     * Getter
     *
     * @codeCoverageIgnore
     * @return int
     */
    public function getFailedBlockTime()
    {
        return $this->_failedBlockTime;
    }

    /**
     * Helper method, gets config parameter and returns value
     *
     * @param string $paramId Configuration parameter id
     *
     * @codeCoverageIgnore
     * @return mixed
     */
    public function getConf($paramId)
    {
        return GFConf::get($paramId);
    }
} // END class

/**
 * Exception throws on bad 'from' number configuration.
 *
 * @category Exception
 * @package  Default
 * @author   Hector Pizarro <hpizarro@gforgegroup.com>
 * @license  [url] [description]
 * @link     (target, link)
 */
class TFAConfigurationException extends Exception
{
    /**
     * Constructor
     *
     * @param int $errorCode API error code
     */
    public function __construct($errorCode)
    {
        switch ($errorCode) {
        case 20003:
            error_log("Twilio: API access denied");
            break;
        case 21611:
            // Twilio sends 1 msg x second, if your traffic is bigger
            // messages go to a queue. This error is thrown if your
            // queue is bigger than 4 hours
            error_log("Twilio: 'From' number SMS queue full");
            break;
        case 21212:
            error_log("Twilio: 'From' number invalid");
            break;
        case 21603:
            error_log("Twilio: 'From' number empty");
            break;
        case 21606:
            error_log("Twilio: 'From' number not in your account");
            break;
        }
        parent::__construct('SMS service error: configuration error detected, please contact GForge Group.', $errorCode);
    }
}

/**
 * Exception throws on bad cell destination.
 *
 * @category Exception
 * @package  Default
 * @author   Hector Pizarro <hpizarro@gforgegroup.com>
 * @license  [url] [description]
 * @link     (target, link)
 */
class TFACellException extends Exception
{
    /**
     * Constructor
     *
     * @param int $errorCode API error code
     */
    public function __construct($errorCode)
    {
        $error = 'cell number rejected';
        switch ($errorCode) {
        case 21211:
            error_log("Twilio: 'To' number invalid");
            break;
        case 21612:
            error_log("Twilio: 'To' number cannot be routed");
            break;
        case 21408:
            error_log("Twilio: 'To' international number not supported");
            break;
        case 21610:
            error_log("Twilio: 'To' number blacklisted");
            break;

        case 21614:
            error_log("Twilio: 'To' number can't receive SMS");
            $error = "cell number can't receive SMS messages";
            break;
        }
        parent::__construct("SMS service error: $error.", $errorCode);
    }
}
