<?php
/**
 * Tests Two Factor Authentication (lib/tfa/TwoFactorAuthenticator.php)
 *
 * @category Class
 * @package  Default
 * @author   Hector Pizarro <hpizarro@gforgegroup.com>
 * @license  [url] [description]
 * @link     [url] [description]
 */

$basePath = dirname(dirname(__DIR__));
require_once $basePath . '/lib/tfa/TwoFactorAuthenticator.php';

use \Mockery as m;

/**
 * Tests Two Factor Authentication (lib/tfa/TwoFactorAuthenticator.php)
 *
 * @category Class
 * @package  Default
 * @author   Hector Pizarro <hpizarro@gforgegroup.com>
 * @license  [url] [description]
 * @link     [url] [description]
 */
class TwoFactorAuthenticatorTest extends PHPUnit_Framework_TestCase
{
    const FROM_INVALID = '+15005550001';
    const FROM_NOT_SMS = '+15005550007';
    const FROM_SMS_FULL = '+15005550008';
    const FROM_VALID = '+15005550006';
    const TO_INVALID = '+15005550001';
    const TO_NO_ROUTE = '+15005550002';
    const TO_NO_INT = '+15005550003';
    const TO_BLACKLISTED = '+15005550004';
    const TO_NO_SMS = '+15005550009';
    const TO_VALID = '+15005550000';

    /**
     * Tests instance creation
     *
     * @return void
     */
    public function testServiceCreated()
    {
        $tfa = new TwoFactorAuthenticator();
        $this->assertNotNull($tfa);
    }

    /**
     * Tests isValidCell()
     *
     * @return void
     */
    public function testIsValidCell()
    {
        $tfa = new TwoFactorAuthenticator();
        $this->assertFalse($tfa->isValidCell(''));
        $this->assertFalse($tfa->isValidCell('abc'));
        $this->assertFalse($tfa->isValidCell('123456abcdefgh'));
        $this->assertFalse($tfa->isValidCell('123456'));
        $this->assertTrue($tfa->isValidCell('1234567890'));
    }

    /**
     * tests sendToken(), bad configuration
     *
     * @expectedExceptionCode 21212
     * @expectedException TFAConfigurationException
     * @return void
     */
    public function testSendTokenBadFrom1()
    {
        $_SESSION = array();
        $tfa = $this->_getMockedTFA(self::FROM_INVALID);
        $tfa->sendToken('', '');
    }

    /**
     * tests sendToken(), bad configuration
     *
     * @expectedExceptionCode 21606
     * @expectedException TFAConfigurationException
     * @return void
     */
    public function testSendTokenBadFrom2()
    {
        $_SESSION = array();
        $tfa = $this->_getMockedTFA(self::FROM_NOT_SMS);
        $tfa->sendToken('', '');
    }

    /**
     * tests sendToken(), bad configuration
     *
     * @expectedExceptionCode 21611
     * @expectedException TFAConfigurationException
     * @return void
     */
    public function testSendTokenBadFrom3()
    {
        $_SESSION = array();
        $tfa = $this->_getMockedTFA(self::FROM_SMS_FULL);
        $tfa->sendToken('', '');
    }

    /**
     * tests sendToken(), invalid 'to' cell
     *
     * @expectedExceptionCode 21211
     * @expectedException TFACellException
     * @return void
     */
    public function testSendTokenBadCell1()
    {
        $_SESSION = array();
        $tfa = $this->_getMockedTFA(self::FROM_VALID);
        $tfa->sendToken(self::TO_INVALID, 'testaction');
    }

    /**
     * tests sendToken(), 'to' cell cannot be routed
     *
     * @expectedExceptionCode 21612
     * @expectedException TFACellException
     * @return void
     */
    public function testSendTokenBadCell2()
    {
        $_SESSION = array();
        $tfa = $this->_getMockedTFA(self::FROM_VALID);
        $tfa->sendToken(self::TO_NO_ROUTE, 'testaction');
    }

    /**
     * tests sendToken(), 'to' cell not allowed, international number
     *
     * @expectedExceptionCode 21408
     * @expectedException TFACellException
     * @return void
     */
    public function testSendTokenBadCell3()
    {
        $_SESSION = array();
        $tfa = $this->_getMockedTFA(self::FROM_VALID);
        $tfa->sendToken(self::TO_NO_INT, 'testaction');
    }

    /**
     * tests sendToken(), 'to' cell blacklisted
     *
     * @expectedExceptionCode 21610
     * @expectedException TFACellException
     * @return void
     */
    public function testSendTokenBadCell4()
    {
        $_SESSION = array();
        $tfa = $this->_getMockedTFA(self::FROM_VALID);
        $tfa->sendToken(self::TO_BLACKLISTED, 'testaction');
    }

    /**
     * tests sendToken(), 'to' cell cannot receive sms
     *
     * @expectedExceptionCode 21614
     * @expectedException TFACellException
     * @return void
     */
    public function testSendTokenBadCell5()
    {
        $_SESSION = array();
        $tfa = $this->_getMockedTFA(self::FROM_VALID);
        $tfa->sendToken(self::TO_NO_SMS, 'testaction');
    }

    /**
     * tests sendToken(), successful
     *
     * @return void
     */
    public function testSendToken()
    {
        $_SESSION = array();
        $tfa = $this->_getMockedTFA(self::FROM_VALID);
        $tfa->sendToken(self::TO_VALID, 'testaction');
        $now = time();
        $atfData = $_SESSION['testaction_tfa'];
        $this->assertNotEmpty($atfData['token']);
        $this->assertEquals(self::TO_VALID, $atfData['cell']);
        $this->assertEquals($tfa->getMaxConsecTokenFailures(), $atfData['failsAllowed']);
        $this->assertTrue($atfData['lastTryTime'] >= $now);
    }

    /**
     * Tests verifyToken()
     *
     * @return void
     */
    public function testVerifyToken()
    {
        $tfa = new TwoFactorAuthenticator();
        $fakeToken ='123abc';
        $failTime = $tfa->getFailedBlockTime();
        $_SESSION = array();
        $_SESSION['_tfa'] = array();
        $_SESSION['test_tfa'] = array();
        $this->assertNotEmpty($tfa->verifyToken(''));
        $this->assertNotEmpty($tfa->verifyToken($fakeToken, 'test'));

        $_SESSION['test_tfa'] = array(
            'cell'         => '123',
            'token'        => $fakeToken,
            'lastTryTime'  => time(),
            'failsAllowed' => 5
        );
        $this->assertNotEmpty($tfa->verifyToken('badtoken', 'test'));

        $_SESSION['test_tfa'] = array(
            'cell'         => '123',
            'token'        => $fakeToken,
            'lastTryTime'  => time(),
            'failsAllowed' => 0
        );
        $this->assertNotEmpty($tfa->verifyToken($fakeToken, 'test'));

        $_SESSION['test_tfa'] = array(
            'cell'         => '123',
            'token'        => $fakeToken,
            'lastTryTime'  => time() - $failTime - 10,
            'failsAllowed' => 0
        );
        $this->assertEmpty($tfa->verifyToken($fakeToken, 'test'));
    }

    /**
     * Returns partial mock for TwoFactorAuthenticator instance
     *
     * @param string $from From number
     *
     * @return TwoFactorAuthenticator
     */
    private function _getMockedTFA($from)
    {
        $GLOBALS['from_cell'] = $from;
        $closureReturnConf = function ($paramId) { //create config closure
            switch ($paramId) {
            case 'system.tfaTwilioCellNumber':
                return $GLOBALS['from_cell'];
            case 'system.tfaTwilioSID':
                return 'AC820ea5a1f51b3d4c2490b8d12c8011dc';
            case 'system.tfaTwilioToken':
                return 'c4932801b20978c5ecb227b888c2a2d0';
            default:
                return '';
            }
        };
        $tfa = \Mockery::mock('\TwoFactorAuthenticator[getConf]');
        $tfa->shouldReceive('getConf')->andReturnUsing($closureReturnConf);
        return $tfa;
    }
}
