<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Ruben Gutierrez	<ruben@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
 
require_once('lib/SystemUtils.php');
require_once('GFConf.php');
 
/**
 * Cruisecontrol daemon. 
 */
class CruisecontrolDaemon {
	/** PID file */
	const PID_FILE = '/tmp/gforge-cc.pid';
	const WAKEUP_TIME = 60;		// Spawn daemon every minute
	const STOP_WAIT_TIME = 15;		// Seconds to wait to retry stopping CC
	const STOP_RETRY = 5;		// How many times we'll try to stop CC

	/** Do not instantiate directly */
	private function __construct() {
	}
	
	public static function start() {
		if (!function_exists("posix_setgid")) {
			throw new Exception("POSIX functions not present in PHP");
		}
		if (!function_exists("pcntl_fork")) {
			throw new Exception("pcntl_fork function not present in PHP. Please recompile with --enable-pcntl.");
		}
		if (self::isRunning()) {
			throw new Exception("CruiseControl daemon is already running");
		}

		$userData = posix_getpwuid(posix_getuid());
		if ($userData['name'] != 'root') {
			throw new Exception("Must be root for running the daemon");
		}
		
		if (!self::fork()) {
			throw new Exception("Couldn't fork process");
		}
	}
	
	public static function stop() {
		if (!self::isRunning()) {
			throw new Exception('CruiseControl daemon is not running');
		}
		$abort = false;
		$retry = 0;
		while(!self::canStop() && $abort !== true) {
			if ($retry > (self::STOP_RETRY-2)) {
				$abort = true;
				break;
			}
			sleep(self::STOP_WAIT_TIME);
			$retry++;
		}
		if ($abort !== true) {
			$pid = self::getPID();
			if (!posix_kill($pid, SIGTERM)) {
				throw new Exception("Could not kill PID $pid");
			} else {
				self::killCC();
				for($i=0;$i<5;$i++) {
					sleep(1);					
					if (!self::isRunning()) {
						break;
					}
				}
				if (self::isRunning()) {
					throw new Exception("Could not stop CruiseControl.");
				}
			}
		} else {
			throw new Exception("Could not stop. CruiseControl is currently building");
		}
	}
	
	public static function isRunning() {
		if (file_exists(self::PID_FILE) && ($pid = @file_get_contents(self::PID_FILE))) {
			// this will tell if process is running without actually killing it
			if (posix_kill($pid, 0)) {
				return true;
			} else {
				$cc = self::ccIsRunning();
				if (count($cc) > 0) {
					return true;
				} else {
					return false;
				}
			}
		} else {
			return false;
		}
	}
	
	/**
	 * Returns the PID
	 */
	public static function getPID() {
		if (!file_exists(self::PID_FILE)) {
			return 0;
		}
		$pid = file_get_contents(self::PID_FILE);
		return $pid;
	}

	/**
	 * Saves the PID
	 */
	private static function savePID() {
		$fh = fopen(self::PID_FILE, "w");
		if (!$fh) {
			echo "Could not open ".self::PID_FILE." for writing. Aborting.\n";
			exit(1);
		}
		fwrite($fh, posix_getpid());
		fclose($fh);
	}
	
	private static function killCC() {
		$cc = self::ccIsRunning();
		if (count($cc) > 0) {
			foreach($cc as $line) {
				$parts = explode(' ', trim($line));
				if (count($parts) > 0 && is_numeric($parts[0])) {
					posix_kill($parts[0], SIGTERM);
				}
			}
		}
	}
	
	private static function ccIsRunning() {
		$cc = array();
		$cc_exec = 'ps a | grep "java.* CruiseControl" | grep -v "grep"';
		SystemUtils::execute($cc_exec, $cc);
		return $cc;
	}
	
	
	/**
	 * Begins the daemon process
	 */
	private static function fork() {
		$pid = pcntl_fork();
		if ($pid == -1) {
			return false;
		} else if ($pid != 0) {
			return true;
		}
		self::handleDaemon();
	}
	
	public static function handleSignal($signal) {
		// Clean up
        @unlink(self::PID_FILE);
		exit(0);
	}
	
	/**
	 * Handler for the daemon process
	 */
	private static function handleDaemon() {
		// Install the signal handler
		declare(ticks = 1);		// Somehow this is needed... else it won't work
		$signals = array(SIGTERM, SIGHUP, SIGQUIT);
		foreach ($signals as $signal) {
			if (!pcntl_signal($signal, array('self', "handleSignal"))) {
				echo "Could not install signal handler. Aborting.\n";
				exit(1);
			}
		}
		self::daemonLoop();
	}


	/**
	 * Main daemon loop
	 */
	private static function daemonLoop() {
		$su_exec = trim(`which su`);
		$ccdir = GFConf::get('build.pathToCC');
		$ccuser = GFConf::get('build.unixUser');
		$java_home = GFConf::get('build.pathToJava');
		$path = $java_home.'/bin:' . trim(`echo \$PATH`);
		$lib_dir = "$ccdir/lib";
		$class_path = "$java_home/lib/tools.jar:$lib_dir/cruisecontrol.jar:$lib_dir/log4j.jar:$lib_dir/jdom.jar:$lib_dir/ant.jar:$lib_dir/ant-launcher.jar:$lib_dir/jasper-compiler.jar:$lib_dir/jasper-runtime.jar:$lib_dir/xercesImpl-2.8.0.jar:$lib_dir/xml-apis-2.8.0.jar:$lib_dir/xmlrpc-2.0.1.jar:$lib_dir/xalan-2.7.0.jar:$lib_dir/serializer-2.7.0.jar:$lib_dir/jakarta-oro-2.0.3.jar:$lib_dir/mail.jar:$lib_dir/junit.jar:$lib_dir/activation.jar:$lib_dir/commons-net-1.1.0.jar:$lib_dir/starteam-sdk.jar:$lib_dir/mx4j.jar:$lib_dir/mx4j-tools.jar:$lib_dir/mx4j-remote.jar:$lib_dir/smack.jar:$lib_dir/comm.jar:$lib_dir/x10.jar:$lib_dir/fast-md5.jar:$lib_dir/maven-embedder-2.0.3-dep.jar:$lib_dir/javax.servlet.jar:$lib_dir/org.mortbay.jetty.jar:$lib_dir/commons-logging.jar:$lib_dir/commons-el.jar:$lib_dir/gforge-publisher.jar:$lib_dir/gforge-axis.jar:$lib_dir/jaxrpc.jar:$lib_dir/axis.jar:$lib_dir/commons-discovery-0.2.jar:$lib_dir/saaj.jar:$lib_dir/wsdl4j-1.5.1.jar:$lib_dir/commons-httpclient-3.0-rc3.jar:$lib_dir/commons-codec-1.3.jar:$lib_dir/crimson.jar:.";
		$options = array('-', $ccuser, 
						'--command=cd '.$ccdir.';'.$java_home.'/bin/java -cp '.$class_path.' -Djavax.management.builder.initial=mx4j.server.MX4JMBeanServerBuilder CruiseControl -jmxport 8000');
		chdir($ccdir);
		self::savePID();
		pcntl_exec($su_exec, $options, array('CLASSPATH'=>$class_path,'JAVA_HOME'=>$java_home, 'PATH'=>$path));
	}

	private static function canStop() {
		$today = date('Y-m-d');
		$pathToCC = GFConf::get('build.pathToCC');
		chdir($pathToCC);
		$build = array();
		$idle = array();
		$build_exec = 'grep "'.$today.'.* now building" cruisecontrol.log';
		$idle_exec = 'grep "'.$today.'.* idle" cruisecontrol.log';
		SystemUtils::execute($build_exec, $build);
		SystemUtils::execute($idle_exec, $idle);
		
		$max_build_time = 0.0;
		foreach ($build as $b) {
			if (strlen($b) > 23) {
				$date = substr($b, 0, 23);
				list($date, $sec) = explode(',', $date);
				$auxtime = (strtotime($date) * 100) + ($sec/10);
				if ($auxtime >= $max_build_time) {
					$max_build_time = $auxtime;
				}
			}
		}
		$max_idle_time = 0.0;
		foreach ($idle as $i) {
			if (strlen($i) > 23) {
				$date = substr($i, 0, 23);
				list($date, $sec) = explode(',', $date);
				$auxtime = (strtotime($date) * 100) + ($sec/10);
				if ($auxtime >= $max_idle_time) {
					$max_idle_time = $auxtime;
				}
			}
		}
		if ($max_build_time == 0) {
			return true;
		} else if ($max_idle_time > $max_build_time) {
			return true;
		} else {
			return false;
		}
	}
}
?>