<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version $Id$
 * @author Daniel Perez	danielperez.arg@gmail.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('notification/GFNotification.php');
require_once("plugin/ActionManager.php");
require_once('StringUtils.php');
require_once('GFUrl.php');

/**
 * Note that the "Forum message notification" is actually a Forum notification. When
 * ANY message is added to the forum, we must notify the user
 */
class ForumMessageNotification implements GFNotification {
	private $ForumMessage;
	private $Forum;
	private $appendString;


	function __construct(ForumMessage $ForumMessage, $appendString='') {
		$this->ForumMessage = $ForumMessage;
		$this->Forum = $ForumMessage->getForumThread()->getForum();
		$this->appendString = $appendString;
	}

	/**
	 *	getNotificationMessage - construct a string of this message content.
	 *
	 *	@param	User
	 *	@return string
	 */
	function getNotificationMessage(User $User) {
		return $this->getGenericMessage();
	}

	function getPosterName() {
		$user = $this->ForumMessage->getUser();
		return $user->getFirstname() . ' ' . $user->getLastname() ;
	}

	/**
	 * This "generic" message will be used for both users and non-users when
	 * a new post is made
	 * @throws InvalidDataException
	 * @throws Exception
	 * @return   string
	 */
	private function getGenericMessage() {
		$url = GFurl::getUrl();

		if ($this->Forum->getSection() != 'project') {
			throw new Exception('Tried to notify about a non-project forum');
		}

		/** @var Project $project  */
		$project = ProjectPeer::retrieveByPk($this->Forum->getRefId());
		$poster = UserPeer::retrieveByPk($this->ForumMessage->getCreatedBy());

		if (!$project || !$poster) {
			throw new InvalidDataException("Invalid project or user name in " . __FUNCTION__);
		}

		// make the body
		$body = 'A new message was posted to: ' . $project->getProjectName() .' - ' .
			$this->ForumMessage->getForumThread()->getThreadName() . "\n";
		$body .= "By " . $poster->getFirstname() . ' ' .$poster->getLastname() . ' at ' .
			$this->ForumMessage->getPostDate() . "\n";

		$body .= StringUtils::getPlainUTF8FromHtml($this->ForumMessage->getBody());
		if (!empty($this->appendString)) {
			$body .= $this->appendString;
		}
		$body .= "\n\n";


		$body .=  "To respond simply reply this email or visit:\n" .
			$url->getFullURL($url->projectUrl($project,'ForumBrowse', 'forum', '',
				array('_forum_action'=>'MessageReply','message_id' => $this->ForumMessage->getForumMessageId())),true) . "\n";
		$body .= "\nYou are receiving this email because you elected to monitor this forum.".
		"\nTo stop monitoring this forum, please visit: ".
		$url->getFullUrl($url->getMonitorUrl('forum', $this->Forum->getPrimaryKey(),$url->projectUrl($project, '', 'forum'), false),true);
		$body .= "\n\n______________________________________________________________________";


		return $body;
	}

	/**
	 *	getNotificationSubject - construct a string of this message subject.
	 *
	 *	@param	User
	 *	@return string
	 */
	function getNotificationSubject(User $User) {
		return $this->getGenericSubject();
	}

	/**
	 * This "generic" subject will be used for both users and non-users when
	 * a new post is made
	 * @return	string
	 */
	private function getGenericSubject() {
		$subject = "[".$this->Forum->getForumName()."] ". StringUtils::unHtmlSpecialChars($this->ForumMessage->getSubject());
		return $subject;
	}


	/**
	 *	getUsers - get an array of user objects
	 *	@return array
	 */
	public function getUsers() {
		return $this->ForumMessage->getForumThread()->getForum()->getMonitoringUsers(); // thread or forum or project monitors
	}

	/**
	 *	getSection - retrieves the section name of the forum message
	 *	@return string
	 */	
	function getSection() {
		return ForumMessage::SECTION;
	}
	
	/**
	 *	getRefId - retrieves the id of the forum message that the notification belongs to
	 *	@return int
	 */	
	function getRefId() {
		return $this->ForumMessage->getForumMessageId();
	}

	function getMessageId() {
		return '<'.$this->getSection().'-'.$this->getRefId().'@'.GFConf::get('system.hostName').'>';
	}

	function getInReplyTo() {
			//Always use the last message
			$c = new Criteria();
			// Same thread
			$c->add(ForumMessagePeer::FORUM_THREAD_ID, $this->ForumMessage->getForumThread()->getPrimaryKey());
			// Must be before this message (ID)
			$c->add(ForumMessagePeer::FORUM_MESSAGE_ID, $this->ForumMessage->getPrimaryKey(), Criteria::LESS_THAN);
			$c->addDescendingOrderByColumn(ForumMessagePeer::FORUM_MESSAGE_ID);
			// We only need one record
			$c->setLimit(1);

			/** @var ForumMessage[] $messages  */
			$messages = $this->ForumMessage->getForumThread()->getForumMessages($c);
			if (count($messages) > 1) {
				$inReplyTo = $messages[1]->getForumMessageId();
			} else {
				//First message in thread, don't set the header
				return '';
			}
		return '<'.$this->getSection() . '-' . $inReplyTo.'@'.GFConf::get('system.hostName').'>';
	}

	function getReferences() {
		return $this->getInReplyTo();
	}

	/**
	 * This will notify of the new message to the addresses defined in 
	 * "send_all_posts_to" field
	 */
	public function notifyAllPosts() {
		$addresses = StringUtils::getAddressesFromList($this->Forum->getSendAllPostsTo());
		
		foreach ($addresses as $address) {
			$message = $this->getGenericMessage();
			$subject = $this->getGenericSubject().' (forummessage-'.$this->ForumMessage->getPrimaryKey().')';
			
			SystemUtils::sendMail($subject, $message, $address,false,SystemUtils::getFromAddress('gforge-gateway'));
		}
	}
}

?>
