<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Marcelo Mottalli <marcelo@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('system/Cronjob.php');
require_once('plugins/mailman/lib/MailmanPlugin.php');
require_once('plugins/mailman/lib/Mailman.php');
require_once('plugins/mailman/lib/MailboxParser.php');
require_once('plugins/forum/lib/Forum.php');
require_once('plugins/forum/lib/ForumMessage.php');
require_once('core/Plugin.php');
require_once('core/Project.php');
require_once('core/ProjectPlugin.php');
require_once('Validator.php');
require_once('lib/StringUtils.php');

class CreateThreadsCronjob extends Cronjob {
	public function __construct() {
		$this->jobname = basename(__FILE__);
		parent::__construct();
	}
	
	public function run() {
		// retrieve all the projects that have mailman plugin
		$c = new Criteria();
		$c->add(PluginPeer::PLUGIN_NAME, 'mailman');
		$c->addJoin(ProjectPluginPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->addJoin(ProjectPeer::PROJECT_ID, ProjectPluginPeer::PROJECT_ID);
		$projects = ProjectPeer::doSelect($c);
		
		foreach ($projects as $project) {
			$this->parseLists($project);
		}
		$this->commitLog();
	}
	
	private function parseLists(Project $project) {
		// Get all the lists for this project
		$c = new Criteria();
		$c->add(MailmanPeer::PROJECT_ID, $project->getPrimaryKey());
		$lists = MailmanPeer::doSelectJoinProject($c);
		
		foreach ($lists as $list) {
			$this->parseList($list, $project);
		}
	}
	
	private function parseList(Mailman $list) {
		$listname = $list->getListName();
		$listdir = SystemUtils::buildFilePath(GFConf::get('mailman.mailmanListsDir'), '/archives/private/'.$listname.'.mbox');
		if (!is_dir($listdir)) {
			$this->logCron('Directory for list \''.$listname.'\' ('.$listdir.') doesn\'t exist. Skipping.');
			return;
		}
		
		// Load the existing messages in an array for quick access
		$existing_messages = array();
		$c = new Criteria();
		$c->addSelectColumn(ForumMessagePeer::FORUM_MESSAGE_ID);
		$c->addSelectColumn(ForumMessagePeer::MAIL_MSG_ID);
		$c->addSelectColumn(ForumMessagePeer::FORUM_THREAD_ID);
		$c->addJoin(ForumMessagePeer::FORUM_THREAD_ID, ForumThreadPeer::FORUM_THREAD_ID);
		$c->addJoin(ForumThreadPeer::FORUM_ID, ForumPeer::FORUM_ID);
		$c->add(ForumPeer::SECTION, 'mailman');
		$c->add(ForumPeer::REF_ID, $list->getPrimaryKey());
		$rs = ForumMessagePeer::doSelectRS($c);
		
		while ($rs->next()) {
			$message_id = $rs->get(1);
			$mail_msg_id = $rs->get(2);
			$thread_id = $rs->get(3);
			$existing_messages[$mail_msg_id] = array('message_id' => $message_id, 'thread_id' => $thread_id);
		}
		
		// Get the forum for this mailing list. Create one if it doesn't exist.
		$c = new Criteria();
		$c->add(ForumPeer::SECTION, 'mailman');
		$c->add(ForumPeer::REF_ID, $list->getPrimaryKey());
		$forum = ForumPeer::doSelectOne($c);
		if (is_null($forum)) {
			$forum = new Forum();
			$forum->setSection('mailman');
			$forum->setRefId($list->getPrimaryKey());
		}
		
		// Update the forum data, if needed		
		$forum->setForumName($list->getListName());
		$forum->setIsPublic($list->getIsPublic());
		$forum->setDescription($list->getListDescription());
		$forum->save();


		// Get the mailboxes for this list
		$dh = opendir($listdir);
		while ($boxname = readdir($dh)) {
			if ($boxname == '.' || $boxname == '..') {
				continue;
			}
			
			$mailbox = SystemUtils::buildFilePath($listdir, $boxname);
			if (!is_file($mailbox)) {
				continue;
			}
			
			// Ok, at this point we've got the mailbox file. Parse it.
			$this->parseMailbox($mailbox, $list, $forum, $existing_messages);
		}
		closedir($dh);
		
	}
	
	private function parseMailbox($mailbox, Mailman $list, Forum $forum, &$existing_messages) {
		$parser = new MailboxParser($mailbox);
		$parser->setSizeLimit(1024*100);		// Limit messages to 100 kilobytes

		while ($mail = $parser->nextMessage()) {
			$message_id = $mail->getMessageId();
			if (!$mail->isValid()) {
				// Mail could not be parsed (usually weird multipart stuff)
				$this->logCron("Message ID $message_id could not be parsed. Error is: \"".$mail->invalid_message."\". Skipping...");
				continue;
			}

			if (!$message_id || array_key_exists($message_id, $existing_messages)) {
				// Mail was alread parsed or it doesn't have an ID.
				// If it doesn't have an ID (shouldn't happen) we won't be able to track
				// if it was already parsed or not)
				continue;
			}
			
			$users = UserPeer::retrieveByEmail($mail->getFrom());
			if (count($users) == 0) {
				$user = null;
			} else {
				$user = $users[0];		// Assume only one was returned
			}
			
			$message = new ForumMessage();
			
			// See if this is a reply to another message (RFC 2822)
			$reply_to_id = $mail->getMainHeader('in-reply-to');
			if (!$reply_to_id || !is_object($reply_to_id)) {
				$reply_to_id = 0;
			} else {
				$reply_to_id = $reply_to_id[0]->text;
			}

			if (!array_key_exists($reply_to_id, $existing_messages)) {
				// If this message isn't a reply to any message, we should start a new thread
				// for it
				$thread = new ForumThread();
				$thread->setThreadName(Validator::sanitizeString($mail->getSubject()));
				$thread->setForum($forum);
				if (is_null($user)) {
					$thread->setCreatedBy(User::UNDEFINED_USER_ID);
				} else {
					$thread->setCreatedBy($user->getUserId());
				}
				$thread->save();
				
				$message->setForumThreadId($thread->getForumThreadId());
			} else {
				// it's a reply to some message, and that message belongs to a thread
				$message->setForumThreadId($existing_messages[$reply_to_id[0]->text]['thread_id']);
				$message->setParentForumMessageId($existing_messages[$reply_to_id[0]->text]['message_id']);
			}
			
			if (is_null($user)) {
				$message->setCreatedBy(User::UNDEFINED_USER_ID);
			} else {
				$message->setCreatedBy($user->getUserId());
			}

			$message->setSubject(Validator::sanitizeString($mail->getSubject()));
			
			$body = Validator::sanitizeString($mail->getBody());
			$body = trim($body);
			$body = nl2br($body);
			if (empty($body)) {
				$message->setBody(Language::getSessText('Mailman.EmptyMessage'));
			} else {
				$message->setBody($body);
			}
			
			$message->setMailMsgId($mail->getMessageId());
			$message->setMailSender(Validator::sanitizeString($mail->getFullFrom()));
			$message->setPostDate(date('Y-m-d h:i:s', $mail->getDate()));
			$message->save();
			
			$existing_messages[$mail->getMessageId()] = array('message_id' => $message->getForumMessageId(), 
																'thread_id' => $message->getForumThreadId());
		}
	}
}

$cronjob = new CreateThreadsCronjob();
$cronjob->run();

?>
