<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Ruben Gutierrez ruben@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('lib/initialize.php');
require_once('system/Cronjob.php');
require_once('core/ProjectPeer.php');
require_once('core/PluginPeer.php');
require_once('core/ProjectPluginPeer.php');
require_once('GFConf.php');
require_once('plugins/scm/lib/ProjectScmPeer.php');
require_once('plugins/scm/lib/ScmModulePeer.php');
require_once('plugins/scm/lib/SCMRepositoryFactory.php');

class ScmCheckoutWww extends Cronjob {

	public function __construct() {
		$this->jobname = basename(__FILE__);
		parent::__construct();
	}
	
	public function run() {
		$c = new Criteria();
		$c->add(PluginPeer::EXCLUSIVE_TYPE, ScmPlugin::SECTION);
		// For now only CVS and SVN supported
		$criterion = $c->getNewCriterion(PluginPeer::PLUGIN_NAME, ScmcvsPlugin::getName());
		$criterion->addOr($c->getNewCriterion(PluginPeer::PLUGIN_NAME, ScmsvnPlugin::getName()));
		$c->add($criterion);
		$ProjectPlugins = ProjectPluginPeer::doSelectJoinPlugin($c);

		foreach ($ProjectPlugins as $ProjectPlugin) {
			$Project = $ProjectPlugin->getProject(); // get the project
			$Repo = SCMRepositoryFactory::getRepository($Project);
			if ($Repo != null && is_object($Repo)) {
				$c = new Criteria();
				$c->add(ScmModulePeer::PROJECT_ID, $Project->getProjectId());
				$c->add(ScmModulePeer::PUBLISH_TO_WWW, true);
				$ScmModule = ScmModulePeer::doSelectOne($c);
				if ($ScmModule != null && is_object($ScmModule)) {
					try {
						$this->doCheckout($Project, $Repo, $ScmModule);
					} catch (Exception $ex) {
						$this->logCron('Web Publishing checkout for project: ' . $Project->getProjectName() . '. Failed: '.$ex->getMessage());
					}
				}
			}
		}
		$this->commitLog();
	}
	
	private function doCheckout($Project, $Repo, $ScmModule) {
		if (!$Repo->isConfigured() || (GFConf::get('system.splitScmServers') == false && !$Repo->isCreated())) {
			throw new Exception('SCM Repository for project ' . $Project->getProjectName() . ' is not yet configured or created');
		}
		
		$destDir = $this->prepareFolder($Project);
		if ($Repo->getType() == 'cvs') {
			$this->doCVSCheckout($Project, $Repo, $ScmModule, $destDir);
		} else if ($Repo->getType() == 'svn') {
			$this->doSVNCheckout($Project, $Repo, $ScmModule, $destDir);
		}		
	}
	
	private function prepareFolder($Project) {
		$unixName = $Project->getUnixName();
		$ret = false;
		
		$apache_user = GFConf::get('system.apacheUser');
		$apache_group = GFConf::get('system.apacheGroup');

		$vhostDir = GFConf::get('system.projectsVHostDir');
		$projectVHostDir = SystemUtils::buildFilePath($vhostDir, $unixName);
		
		if (is_dir($projectVHostDir)) {
			return $projectVHostDir;
		}
		
		$this->logCron("Creating directory $projectVHostDir...");
		if (!mkdir($projectVHostDir, 0770, true)) {
			throw new Exception("Unable to create directory $projectVHostDir");
		}

		return $projectVHostDir;
	}
	
	private function doCVSCheckout($Project, $Repo, $ScmModule, $destDir) {
		$apache_user = GFConf::get('system.apacheUser');
		$apache_group = GFConf::get('system.apacheGroup');
		
		$projectScm = $Repo->getScmConfiguration();

		if (!$projectScm->getControlled()) {
			//@todo	This will be used when we add support for external SCM
			$this->logCron('Controlled SCM is not supported yet.');
			return false;
		} else {
			$binary = GFConf::get('scmcvs.cvsbin');
			$moduleName = $ScmModule->getPath();
			$moduleName = ($moduleName{0} == '/') ? substr($moduleName, 1, strlen($moduleName)-1) : $moduleName;
			$output = array();			

			chdir($destDir);
			if (file_exists($destDir.'/CVS') && is_dir($destDir.'/CVS')) {
				// Update the existing repository
				$cmd = 'export CVS_RSH=ssh && '.$binary.' update -dP';
				$msg = 'Web Publishing update for project: ' . $Project->getProjectName() . '. OK!';
			} else {
				if (GFConf::get('system.splitScmServers')) {
					// If using split SCM we must use SSH access. NOTE: needs SSH keys to access the remote server as root!
					$location = ':ext:root@'.$Repo->getSCMHostname().':'.$Repo->getRoot();
					// For some weird reason CVS does not allow the -d $destDir command when performing a checkout via ssh
					// so we must fake that functionality by checking out the module and moving al the files inside the
					// module to $destDir
					$cmd = "export CVS_RSH=ssh && cd $destDir && mkdir __tmp__ && cd __tmp__ && $binary -d $location co $moduleName && cd .. && mv __tmp__/$moduleName/* . && rm -rf __tmp__";
				} else {
					// This means we're performing a local checkout as root
					$location = $Repo->getRoot();
					$cmd = "$binary -d $location co -d $destDir/ $moduleName";
				}
				
				$msg = 'Web Publishing checkout for project: ' . $Project->getProjectName() . '. OK!';
			}
			
			$this->logCron("Executing command: $cmd");
			$error = SystemUtils::execute($cmd, $output);

			if ($error) {
				throw new Exception("Unable to perform checkout/update for project (error code [$error]): ".implode("\n", $output));
			}

			$this->logCron($msg);
			
			SystemUtils::execute("chown -h -R {$apache_user}:{$apache_group} {$destDir}", $output);
		}
	}
	
	private function doSVNCheckout($Project, $Repo, $ScmModule, $destDir) {
		$apache_user = GFConf::get('system.apacheUser');
		$apache_group = GFConf::get('system.apacheGroup');

		$projectScm = $Repo->getScmConfiguration();

		if (!$projectScm->getControlled()) {
			//@todo	This will be used when we add support for external SCM
			$this->logCron('Controlled SCM is not supported yet.');
			return false;
		} else {
			$binary = GFConf::get('scmsvn.svnpath') . '/svn';
			$output = array();

			chdir($destDir);
			if (file_exists($destDir.'/.svn') && is_dir($destDir.'/.svn')) {
				$cmd = $binary.' update';
				$msg = 'Web Publishing update for project: ' . $Project->getProjectName() . '. OK!';
			} else {
				if (GFConf::get('system.splitScmServers')) {
					// If using split SCM we must use SSH access. NOTE: needs SSH keys to access the remote server as root!
					$location = 'svn+ssh://root@'.$Repo->getSCMHostname().$Repo->getRoot().$ScmModule->getPath();
				} else {
					// This means we're performing a local checkout as root
					$location = 'file://'.$Repo->getRoot().$ScmModule->getPath();
				}
				
				$cmd = "$binary co $location $destDir";
				$msg = 'Web Publishing checkout for project: ' . $Project->getProjectName() . '. OK!';
			}

			$this->logCron("Executing command: $cmd");
			$error = SystemUtils::execute($cmd, $output);

			if ($error) {
				throw new Exception("Unable to perform checkout/update for project: ".implode("\n", $output));
			}
			
			$this->logCron($msg);
			
			SystemUtils::execute("chown -h -R {$apache_user}:{$apache_group} {$destDir}", $output);
		}
	}
}

$cronjob = new ScmCheckoutWww();
echo $cronjob->run();
?>
