<?php

/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Marcelo Mottalli <marcelo@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
require_once('plugins/scm/lib/ScmCommit.php');
require_once('plugins/scm/lib/ScmCommitFile.php');
require_once('lib/Validator.php');

/**
 * This class is in charge of creating the correct entries on the database after a
 * new SCM commit is made.
 */
class ScmCommitWrapper {

	private $repository;
	private $user;
	private $log;
	private $date;
	private $files;
	private $project;
	private $commitIds;
	private $extraIds;

	/**
	 * Constructor.
	 * A special class is needed for this since usually the commits are done under a special user,
	 * so we need a way to log in in the system as the user that's making the commit.
	 *
	 *
	 * @param SCMRepository $repository   Repositoy in which the commit is being made.
	 * @param User $user   User that is making the commit. If it's not a GForge user, must be null.
	 * @param string $log   Commit log
	 * @param mixed $files   Array with information about the changed files
	 * @param string $date
	 * The following two are only used by git repos
	 * @param array $commitIds
	 * @param array $extraIds
	 */
	public function __construct(SCMRepository $repository, $user, $log, $files, $date = null, $commitIds=array(),$extraIds=array()) {
		$this->repository = $repository;
		$this->commitIds = $commitIds;
		$this->extraIds = $extraIds;

		if (!is_object($user)) {
			$this->user = UserPeer::retrieveByPk(User::UNDEFINED_USER_ID);
		} else {
			$this->user = $user;
		}

		$this->log = Validator::sanitizeString($log);
		$this->date = (!$date) ? time() : $date;
		/* Each of $files must have the following format:
		  array(
		  'path' => Path to the file inside the repository (i.e. /module1/README
		  'prev_version' => Version of the file in the repository before the commit
		  'new_version' => New version of the file afeter the commit
		  'changed_lines' => Number of changed lines from previous version (positive/negative number)
		  )
		 */
		$this->files = $files;
		$this->project = $repository->getProject();
	}

	/**
	 * Parse the commit data 
	 */
	public function generateCommitData() {
		// Sets the session user as the user that committed the files
		$session = Session::getSession();
		if ($session->getUser() == null)
			$session->setUser($this->user);

		$commit = null;

		$type = $this->repository->getType();
		if (strtolower($type) == 'cvs') {
			// Unfortunately, when we make a commit with files that are on different directories in the repo, cvs 
			// makes one call to the loginfo script per each directory set. to avoid multiple entries in the
			// scm_commit table, we check that the same exact log for the same user doesn't appear in there. if it does, it's the
			// case of multiple post by cvs.
			$c = new Criteria();
			$c->add(ScmCommitPeer::MESSAGE_LOG, $this->log);
			$c->add(ScmCommitPeer::USER_ID, $this->user->getUserId());
			$c->add(ScmCommitPeer::PROJECT_ID, $this->project->getProjectId());

			// Assume that committs separated by 1 minute are different commits
			// This may cause problems with timezones?
			$c->add(ScmCommitPeer::COMMIT_DATE, time() - 60, Criteria::GREATER_THAN);
			$commit = ScmCommitPeer::doSelectOne($c);
		}

		if (is_null($commit)) {
			$commit = new ScmCommit();
			$commit->setCommitDate($this->date);
			$commit->setUser($this->user);
			$commit->setMessageLog($this->log);
			$commit->setProject($this->project);
			if ($type == 'cvs' || $type = 'svn') {
				$commit->save(null, false);
			} else {
				$commit->save();
			}
		}

		foreach ($this->files as $file) {
			$this->addFileLog($file, $commit);
		}

		$associations = array();
		if ($this->repository->getType() == 'git') {
			$tmp = array_merge($this->commitIds, $this->extraIds);
			foreach ($tmp as $id) {
				$associations['[#'.$id.']'] = array(TrackerItem::SECTION, $id);

			}
		} else {
			$associations = StringUtils::extractAssociationsFromString($this->log);
		}

		// Trigger the action
		$params = array(
			'commit' => $commit,
			'project' => $this->project,
			'associations' => $associations
		);
		ActionManager::notify("SCM_COMMIT", $params);

		//Extract status change from commit logs
		$status_changes = ScmCommit::getStatusFromMessages($this->log);

		//Set status to tracker items
		foreach($status_changes['valid'] as $tid => $status) {
			TrackerItemPeer::changeTiStatus($tid, $status);
		}

		//Extract time entries from commit logs
		$time_entries = ScmCommit::getTimeEntriesFromMessages($this->log);

		//Create time entries on tracker items
		foreach($time_entries['valid'] as $tid => $time_entry) {
			TrackerItemPeer::createTimeEntry($tid, $time_entry);
		}

		try {
			$mentions = StringUtils::getMentionedUsers($this->log);
			if (count($mentions) > 0) {
				require_once('RepositoryBrowser.php');
				$browser = new RepositoryBrowser($this->repository);

				$userUrl = GFUrl::getUrl()->userURL($this->user);
				$userlink = "<a href=\"{$userUrl}\" target=\"_new\" title=\"User profile for {$this->user->getUnixName()}\">{$this->user->getUnixName()}</a>";
				$commit_url = $browser->getCommitDetailUrl($commit->getScmCommitId());
				$link = "<a href=\"{$commit_url}\" title=\"View Commit\" >Commit {$commit->getScmCommitId()}</a>";
				$commit_text = $this->log;
				if (strlen($commit_text) > 140) {
					$commit_text = substr($commit_text, 0, 140) . '...';
				}

				$wall_comment = "$userlink mentioned you in $link: <br /><br />{$commit_text}";
				ForumPeer::sendUserWallMentions($mentions, $this->user->getUserId(), $wall_comment);
			}
		} catch (Exception $ex) {
			error_log("Exception while trying to send SCM commit @mentions: {$ex->getMessage()}");
			error_log($ex->getTraceAsString());
		}
	}

	/**
	 * Add a entry in the database for a file associated to a commit
	 *
	 * @param array	The file array with name and version numbers
	 * @param ScmCommit $commit
	 *
	 */
	private function addFileLog($file, ScmCommit $commit) {
		// first see if there's a duplicate file
		$c = new Criteria();
		$c->add(ScmCommitFilePeer::SCM_COMMIT_ID, $commit->getScmCommitId());
		$c->add(ScmCommitFilePeer::MODIFIED_FILE, $file['path']);
		$f = ScmCommitFilePeer::doSelectOne($c);
		if (!is_null($f)) {
			$f->delete();
		}


		$commitfile = new ScmCommitFile();
		//TODO: now, the path relative to the project is stored, that means the repository name doesn't go into the file field
		// this is different from cvstracker, which has the cvs repo name already in the file. perhaps we will have to
		// adjust this (or perhaps not). look into this issue when whe integrate the data in the db with viewcvs or whatever
		$commitfile->setNewVersion($file['new_version']);
		$commitfile->setPrevVersion($file['prev_version']);
		$commitfile->setModifiedFile($file['path']);

		$commitfile->setScmCommitId($commit->getPrimaryKey());
		$commitfile->save();
	}

}

?>
