<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version $Id$
 * @author Marcelo Mottalli <marcelo@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('system/SystemConfigManager.php');
require_once('system/DAVConfigManager.php');
require_once('plugins/scm/lib/ScmPlugin.php');
require_once('lib/core/UserUnixPeer.php');


/**
 * @property Project $project
 * @property SCMRepository $repository
 * @property ProjectScm $projectScm
 */
class ScmUserManager
{
	private $repository, $project, $projectScm;
	private $davAccessFile, $davAuthFile;
	
	public function __construct($repository) {
		$this->repository = $repository;
		$this->project = $repository->getProject();
		$this->projectScm = $repository->getScmConfiguration();
		
		$this->davAccessFile = SystemUtils::buildFilePath(GFConf::get('system.varDir'), $repository->getType().'root-access');
		$this->davAuthFile = SystemUtils::buildFilePath(GFConf::get('system.varDir'), $repository->getType().'root-authfile');
	}
	
	public function updateSystemUsers() {
		if ($this->repository->getAccessType() == SCMRepository::ACCESS_TYPE_SSH) {
			$this->updateSSHUsers();
			$this->deleteDAVUsers();
		} elseif ($this->repository->getAccessType() == SCMRepository::ACCESS_TYPE_DAV) {
			$this->updateDAVUsers();
			$this->deleteSSHUsers();
		}
	}
	
	private function updateSSHUsers() {
		// If the permissions are handled by an external NSS like nss-pgsql, don't update the system files
		if (GFConf::get('system.useNSS')) {
			return;
		}
		
		$configManager = new SystemConfigManager();
		
		// Create the corresponding UNIX group
		$groupName = $this->project->getUnixName();
		$configManager->deleteGForgeGroup($groupName);
		$configManager->createGForgeGroup($groupName, $this->projectScm->getUnixGroupId());
		
		// Add all the users with at least read access
		/** @var User[] $users */
		$users = $this->project->getMembers();
		foreach ($users as $user) {
			$up = $this->project->getUserPermission($user);
			if (!$up->check('scm', 0, ScmPlugin::READ)) {
				continue;
			}
			
			// See if the UNIX account is configured for this user
			// IMPORTANT: Note that the entry is automatically created if it isn't found
			if (!GFConf::get('system.usesNis')) {
				try {
					//remove from cache, we need to pull the record from db
					UserUnixPeer::invalidate($user->getUserId());
					$user_unix = UserUnixPeer::retrieveByPkOrCreate($user->getUserId());
					$user_shell = $user_unix->getUnixShell();

					$uid = $user_unix->getUnixUid();
					$gid = $user_unix->getUnixGid();
					if (!$gid) $gid = $uid;
					$unix_password = $user->getPasswordCrypt();
					$home = SystemUtils::buildFilePath(GFConf::get('system.homeDir'), '/'.$user->getUnixName());
					$shell = $user_shell->getPath();
					$gcos = $user->getUnixName();

					$configManager->createGForgeUser($user->getUnixName(), $uid, $unix_password, $gid, $home, $shell, $gcos);
				} catch (PropelException $e) {
					syslog(LOG_ERR, $e->getMessage());
				}
			}
			
			$configManager->addUserToGroup($user->getUnixName(), $groupName);


		}
		
		// Also add the apache user (enables browsing)
		$apache_user = GFConf::get('system.apacheUser');
		$configManager->addUsertoGroup($apache_user, $groupName);
		
		// Adds access to the "anonymous" user for projects with anonymous access
		if ($this->repository->hasAnonymousAccess() && $configManager->userExists('anonymous')) {
			$configManager->addUserToGroup('anonymous', $groupName);
		}
		
		$configManager->writeChanges();
		//Restart apache to grab new membership, just if we are using the daemon
		if (GFConf::get('system.useDaemon')) {
			SystemUtils::restartApache();
		}

		//Create home directories

		foreach ($users as $user) {
			$homePath = GFConf::get('system.homeDir');
			$home = $homePath.DIRECTORY_SEPARATOR.$user->getUnixName();
			if (!file_exists($home)) {
				mkdir($home);
				chown($home, $user->getUnixName());
				chgrp($home, 'users');
				chmod($home, 0700);
			}
		}


	}
	
	private function deleteDAVUsers() {
		// This effectively removes the DAV configuration for this project
		$manager = new DAVConfigManager($this->project, $this->davAccessFile, $this->davAuthFile);
		$manager->writeChanges();
	}
	
	private function deleteSSHUsers() {
		// Remove the UNIX group for this project.
		$configManager = new SystemConfigManager();
		$configManager->deleteGForgeGroup($this->project->getUnixName());
		$configManager->writeChanges();
	}
	
	private function updateDAVUsers() {
		if ($this->project->usesPlugin('scmsvn') && GFConf::get('scmsvn.useModAuthGforge')) {
			//Ignore notification, system is using gforge module for authentication
			return;
		}
		$davManager = new DAVConfigManager($this->project, $this->davAccessFile, $this->davAuthFile);
		$davManager->addUserGroup('committers');
		$davManager->addUserGroup('readers');
		$davManager->addUser('anonymous', '');

		$users = $this->project->getMembers();

		foreach ($users as $user) {
			$davManager->addUser($user);

			$permission = $this->project->getUserPermission($user);
			
			if ($permission->check('scm', 0, ScmPlugin::WRITE)) {
				$davManager->addUserToGroup($user, 'committers');
			} else if ($permission->check('scm', 0, ScmPlugin::READ)) {
				$davManager->addUserToGroup($user, 'readers');
			}
		}
		
		
		// Set the repository-wide permissions
		$davManager->setGroupAccess('readers', '/', ScmPlugin::READ);
		$davManager->setGroupAccess('committers', '/', ScmPlugin::WRITE);

		// See if we need anonymous access
		if ($this->repository->hasAnonymousAccess()) {
			$davManager->setMinUserAccess('*', '/', ScmPlugin::READ);		// Give read access to all the users
		}
		
		// Now set the ACL (path) permissions
		$modules = $this->repository->getModules();
		foreach ($modules as $module) {
			$path = $module->getPath();
			$modulePerms = $this->repository->getModulePermissions($module->getScmModuleId());
			
			foreach ($modulePerms as $modulePerm) {
				if (is_null($modulePerm['module_access'])) {
					// this may happen if the user's permission for the module aren't set
					continue;
				}
				
				$user = $modulePerm['user'];
				
				$users = array();
				if ($user->getIsGroup()) {		// If it's a group, we must give the correct permissions to the users inside the group
					$ugus = $user->getUsergroupUsersRelatedByUsergroupId();
					foreach ($ugus as $ugu) {
						$users[] = UserPeer::retrieveByPk($ugu->getMemberUserId());
					}
				} else {
					$users[] = $user;
				}
				
				foreach ($users as $user) {
					$davManager->setMinUserAccess($user->getUnixName(), $path, $modulePerm['module_access']);
				}
			}
		}
		
		$davManager->writeChanges();
	}

	public function removeProjectGroups() {
		if ($this->repository->getAccessType() == SCMRepository::ACCESS_TYPE_SSH) {
			$this->deleteSSHUsers();
		}

		if ($this->repository->getAccessType() == SCMRepository::ACCESS_TYPE_DAV) {
			$this->deleteDAVUsers();
		}
	}
}
?>