<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Marcelo Mottalli	marcelo@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
 

require_once("system/Cronjob.php");
require_once('core/CronHistory.php');
require_once('plugins/scmgit/lib/GITRepository.php');
require_once('plugins/scm/lib/ScmProjectStats.php');
require_once('plugins/scm/lib/ScmUserStats.php');

// this could be a lengthy process
set_time_limit(0);

/**
 * Cronjob that parses the Git history
 */
class ParseGitHistoryCronjob extends Cronjob {
	public function __construct() {
		$this->jobname = basename(__FILE__);
		parent::__construct();
	}	
	
	public function run() {
		// Get all projects that use the plugin
		$c = new Criteria();
		$c->add(PluginPeer::PLUGIN_NAME, 'scmgit');
		$project_plugins = ProjectPluginPeer::doSelectJoinPlugin($c);
		
		foreach ($project_plugins as $project_plugin) {
			$project = $project_plugin->getProject(); // get the project
			$repos = new GitRepository($project);
			
			// repository hasn't been created yet... there isn't a history file
			if (!$repos->isCreated()) continue;
			
			$this->parseHistory($repos, $project);
		}
		$this->commitLog();
	}
	
	public function parseHistory(GitRepository $repos, Project $project) {
		$root = $repos->getRoot();
		if (!chdir($root)) {
			$this->logCron("Couldn't change directory to $root. Aborting.");
			return;
		}
		
		$output = array();

		$day_parameter = $this->getCronjobParameter(0);
		$bin = GFConf::get('scmgit.gitpath') . '/git rev-list ';
		if (strtolower($day_parameter) == 'all') {
			// Parsing all days -- retrieve the full log
			$cmd = 'cd /gitroot/' . $project->getUnixName() . '; '.
				$bin .' --all --pretty=format:%ae,%ct --encoding=utf8 &> /dev/null && ' .
				$bin . '--all --pretty=format:%ae,%ct --encoding=utf8';
		} else {
			$beg =(time()-(4*86400));
			$end = time()+86400;
			$cmd =  'cd /gitroot/' . $project->getUnixName() . '; ' .
				$bin . " --all --pretty=format:%ae,%ct --encoding=utf8 --max-age=$beg --min-age=$end &> /dev/null && " .
				$bin . " --all --pretty=format:%ae,%ct --encoding=utf8 --max-age=$beg --min-age=$end";
		}

		$res = SystemUtils::execute($cmd, $output);

		if ($res != 0) {		// error in command
			return;
		}
		
		$days_data = $this->parseDates($output);
		
		foreach ($days_data as $day => $day_data) {
			$c = new Criteria();
			$c->add(ScmProjectStatsPeer::STAT_DATE, $day);
			$c->add(ScmProjectStatsPeer::PROJECT_ID, $project->getProjectId());
			$stats = ScmProjectStatsPeer::doSelectOne($c);
			if (is_null($stats)) {
				$stats = new ScmProjectStats();
				$stats->setProjectId($project->getProjectId());
				$stats->setStatDate($day);
			}
			
			$stats->setCommits($day_data['commits']);
			$stats->setAdds(0);		// No adds information for Git
			$stats->setCheckouts(0);		// No checkouts information for Git
			$stats->save();
			
			// now store the user information
			foreach ($day_data['user_data'] as $email => $user_data) {
				$user = UserPeer::retrieveByEmail($email);
				
				if (is_null($user) || (is_array($user) && count($user) != 1)) {
					// Not a GForge user -- use the undefined user
					$user = UserPeer::retrieveByPk(User::UNDEFINED_USER_ID);
					if (is_null($user)) continue;
				} elseif (is_array($user)) {
					// $user is an array with only one user
					$user = $user[0];
				}
				
				$c = new Criteria();
				$c->add(ScmUserStatsPeer::STAT_DATE, $day);
				$c->add(ScmUserStatsPeer::PROJECT_ID, $project->getProjectId());
				$c->add(ScmUserStatsPeer::USER_ID, $user->getUserId());
				
				$stats = ScmUserStatsPeer::doSelectOne($c);
				if (!$stats) {
					$stats = new ScmUserStats();
					$stats->setStatDate($day);
					$stats->setProjectId($project->getProjectId());
					$stats->setUserId($user->getUserId());
				}
				
				$stats->setCommits($user_data['commits']);
				$stats->setAdds(0);
				$stats->setCheckouts(0);
				$stats->save();
			}
		}
	}
	
	public function parseDates($lines) {
		$days = array();

		// See which day(s) we want to parse
		$day_parameter = $this->getCronjobParameter(0);
		if (strtolower($day_parameter) == 'all') {
			// Parse all days?
			$parse_all = true;
		} else {
			$parse_all = false;
			
			$days_to_parse = array();			
			
			if ($day_parameter && ($time = strtotime($day_parameter)) !== -1) {
				$days_to_parse[] = date('Y-m-d', $time);
			} else {
				// Parse today and yesterday. This script is usually ran once per day. Parsing
				// the two days will ensure yesterday's information is correct, otherwise
				// the information between this script's last run and yesterday midnight
				// would get lost.
				$today = time();
				$yesterday = time()-(60*60*24);
				
				$days_to_parse[] = date('Y-m-d', $today);
				$days_to_parse[] = date('Y-m-d', $yesterday);
			}
		}
		
		for ($i = 0; $i < count($lines); $i++) {
			$line = trim($lines[$i]);
			// The line must be "commit <hash number>"
			if (!preg_match('/^commit /i', $line)) {
				// Shouldn't happen
				continue;
			}
			
			// Read the following line, which has the user email address and timestamp
			$i++;
			$line = trim($lines[$i]);
			list($email, $timestamp) = explode(',', $line);
			$day = date('Y-m-d', $timestamp);

			if (!$parse_all && !in_array($day, $days_to_parse)) {
				// we don't want to parse this day
				continue;
			}

			if (!array_key_exists($day, $days)) {
				$days[$day] = array();
				$days[$day]['adds'] = 0;
				$days[$day]['commits'] = 0;
				$days[$day]['user_data'] = array();
			}

			if (!array_key_exists($email, $days[$day]['user_data'])) {
				$days[$day]['user_data'][$email] = array();
				$days[$day]['user_data'][$email]['adds'] = 0;
				$days[$day]['user_data'][$email]['commits'] = 0;
			}
			
			$days[$day]['commits']++;
			$days[$day]['user_data'][$email]['commits']++;
		}
		
		return $days;
	}
}

$cronjob = new ParseGitHistoryCronjob();
$cronjob->run();
?>
