<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Marcelo Mottalli <marcelo@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

//require_once('SystemUtils.php');
require_once('plugins/scmgit/lib/exceptions/GITException.php');
require_once('plugins/scm/lib/SCMRepository.php');


class GitRepository extends SCMRepository {
	private $gitRoot;
	public $hostname;
	private $revisionCache = array();
	private $preReceiveFile, $postReceiveFile;
	private $rawInput;

	//Used by the git receive parser
	const REF_DELETE = 1;
	const REF_UPDATE = 2;
	const REF_CREATE = 3;


	public function __construct(Project $project) {
		parent::__construct($project);

		$this->gitRoot = GFConf::get('scmgit.localGitRoot');
		$this->gitpath = GFConf::get('scmgit.gitpath');
		$this->preReceiveFile = SystemUtils::buildFilePath($this->repositoryPath, '/hooks/pre-receive');
		$this->postReceiveFile = SystemUtils::buildFilePath($this->repositoryPath, '/hooks/post-receive');
	}

	/**
	 * Return the type of this repository (ie, 'cvs', 'svn', etc)
	 * @return	string
	 */
	public function getType(){
		return 'git';
	}
	
	
	/**
	 * Returns the root for ALL repositories
	 * @return	string
	 */
	public function getScmRoot() {
		return GFConf::get('scmgit.localGitRoot');
	}
	
	public function initializeRepository() {
		$commands = array(
			// Sequence of commands to execute in order to create the repository
			'cd '.$this->repositoryPath." && ".$this->gitpath.'/git --bare init --shared',
			'echo "'.$this->project->getProjectName().'" > '.$this->repositoryPath.'/description'
		);
		
		foreach ($commands as $command) {
			if (SystemUtils::execute($command, $output) != 0) {
				throw new GitException('Error creating Git repository: '.implode("\n", $output));
			}
		}
	}
	
	public function updateHooks() {
		$this->updatePreReceive();
		$this->updatePostReceive();
	}
	
	protected function updateExtraPermissions() {
		$output = array();
		
		SystemUtils::execute("chmod 770 {$this->preReceiveFile}", $output);
		SystemUtils::execute("chmod 770 {$this->postReceiveFile}", $output);
	}
	
	private function updatePreReceive() {
		// GForge manages the pre-receive script
		$unix_group_name = $this->project->getUnixName();
		$preReceiveCheckScript = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmgit/bin/pre_receive_checks.php');
		
		$old_content = array();
		$lines = array();
		$lines[] = '#!/bin/sh';
		if (is_readable($this->preReceiveFile)){
			$old_content = file($this->preReceiveFile);
			for ($i = 0; $i < count($old_content); $i++) {
				$line = trim($old_content[$i]);
				if (strstr($line,'#!/bin/sh')){
					continue;
				}
				if(empty($line)){
					continue;
				}
				if (strstr($line,'#BEGIN GForge pre-receive checks')) {
					//Preserve the lines BEFORE the Gforge markers
					break;
				}
				$lines[] = $line;
			}
		}

		$lines[] = '#BEGIN GForge pre-receive checks';
		$lines[] = '#if you wish to include any custom filter, include it ABOVE the BEGIN line';
		$lines[] = 'php -q -d include_path='.ini_get('include_path').' '.$preReceiveCheckScript;
		$lines[] = '#END GForge pre-receive checks';
		
		$this->insertCommitFilters($lines, ScmCommitFilter::PRE_COMMIT);

 		SystemUtils::writeFile($this->preReceiveFile, $lines);
	}
	
	private function updatePostReceive() {
		// GForge manages the post-receive script
		$unix_group_name = $this->project->getUnixName();
		$postReceiveScript = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmgit/bin/post_receive.php');

		$old_content = array();
		$lines = array();
		$lines[] = '#!/bin/sh';
		if (is_readable($this->postReceiveFile)) {
			$old_content = file($this->postReceiveFile);
			for ($i = 0; $i < count($old_content); $i++) {
				$line = trim($old_content[$i]);
				if (strstr($line,'#!/bin/sh')){
					continue;
				}
				if (empty($line)) {
					continue;
				}
				if(strstr($line,'#BEGIN GForge post-receive hooks')) {
					//Preserve the lines BEFORE the Gforge markers
					break;
				}
				$lines[] = $line;
				
			}
			
		}

		$lines[] = '#BEGIN GForge post-receive hooks';
		$lines[] = '#if you wish to include any custom filter, include it ABOVE the BEGIN line';
		$lines[] = 'php -q -d include_path='.ini_get('include_path').' '.$postReceiveScript;
		$lines[] = '#END GForge post-receive hooks';
		
		$this->updateMail($lines);

		$this->insertCommitFilters($lines, ScmCommitFilter::POST_COMMIT);

		SystemUtils::writeFile($this->postReceiveFile, $lines);
	}

	/**
	 * Inserts the commit filter commands
	 */
	private function insertCommitFilters(&$lines_src, $type = ScmCommitFilter::PRE_COMMIT) {
		$lines_dest = array();

		for ($i = 0; $i < count($lines_src); $i++) {
			$line = $lines_src[$i];
			if (strpos($line, '#BEGIN GForge filters') !== false) {
				// Seek to the end of the marker
				for (; $i < count($lines_src); $i++) {
					$line = $lines_src[$i];
					if (strpos($line, '#END GForge filters') !== false) {
						break;
					}
				}
			} else {
				$lines_dest[] = $line;
			}
		}
		
		// Now append the filters

		$c = new Criteria();
		$c->add(ProjectPluginPeer::PROJECT_ID, $this->project->getProjectId());
		$c->addJoin(ProjectPluginPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->add(PluginPeer::PLUGIN_NAME, ScmgitPlugin::getName());		// Make sure we're not including filters we shouldn't
		$c->addJoin(ScmCommitFilterProjectPeer::PROJECT_ID, ProjectPluginPeer::PROJECT_ID);
		$c->addJoin(ScmCommitFilterPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->addJoin(ScmCommitFilterProjectPeer::SCM_COMMIT_FILTER_ID, ScmCommitFilterPeer::SCM_COMMIT_FILTER_ID);
		//Only update the proper set of hooks
		$c->add(ScmCommitFilterPeer::FILTER_TYPE, $type);
		$filters = ScmCommitFilterPeer::doSelect($c);
		
		$lines_dest[] = '#BEGIN GForge filters';
		foreach ($filters as $filter) {
			$command = $filter->getFilterString();
			$command = StringUtils::unHtmlSpecialChars($command);
			$command = preg_replace('/\\$PROJECT/i', $this->project->getUnixName(), $command);
			$command = preg_replace('/\\$PATH/i', $this->getroot(), $command);
			//$command = preg_replace('/\\$FILE/i', '\\$2', $command);
			$lines_dest[] = $command . ' || exit $? ';
		}
		$lines_dest[] = '#END GForge filters';
		
		$lines_src = $lines_dest;
	}
	
	public function updateMail(&$lines) {
		// Read the preferences
		$preferences = $this->getScmPreferences();

 		if (!is_null($preferences) && $preferences->getMailCommits()) {
			$this->addMail($lines);
		} else {
			$this->removeMail();
		}
	}

	public function removeMail() {
		$cmd = $this->gitpath . '/git config';
		$repoPath = $this->repositoryPath;
		$commands = array(
			"cd $repoPath && $cmd -l | grep -q hooks.mailinglist && $cmd --unset hooks.mailinglist",
			"cd $repoPath && $cmd -l | grep -q hooks.emailprefix && $cmd --unset hooks.emailprefix",
			"cd $repoPath && $cmd -l | grep -q hooks.emailmaxlines && $cmd --unset hooks.emailmaxlines",
			"cd $repoPath && sed -i 's/\\[hooks\\]//' config"
		);

		foreach ($commands as $command) {
			SystemUtils::execute($command, $output);

			/*
			if (SystemUtils::execute($command, $output) != 0) {
				throw new GitException('Error adding Git main configuration: '.implode("\n", $output) . implode("\n", $commands));
			}*/
		}

	}
	
	public function addMail(&$lines) {
		// First check if we have mailman configured
		if (!$this->project->usesPlugin('mailman')) {
			return;
		}

		$cmd = $this->gitpath . '/git config';
		$repoPath = $this->repositoryPath;
		$mailingList = $this->repositoryName.'-commits@'.GFConf::get('mailman.listsHost');
		$maxlines = GFConf::get('scmgit.emailMaxLines');

		// Set up the mailing lists on the Git repository
 		$commands = array(
 			// Sequence of commands to execute
		   //"(cd $repoPath && grep -q \"[hooks]\" config) ||  $cmd --remove-section hooks",
		   "cd $repoPath && $cmd  hooks.mailinglist $mailingList",
		   "cd $repoPath && $cmd  hooks.emailprefix \"[Gforge-Git]\"",
		   "cd $repoPath && $cmd  hooks.emailmaxlines $maxlines",
		   "cd $repoPath && $cmd  hooks.envelopesender '".GFConf::get('system.name')."<noreply@".GFConf::get('system.hostName').">'"
		   //'cd '.$this->repositoryPath." && ".$this->gitpath.'/git config hooks.mailinglist "'.$this->repositoryName.'-commits@'.GFConf::get('mailman.listsHost').'"',
 		   //'cd '.$this->repositoryPath." && ".$this->gitpath.'/git config hooks.emailprefix "[GForge-Git]"',
 		);
 		
 		foreach ($commands as $command) {
		   SystemUtils::execute($command, $output) != 0;
		   /*
 			if (SystemUtils::execute($command, $output) != 0) {
 				throw new GitException('Error adding Git main configuration: '.implode("\n", $output));
 			}*/
 		}
 		
 		// Note that the actual email is sent by the post_recieve.php script
	}
	
	
	public function doLocalCheckout($destdir, $module) {
		// $module here is the branch name
		$path = $this->repositoryPath;

		// If the repo is not on the local filesystem, it will have a scheme prefix
		// (like git:// or ssh:// or http(s)://).
		// If it doesn't have this prefix, we need to prepend file:// scheme to 
		// prevent issue [#10137] from happening
		if (stristr($path, '://') == FALSE) {
			$path = "file://{$path}";
		}

		if (!$module) {
			$module = 'master';			// The default module
		}
		
		if (is_dir($destdir)) {
			rmdir($destdir);			// Directory must not exist
		}
		
		$output = array();
		$cmd = "{$this->gitpath}/git clone -n $path $destdir";
		$error = SystemUtils::execute($cmd, $output);
		if ($error) {
			throw new Exception("Unable to checkout the code - Error code $error - Command was '$cmd'");
		}
		
		chdir($destdir);

		$cmd = "{$this->gitpath}/git checkout $module";
		$error = SystemUtils::execute($cmd, $output);
		if ($error) {
			throw new Exception("Unable to checkout the code - Error code $error - Command was '$cmd'");
		}
	}
	
	public function getCheckedoutFileRevision($checkout_dir, $file) {
		if (!isset($this->revisionCache[$checkout_dir])) {
			chdir($checkout_dir);
			
			$output = array();
			$cmd = $this->gitpath.'/git rev-list --max-count=1 HEAD';
			$error = SystemUtils::execute($cmd, $output);
			if ($error) {
				return -1;
			}
			
			$revision = trim($output[0]);
			$this->revisionCache[$checkout_dir] = $revision;
		}
		
		return $this->revisionCache[$checkout_dir];
	}
	
	protected function getGlobalAccessType() {
		return (GFConf::get('scmgit.useSSH') ? SCMRepository::ACCESS_TYPE_SSH : SCMRepository::ACCESS_TYPE_DAV);
	}


	/**
	 * Parses the output from git-receive-pack
	 * @param bool $ignoreMerges
	 *
	 * @return array
	 * Structure: array('ref_name'=> array(
	 *   'old' => string
	 *   'new' => string
	 *   'type' => int (REF_DELETE | REF_UPDATE |REF_CREATE)
	 *   'hashes'=>array(hash_1, hash_2 ... hash_n-1, hash_n)
	 * ))
	 * Each hash actually stores the commit hash and the commit message, separated by a space
	 * @throws Exception
	 */
	public function parseReceivePack($ignoreMerges=true, $type=GitReceive::PRE_SCRIPT) {
		$this->rawInput = $input = file_get_contents("php://stdin");
		$lines = preg_split("/\n/", $input, NULL, PREG_SPLIT_NO_EMPTY);
		$refs = array();
		$mergeSwitch = '';

		if ($ignoreMerges) {
			$mergeSwitch = ' --no-merges ';
		}

		$cmd = "git for-each-ref --format '^%(refname:short)' refs/heads/";
		SystemUtils::execute($cmd, $excludeBranches);

		//Escape branch names
		foreach($excludeBranches as $key => $branchName) {
			$excludeBranches[$key] = addslashes($branchName);
		}

		// $line should be something like
		// "375afa809bf83b858db3033d51e67d1a9020633c 9c13a7a4f8c8e37747f9387ac688830b054eceed refs/heads/master"
		// one for each ref being updated/deleted/created
		foreach($lines as $line) {
			$new_hashes = array();

			list($old_value, $new_value, $ref_name) = explode(" ", trim($line));

			$new_zero = preg_match('/^0000000000/', $new_value);
			$old_zero = preg_match('/^0000000000/', $old_value);
			// Get the revision logs up to the previous update

			if ($type == GitReceive::POST_SCRIPT) {
				//IF this is called from post_receive, we need to exclude the current branch
				$ref_short = preg_replace('/refs\/heads\//', '', $ref_name);

				$key = array_search('^'.$ref_short, $excludeBranches);
				unset($excludeBranches[$key]);
			}

			if ($old_zero && $new_zero) {
				// Branch delete using the command "git push origin :refs/branches/[branch]"
				$change_type = self::REF_DELETE;
				$cmd = "";
			} elseif ($old_zero) {
				// Initial revision?

				$cmd = "git rev-list --pretty=oneline  $new_value $mergeSwitch ". implode(' ',$excludeBranches);
				$change_type = self::REF_CREATE;
			} elseif ($new_zero) {
				// Branch delete using the command "git push origin :[branch]
				$change_type = self::REF_DELETE;
				$cmd = "git rev-list --pretty=oneline $old_value";
			} else {
				$change_type = self::REF_UPDATE;
				$cmd = "git rev-list --pretty=oneline $mergeSwitch $new_value ^$old_value " . implode(' ', $excludeBranches);
			}

			if ($cmd) {
				$error = SystemUtils::execute($cmd, $new_hashes);
				if ($error) {
					throw new Exception("Error executing the command \"$cmd\": \n". implode("\n", $new_hashes));
				}
			}

			$refs[$ref_name] = array(
				'old' => $old_value,
				'new' => $new_value,
				'type' => $change_type,
				'hashes' => $new_hashes
			);

		}

		return $refs;
	}

	public function getRawInput() {
		return $this->rawInput;
	}



	/**
	 * Returns a list of branches where this commit is present
	 * @param string $hash
	 * @return array
	 */
	public function getBranchesForCommit($hash) {
		$repoPath = $this->repositoryPath;
		$gitExec = $this->gitpath.'/git';
		$ret = array();
		$output = array();

		$cmd = "cd $repoPath && $gitExec branch  --contains $hash";

		if (SystemUtils::execute($cmd, $output) == 0) {
			$ret = $output;
		}

		return $ret;

	}


}

?>
