<?php

require_once('core/Project.php');
require_once('core/User.php');
require_once('SystemUtils.php');
require_once('plugins/scm/lib/ScmCommit.php');
require_once('plugins/scm/lib/ScmCommitWrapper.php');
require_once('plugins/scm/lib/ScmPlugin.php');
require_once('plugins/scmgit/lib/GITRepository.php');




/**
 * Class GitReceive
 *
 * @property GITRepository repository
 * @property Project project
 * @property User user
 * @property array refs
 * array('ref_name'=> array(
 *   'old' => string,
 *   'new' => string,
 *   'type' => int (REF_DELETE | REF_UPDATE |REF_CREATE),
 *   'hashes'=>array('hash_1','hash_2',...,'hash_n')
 *    )
 * )
 * @property array processed_refs
 * array('ref_name'=>array(
 *    'invalidIds'=>array('id'=>'message', 'id'=>'message'...),
 *    'validIds'=>array('tid1', 'tid2',...)
 *    'hashes'=>
 *       'hash1'=>array(
 *          'invalidIds'=>array('id'=>'message'. 'id2'=>'message'...)
 *          'validIds'=>array('id'=>'message', 'id2'=>'message'...)
 *        )
 *        ...
 * )
 * @property integer scriptType
 *
 */
class GitReceive {
	private $project;
	private $repository;
	private $refs;
	private $user;
	private $processedRefs;
	private $scriptType;

	const REF_DELETE = 1;
	const REF_UPDATE = 2;
	const REF_CREATE = 3;
	const PRE_SCRIPT = 'pre';
	const POST_SCRIPT = 'post';



	/**
	public function __construct() {
	$this->change_type = '';
	$this->fail = false;
	$this->branch_name;
	$this->messages = array();
	$this->show_suggestions_message = false;
	$this->show_branch_suggestions_message = false;
	}*/

	/**
	 * @param Project $Project
	 * @param User $User
	 * @param int|string $scriptType
	 * @throws Exception
	 */
	public function __construct($Project, $User, $scriptType=self::PRE_SCRIPT) {
		if (!$Project || !($Project instanceof Project)) {
			throw new Exception("Invalid project object\n");
		}
		if (!$User || !($User instanceof User)) {
			throw new Exception("Invalid user object\n");
		}
		$this->project = $Project;
		$this->repository = new GITRepository($this->project);
		$this->user = $User;
		$this->refs = $this->repository->parseReceivePack(true, $scriptType);
		$scmPref = ScmPreferencePeer::retrieveByPk($Project->getId());
		$this->requireTIds = !is_null($scmPref) && $scmPref->getRequireTrackerItemId();
		$this->scriptType = $scriptType;

		//Start session
		$this->createUserSession();
	}

	/**
	 * Retrieve the project for the current working directory
	 *
	 * @return Project
	 * @throws Exception
	 */
	public function getProject() {
		return $this->project;
	}




	/**
	* Prints all refs, for debug purposes only
	*/
	public function printRefs() {
		foreach($this->refs as $branch => $data){
			echo "Branch $branch:\n";
			echo "\tOld ref: {$data['old']}\n";
			echo "\tNew ref: {$data['new']}\n";
			echo "\tType: {$data['type']}\n";
			echo "\t hashes: " . implode(' ', $data['hashes'])."\n\n";
		}
	}


	/**
	 * Creates user session
	 */
	private function createUserSession() {
		// Start the session as the user that's committing
		Session::getSession()->setUser($this->user);

		return;
	}

	/*
	 * Parses ids in the current ref array
	 */
	public function parseIds() {
		foreach($this->refs as $refname=>$data) {

			if (!isset($data['hashes']) || !count($data['hashes'])) {
				//Nothing to parse
				continue;
			}

			if ($data['type'] == GitRepository::REF_DELETE) {
					//Do nothing
				continue;
			}

			$tmp = ScmCommit::processBranchNameTIDs($refname, $this->project);
			$this->processedRefs[$refname]['validIds'] = $tmp['validIds'];

			$this->processedRefs[$refname]['invalidIds'] = $tmp['invalidIds'];
 			$this->processedRefs[$refname]['hashes'] = array();

			foreach ($data['hashes'] as $log) {
				$parsedLog = explode(' ', $log);

				$hash = $parsedLog[0];
				if (!$this->getCommitExists($hash)) {
					$this->processedRefs[$refname]['hashes'][$hash] = ScmCommit::processCommitMsgTIDs($log,$this->project);
				}
			}

		}

	}

	/**
	 * Checks that all the ids in a branch name are valid references to tracker items
	 * @param string $ref the branch name
	 * @return bool
	 */
	private function allIdsInRefNameValid($ref) {
		if (isset($this->processedRefs[$ref])) {
			if(count($this->processedRefs[$ref]['invalidIds']) == 0 && count($this->processedRefs[$ref]['validIds']) == 0){
				//No ids at all
				return false;
			}

			if(count($this->processedRefs[$ref]['invalidIds']) == 0 && count($this->processedRefs[$ref]['validIds']) >= 1) {
				//No invalid ids, at leas one valid id
				return true;
			}
		}

		return false;
	}

	/**
	 * Checks that all ids in a commit log are valid
	 * @param string $ref
	 * @param string $hash
	 * @return bool
	 */
	private function allIdsInLogValid($ref, $hash) {
		if (isset($this->processedRefs[$ref]) && isset($this->processedRefs[$ref]['hashes'][$hash])) {
			if (count($this->processedRefs[$ref]['hashes'][$hash]['invalidIds']) == 0 &&
				count($this->processedRefs[$ref]['hashes'][$hash]['validIds']) == 0) {
				//No ids
				return false;
			}

			if (count($this->processedRefs[$ref]['hashes'][$hash]['invalidIds']) == 0
				&& count($this->processedRefs[$ref]['hashes'][$hash]['validIds']) >= 1
				&& count($this->processedRefs[$ref]['hashes'][$hash]['statuses']['invalid']) === 0
                && count($this->processedRefs[$ref]['hashes'][$hash]['time_entries']['invalid']) === 0) {
				return true;
			}
		}

		return false;

	}


	/**
	 * Checks that all statuses in a commit log are valid
	 * @param string $ref
	 * @param string $hash
	 * @return bool
	 */
	private function allStatusesInLogValid($ref, $hash) {
		if (isset($this->processedRefs[$ref]) && isset($this->processedRefs[$ref]['hashes'][$hash])) {
			if (count($this->processedRefs[$ref]['hashes'][$hash]['statuses']['invalid']) > 0) {
				return false;
			}
		}

		return true;
	}


	/**
	 * Checks that all statuses in a commit log are valid
	 * @param string $ref
	 * @param string $hash
	 * @return bool
	 */
	private function allIdsStatusesLogValid($ref, $hash) {
		if (isset($this->processedRefs[$ref]) && isset($this->processedRefs[$ref]['hashes'][$hash])) {
			if (count($this->processedRefs[$ref]['hashes'][$hash]['invalidIds']) == 0 &&
				count($this->processedRefs[$ref]['hashes'][$hash]['validIds']) == 0) {
				//No ids
				return false;
			}

			if (count($this->processedRefs[$ref]['hashes'][$hash]['invalidIds']) == 0
				&& count($this->processedRefs[$ref]['hashes'][$hash]['validIds']) >= 1) {
				return true;
			}
		}

		return false;

	}



	/**
	 * Check if a branch name has any tracker item id reference
	 * @param string $ref
	 * @return bool
	 */
	private function refContainsIds($ref) {
		assert(isset($this->processedRefs[$ref]));
		if (count($this->processedRefs[$ref]['invalidIds']) == 0
			&& count($this->processedRefs[$ref]['validIds']) == 0 ) {
			return false;
		}

		return true;
	}

	/**
	 * Checks if a commit message has any tracker item id reference
	 * @param string $ref
	 * @param string $hash
	 * @return bool
	 */
	private function logContainsIds($ref,$hash) {
		assert(isset($this->processedRefs[$ref]['hashes'][$hash]));

		if (count($this->processedRefs[$ref]['hashes'][$hash]['invalidIds']) == 0 &&
			count($this->processedRefs[$ref]['hashes'][$hash]['validIds']) == 0) {
			return false;
		}

		return true;
	}


	/**
	 * Checks that all logs (commit messages and branch names) are valid
	 * @return bool
	 */
	public function allLogsValid() {
		return count($this->getInvalidCommits()) == 0;
	}


    public function processInvalidStatusMessages($data, $refname) {
        //Process invalid statuses, if any
        $messages_tmp = '';
        foreach ($data['hashes'] as $hash => $data2) {
            if (count($data2['statuses']['invalid']) >= 1) {
                $messages_tmp .= "\n\n" . $this->convertShortHash($data2['hash']) . ":" . $this->getMessageLog($data2['log'])."\n";
                foreach($data2['statuses']['invalid'] as $tid => $arr_data) {
	                $status = $arr_data['status'];
	                $err_code = $arr_data['errcode'];

					$error_msg = "";
	                switch($err_code) {
		                case ScmCommit::STATUS_UPDATE_INVALID_ALIAS:
			                    $error_msg = Language::getSessText('ScmCommitMessage.invalidAlias');
			                break;

		                case ScmCommit::STATUS_UPDATE_CANT_TRANSITION:
			                $error_msg = Language::getSessText('ScmCommitMessage.cannotTransition');
			                break;

		                case ScmCommit::STATUS_UPDATE_COMMIT_ACTION:
			                $error_msg = Language::getSessText('ScmCommitMessage.commitAction');
			                break;

		                case ScmCommit::STATUS_UPDATE_NO_NEXT_STATUS:
			                $error_msg = Language::getSessText('ScmCommitMessage.noNextStatus');
			                break;
	                }

	                $messages_tmp .= $error_msg . "\n";
	                if ($err_code == ScmCommit::STATUS_UPDATE_INVALID_ALIAS || $err_code == ScmCommit::STATUS_UPDATE_NO_NEXT_STATUS) {
                        $messages_tmp .= " > Id " . $tid . ", status: " . $status . "\n";
                        $messages_tmp .= " > Available statuses: " . ScmCommit::getStatusesByTid($tid);
	                }
                }
            }
        }
        $messages = '';
        if ($messages_tmp != '') {
            $messages .= "Branch " . basename($refname). " contains either invalid or forbidden status updates in commit logs:\n";
            $messages .= $messages_tmp;
            $messages .="\n";
        }

        return $messages;
    }


    public function processInvalidTimeEntryMessages($data, $refname) {
        //Process invalid time entries, if any
        $messages_tmp = '';
        foreach ($data['hashes'] as $hash => $data2) {
            if (count($data2['time_entries']['invalid']) >= 1) {
                $messages_tmp .= "\n\n" . $this->convertShortHash($data2['hash']) . ":" . $this->getMessageLog($data2['log'])."\n";
                foreach($data2['time_entries']['invalid'] as $tid => $time_entry) {
                    $messages_tmp .= " > Id " . $tid . ", time entry: " . $time_entry . "\n";
                    $messages_tmp .= " > Time entries must be greater than: " . ScmCommit::MIN_TIME_ENTRY;
                }
            }
        }

        $messages = '';
        if ($messages_tmp != '') {
            $messages .= "\nBranch " . basename($refname). " contains invalid time entry references in commit logs:\n";
            $messages .= $messages_tmp;
            $messages .="\n";
        }

        return $messages;
    }


	/**
	 * Collects all invalid commits and builds a string that explains the problem(s) to the user
	 * @return string
	 */
	public function getInvalidCommitsMessages() {
		$invalidCommits = $this->getInvalidCommits();
		$messages = '';

		if (count($this->processedRefs)) {
			$messages = "#######################################################################\n";
			$messages .= "Push rejected: Repository requires valid tracker item ids for branches\n";
			$messages .= "and/or commits, also if you specify a status update, it is subject to\n";
			$messages .= "acceptance based on transition validity. If you specify a time entry in\n";
			$messages .= "your commit message, it has to be a valid time entry, in the format:\n";
			$messages .= "'integer.integer', e.g.: 1.25\n";
			$messages .= "#######################################################################\n\n";
		}

		foreach($this->processedRefs as $refname=>$data) {
			if(array_key_exists($refname, $invalidCommits)) {
				if($this->refs[$refname]['type'] == GitRepository::REF_CREATE) {

					//All ids in branch name are valid but there are invalid ids in commits
					if ($this->refContainsIds($refname) && count($data['invalidIds']) == 0) {
						$messages .= "Branch " . basename($refname) . " contains valid tracker item ids \n".
						"but the following commits include references to tracker items that \n" .
						"don't exist or you can't access:\n\n";
						foreach ($data['hashes'] as $hash => $data2) {
							if ( count($data2['invalidIds']) ) {
								$messages .= "   " .$this->convertShortHash($hash) . ": " .
								$this->getMessageLog($data2['log']) . "\n";
							}
						}
						$messages .= "\n";
					} elseif ($this->refContainsIds($refname) && count($data['invalidIds']) >= 1) {
						//Some ids in ref branch are not valid, push denied

						$messages .= "Branch " . basename($refname) . " contains invalid tracker item ids: \n";
						foreach ($data['invalidIds'] as $invalidIds => $reason) {
							$messages .= "   " . $invalidIds . " : " . $reason . "\n";
 						}
						$messages .="\n";

					} else {

						//No ids in branch, but the branch is present in invalid commits array, this means
						//we have invalid ids or statuses in at least one commit
						$messages_tmp = '';
						foreach ($data['hashes'] as $hash => $data2) {
							if (count($data2['invalidIds']) == 0  && count($data2['validIds']) == 0) {
								$messages_tmp .= "   " . $this->convertShortHash($hash) . ": " . $this->getMessageLog($data2['log'])."\n";
								$messages_tmp .= "      doesn't reference any tracker item id\n";
							} elseif (count($data2['invalidIds']) >= 1) {
								$messages_tmp .= "   " .	 $this->convertShortHash($data2['hash']) . ":" . $this->getMessageLog($data2['log'])."\n";
								foreach($data2['invalidIds'] as $invalidIds => $reason) {
									$messages_tmp .= "      " . " Id " . $invalidIds . " : " . $reason . "\n";
								}
							}
						}

						if ($messages_tmp != '') {
							$messages .= "Branch " . basename($refname). " doesn't reference any tracker item id and some commits\n".
							"don't reference valid ids:\n";
							$messages .= $messages_tmp;
							$messages .="\n";
						}
					}
					
                    $messages .= $this->processInvalidStatusMessages($data, $refname);

                    $messages .= $this->processInvalidTimeEntryMessages($data, $refname);
				}

				if($this->refs[$refname]['type'] == GitRepository::REF_UPDATE) {
					//All ids in branch name are valid but there are invalid ids in commits
					if ($this->refContainsIds($refname) && count($data['invalidIds']) == 0) {
						$messages .= "Branch " . basename($refname) . " contains valid tracker item ids \n".
						"but the following commits include references to tracker items that \n" .
						"you can't access:\n\n";
						foreach ($data['hashes'] as $hash => $data2) {
							if ( count($data2['invalidIds']) ) {
								$messages .= "   " .$this->convertShortHash($hash) . ": " .
								$this->getMessageLog($data2['log']) . "\n";
							}
						}
						$messages .= "\n";
					} else {
						//No ids in branch, but the branch is present in invalid commits array, this means
						//we have invalid ids in at least one commit

						$messages_tmp = '';
						foreach ($data['hashes'] as $hash => $data2) {
							if (count($data2['invalidIds']) == 0  && count($data2['validIds']) == 0) {
								$messages_tmp .= "   " . $this->convertShortHash($hash) . ": " . $this->getMessageLog($data2['log'])."\n";
								$messages_tmp .= "      doesn't reference any tracker item id\n";
							} elseif (count($data2['invalidIds']) >= 1) {
								$messages_tmp .= "   " .	 $this->convertShortHash($hash) . ":" . $this->getMessageLog($data2['log'])."\n";
								foreach($data2['invalidIds'] as $invalidIds => $reason) {
									$messages_tmp .= "      " . " Id " . $invalidIds . " : " . $reason . "\n";
								}
							}
						}
						
						if ($messages_tmp != '') {
							$messages .= "Branch " . basename($refname). " doesn't reference any tracker item id and some commits\n".
							"don't reference valid ids:\n";
							$messages .= $messages_tmp;
							$messages .="\n";
						}
						
					}

                    $messages .= $this->processInvalidStatusMessages($data, $refname);

                    $messages .= $this->processInvalidTimeEntryMessages($data, $refname);

				}
			}
		}

		$messages .="\n\n";
		return $messages;

	}

	/**
	 * Collect invalid ids on branch names and commit messages
	 * @return array
	 */
	private function getInvalidCommits() {

		$invalidCommits = array();

		if(!count($this->processedRefs)) {
			//An empty array means a single branch deletion
			return $invalidCommits;
		}

		foreach($this->processedRefs as $refname=>$data) {

			//Ignore initial push to master
			if ($this->refs[$refname]['type'] == GitRepository::REF_CREATE && $refname == 'refs/heads/master') {
				continue;
			}

			//Ignore deletions
			if ($this->refs[$refname]['type'] == GitRepository::REF_DELETE) {
				continue;
			}

			//A new branch is created
			if ($this->refs[$refname]['type'] == GitRepository::REF_CREATE) {
				//Branch id must be valid, if there is any reference
				$containsIDs = $this->refContainsIds($refname);
				$allIdsValid = $this->allIdsInRefNameValid($refname);

				if ( $containsIDs ) {
					//Branch contains ids, all ids must be valid
					// Check all commits, if there is an id, it should be valid, if not, we abort the transaction
					if ($allIdsValid) {
						foreach($data['hashes'] as $hash => $results){
							if ($this->logContainsIds($refname, $hash)) {
								if (!$this->allIdsInLogValid($refname,$hash)) {
									$invalidCommits[$refname][$hash] = $hash;
								}
							}
						}

					} else {
						//Store the ref as invalid
						$invalidCommits[$refname] = $refname;
					}

				} else {
					//Branch doesn't contains IDS, check all commits (empty ids/invalid ids non-allowed)
					foreach($data['hashes'] as $hash => $results) {
						if(!$this->allIdsInLogValid($refname, $hash))
							$invalidCommits[$refname][$hash] = $hash;
					}

				}

			}


			//A push to an existing branch
			if ($this->refs[$refname]['type'] == GitRepository::REF_UPDATE) {
				if($this->allIdsInRefNameValid($refname)) {
					//If we have an empty id, we use the branch id, so only fail if the id is invalid
					foreach ($data['hashes'] as $hash => $results) {
						if ($this->logContainsIds($refname, $hash)) {
							if (!$this->allIdsInLogValid($refname, $hash)) {
								$invalidCommits[$refname][$hash] = $hash;
							}
						}
					}
				} else {
					//Branch doesn't have valid ids, we don't fail but we enforce ids in commits
					foreach ($data['hashes'] as $hash => $results) {
						if (!$this->allIdsInLogValid($refname, $hash)) {
							$invalidCommits[$refname][$hash] = $hash;
						}
					}
				}

			}

		}

		return $invalidCommits;
	}


	/**
	 * Prints the processedRefs property, for debug purposes only
	 */
	public function printProcessedRefs(){
		var_dump($this->processedRefs);
	}

	/**
	 * Prints the parsedRefs property, for debug purposes only
	 */
	public function printParsedRefs(){
		var_dump($this->refs);
	}

	public function getRefs() {
		$ret = array();
		foreach($this->refs as $refname => $data) {
			$ret[] = basename($refname);
		}

		return $ret;
	}

	/**
	 * Returns a shorted hash (7 characters)
	 * @param $hash
	 * @return string
	 */
	public function convertShortHash($hash){
		return substr($hash, 0, 7);
	}

	/**
	 * Returns the message log only (removes the hash from the line)
	 * @param $log
	 * @return string
	 */
	public function getMessageLog($log) {
		return trim(preg_replace('/^[a-z0-9]{5,40}/', '', $log, 1 ));
	}


	/**
	 * Check all acls (should be used for ssh repositories only)
	 * @return bool
	 * @throws Exception
	 */
	public function allValidAcls() {

		$messages = "#######################################################################\n";
		$messages .= "Push rejected: You don't have enough permissions, details follow\n";
		$messages .= "#######################################################################\n";

		foreach($this->refs as $refname => $data) {

			if ($data['type'] == self::REF_CREATE){
				// Initial revision?
				$cmd = "git rev-list --objects {$data['new']}";
			} else if ($data ['type'] == self::REF_UPDATE) {
				$cmd = "git rev-list --objects {$data['new']} ^{$data['old']}";
			} else { //Branch is going to be deleted
				if (!$this->repository->getProject()->userIsAdmin()) {
					//branch deletion is allowed to a project admin only
					$messages .= "\nDelete $refname: Only a project admin can delete a remote branch\n\n";
					echo $messages;
					return false;
				}
				return true;
			}
			$output = array();
			$error = SystemUtils::execute($cmd, $output);
			if ($error) {
				throw new Exception($messages . "\nError executing the command \"$cmd\": \n". implode("\n", $output));
			}

			$changed_objects = array();
			foreach ($output as $line) {
				$line = trim($line);
				// $line should be something like
				// $hash $changed_object
				if (!preg_match('/([^ ]*) (.*)/', $line, $matches)) {
					continue;
				}

				$changed_objects[] = $matches[2];
			}

			// Note that the changed objects can be duplicated if they were changed in more than one commit
			$changed_objects = array_unique($changed_objects);
			foreach ($changed_objects as $file) {
				$file = SystemUtils::buildFilePath('/', $file);
				if (!$this->repository->userCanCommit($file)) {
					$messages .= "\n".Language::getSessText('Scm.NoWriteAccess', array($this->user->getUnixName(), $file))."\n";
					echo $messages;
					return false;
				}
			}
		}

		return true;
	}

	public function parseAssociations() {

		if (!count($this->processedRefs)) {
			return;
		}

		foreach ($this->processedRefs as $refname => $data) {
			if ($this->refs[$refname]['type'] == self::REF_CREATE) {
				if(basename($refname) == "master") {
					//Initial push, we want to parse all commits
					$revspec = $this->refs[$refname]['new'];
				} else {
					$initial = array_pop(array_keys($data['hashes']));
					if (!$initial) {
						// No parent commit, this is a tag
						continue;
					}
					$revspec = $initial .'^1..' . $this->refs[$refname]['new'];
				}
			}
			if ($this->refs[$refname]['type'] == self::REF_DELETE) {
				$revspec = $this->refs[$refname]['old'];
			}
			if ($this->refs[$refname]['type'] == self::REF_UPDATE) {
				$revspec = $this->refs[$refname]['old'] . '..'. $this->refs[$refname]['new'];
			}


			$cmd ="git rev-list --no-merges --pretty=\"format:%H;%ci;%s\" --date-order --reverse --stdin $revspec ";
			$commits = array();
			SystemUtils::execute($cmd, $commits);

			//The branch ids
			$extraIds = $data['validIds'];
			// Process each commit separately


			foreach ($commits as $commit_line) {
				// If the line is "commit <hash>", skip it
				if (strpos($commit_line, 'commit') === 0) {
					continue;
				}

				// $commit_line is "$hash;$timestamp;$log")
				preg_match('/^([^;]*);([^;]*);(.*)$/', $commit_line, $matches);
				$hash = $matches[1];
				$date = $matches[2];
				$commit_log = $matches[3];

				if ($this->getCommitExists($hash)) {
					continue;
				}

				// Get the modified files
				$modified_files = array();

				SystemUtils::execute("git show --name-only --no-merges --pretty=oneline $hash 2> /dev/null", $modified_files);
				if (count($modified_files)) {
					$modified_files = array_slice($modified_files, 1);		// Remove the first line (hash and log)
				}

				$files = array();
				foreach ($modified_files as $onefile) {
					$files[] = array(
					'path' => '/'.$onefile,
					'prev_version' => isset($oldrev) ? $oldrev : ((isset($this->refs[$refname]['old']) && !preg_match('/000000/',$this->refs[$refname]['old']) ) ? $this->refs[$refname]['old'] : $this->getFirstParentCommit($hash)),
					'new_version' => $hash
					);
				}

				// Store the commit information
				$commitIds = $data['hashes'][$hash]['validIds'];
				$wrapper = new ScmCommitWrapper($this->repository, $this->user, $commit_log, $files, strtotime($date), $commitIds,$extraIds);
				$wrapper->generateCommitData();

				// Keep track of the commit hash
				$oldrev = $hash;
			}
		}
	}

	public function getFirstParentCommit($hash) {
		SystemUtils::execute("git log --pretty=%P -n1 $hash 2> /dev/null", $output);
		if (isset($output[0])) {
			return $output[0];
		} else {
			return '';
		}
	}

	public function sendCommitEmail() {
		$input = $this->repository->getRawInput();
		$emailScript = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmgit/bin/post-receive-email');
		// Pipe the input to the email script
		$output = array();
		$input = trim($input);
		SystemUtils::execute("echo $input | $emailScript", $output);
	}


	/**
	 * Checks if commit exists in this repository
	 *
	 * @param string $hash
	 * @return bool
	 */
	public function getCommitExists($hash) {


		if ($this->scriptType == self::PRE_SCRIPT &&
			count($this->repository->getBranchesForCommit($hash)) == 0) {
			//When invoked by the pre-commit hook, the commit is not there yet
			return false;
		}


		if ($this->scriptType == self::POST_SCRIPT &&
			count($this->repository->getBranchesForCommit($hash)) == 1) {
			//When invoked by the post-commit hook, the commit will be in a single branch
			//Ignore subsequent merges
			return false;
		}


		return true;
	}

}