<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id: Group.class,v 1.94 2005/09/06 18:13:29 danper Exp $
 * @author Tim Perdue tim@gforge.org
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('notification/GFNotification.php');
require_once("plugin/ActionManager.php");
require_once('core/ProjectPeer.php');
require_once('TrackerItemPeer.php');
require_once('TrackerItemAssigneePeer.php');
require_once('TrackerItemMessagePeer.php');
require_once('StringUtils.php');
require_once('plugins/frs/lib/FrsReleasePeer.php');
require_once('plugins/frs/lib/FrsRelease.php');

class TrackerItemNotification implements GFNotification {
	const TRACKER_MAIL_MARKER = '#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+';// This string is used when sending the notification mail for identifying the user response

	private $TrackerItem;
	private $Tracker;
	private $Project;
	private $notificationType='trackeritem';
	private $users;
	private $modifiedValues;

	/**
	 * Flag used to show in notification message if a follow-up was added
	 *
	 * @var boolean
	 */
	private $followUpAdded;

	/**
	 * Flag used to show in notification message if tracker item is new
	 *
	 * @var flag
	 */
	private $isNew;

	function __construct(TrackerItem $TrackerItem, $modifiedValues=array(), $followUpAdded = false, $isNew=false) {
		$this->TrackerItem = $TrackerItem;
		$this->Tracker = $TrackerItem->getTracker();
		$this->Project = $this->Tracker->getProject();
		$this->modifiedValues = $modifiedValues;
		$this->followUpAdded = $followUpAdded;
		$this->isNew = $isNew;
	}

	/**
	 * Returns an array containing an assoc. array with markers to fields 
	 * changed and the explanatory text to show at the top of the mail message.
	 *
	 * @return array Containing (markers, bodyDetail)
	 */
	private function getMessageChanges()
	{
		$fields = array(
			'status'    => TrackerItemPeer::STATUS_ID, // not used?
			'priority'  => TrackerItemPeer::PRIORITY,
			'summary'   => TrackerItemPeer::SUMMARY,
			'details'   => TrackerItemPeer::DETAILS,
			'parent id' => TrackerItemPeer::PARENT_ID // not used?
		);
		$markers = array();

		foreach ($fields as $key => $value) {
			$markers[$key] = '';
			if (in_array($value, $this->modifiedValues)) {
				$markers[$key] = '>';
			}
		}
		$strDetails = 'updated'; // default, generic message
		if ($this->isNew) {
			$strDetails = 'created';
		} else if (!empty($this->modifiedValues['assignees'])) {
			$strDetails = 'added assignees in';
		} else if ($this->followUpAdded) {
			$strDetails = 'added a follow-up to';
		}
		return array($markers, $strDetails);
	}
	
	private function getGenericMessage() {
		$url = GFUrl::getUrl();

		//set the main variables for the items according whether they have been changed or not
		list($markers, $strActionDetail) = $this->getMessageChanges();
		$strName = $this->TrackerItem->getLastModifiedByNameUnixname();
		$strTracker = $this->TrackerItem->getTracker()->getTrackerName();
		$strTrackerItemId = $this->TrackerItem->getPrimaryKey();
		$strDate = $this->TrackerItem->getOpenDate();
		$body = <<<TXT
{$strName} {$strActionDetail} item #{$strTrackerItemId} - tracker "{$strTracker}" on {$strDate}


TXT;
		if (!($this->Tracker->getSendAbbreviatedEmail() && $markers['priority'] == "")) {
			$body .= $markers['priority'] .
				"Priority: ". $this->TrackerItem->getPriority() ."\n";
		}
		$body .= "Submitted By: ". $this->TrackerItem->getUser()->getFirstName() . " " . $this->TrackerItem->getUser()->getLastName() .
			" (". $this->TrackerItem->getUser()->getUnixName() . ")"."\n";
		// get the user_id of the assignees
		$assignees  = $this->TrackerItem->getAssignedUsers();

		foreach ($assignees as $assignee) {
			if (array_key_exists('assignees', $this->modifiedValues) && is_array($this->modifiedValues['assignees'])) {
				$marker = (in_array($assignee->getUserId(), $this->modifiedValues['assignees'])?">":"");	
			} else {
				$marker = '';
			}
			if (!($this->Tracker->getSendAbbreviatedEmail() && $marker == "")) {
				$body .=  $marker."Assigned to: ". $assignee->getFirstName() . " " . $assignee->getLastName() .
		 			" (". $assignee->getUnixName() . ")"."\n";
		 	}
		}

		$body .= $markers['summary']. "Summary: ". $this->getGenericSummary() ." \n";

		// add the extra fields
		$extra_fields = $this->TrackerItem->getTracker()->getTrackerExtraFields();
		$datas=$this->TrackerItem->getTrackerExtraFieldDatas();
		$selected=array();
		foreach ($datas as $data) {
			$selected[$data->getTrackerExtraFieldId()][]=$data->getFieldData();
		}
		foreach ($extra_fields as $field) {
			if (array_key_exists('extra_fields', $this->modifiedValues) && is_array($this->modifiedValues['extra_fields'])) {
				$marker = (in_array($field->getTrackerExtraFieldId(), $this->modifiedValues['extra_fields'])?">":"");
			} else {
				$marker = '';
			}
			
			if ($this->Tracker->getSendAbbreviatedEmail() && $marker == "") {
				continue;
			}
			
			$body .= $marker.$field->getFieldName() .": ";
			// $extra_field->getFieldData()."\n";
			$str = 'N/A';
			
			if (array_key_exists($field->getTrackerExtraFieldId(), $selected)) {
				switch ($field->getFieldType()) {
					case TrackerExtraField::TYPE_MEMBERS:
						$crit = new Criteria();
						$crit->add(UserPeer::USER_ID,$selected[$field->getTrackerExtraFieldId()],Criteria::IN);
						$elements=UserPeer::doSelect($crit);
						$str='';
						foreach ($elements as $element) {
							$str .= $element->getFirstname(). ' ' . $element->getLastname() . ', ';
						}
						break;
					case TrackerExtraField::TYPE_SELECT:
					case TrackerExtraField::TYPE_RADIO:
					case TrackerExtraField::TYPE_STATUS: {
						if (isset($selected[$field->getTrackerExtraFieldId()])) {
							$element=TrackerExtraFieldElementPeer::retrieveByPk($selected[$field->getTrackerExtraFieldId()][0]);
							if (is_object($element)) {
								$str=$element->getElementName();
							} else {
								$str='N/A';
							}
						} else {
							$str='N/A';
						}
						break;
					}
					case TrackerExtraField::TYPE_RELEASE: {
						if (isset($selected[$field->getTrackerExtraFieldId()])) {
							$element=FrsReleasePeer::retrieveByPk($selected[$field->getTrackerExtraFieldId()][0]);
							if (is_object($element)) {
								$str=$element->getReleaseName();
							} else {
								$str='N/A';
							}
						} else {
							$str='N/A';
						}
						break;
					} 
					case TrackerExtraField::TYPE_CHECKBOX:
					case TrackerExtraField::TYPE_MULTISELECT: {
						$crit = new Criteria();
						$crit->add(TrackerExtraFieldElementPeer::ELEMENT_ID,$selected[$field->getTrackerExtraFieldId()],Criteria::IN);
						$elements=TrackerExtraFieldElementPeer::doSelect($crit);
						$str='';
						foreach ($elements as $element) {
							$str .= $element->getElementName().', ';
						}
						break;
					}
				//	case TrackerExtraField::TYPE_DATE:
				//	case TrackerExtraField::TYPE_NUMERIC:
				//	case TrackerExtraField::TYPE_TEXT:
				//	case TrackerExtraField::TYPE_TEXTAREA: {
					default: {
						if (isset($selected[$field->getTrackerExtraFieldId()])) {
							$str=$selected[$field->getTrackerExtraFieldId()][0];
						} else {
							$str='N/A';
						}
					}
				} //end switch
			}
			$body .= $str."\n";
			unset($str);
		}

		//construct the body	
		if (!($this->Tracker->getSendAbbreviatedEmail() && $markers['details'] == "")) {
			$body .= "\n\nInitial Comment:".
				"\n". $this->getDetails() .
				"\n\n----------------------------------------------------------------------";
		}

		// include the followups
		$c = new Criteria();
		$c->addAscendingOrderByColumn(TrackerItemMessagePeer::TRACKER_ITEM_MESSAGE_ID);
		$followups = $this->TrackerItem->getTrackerItemMessages($c); // get the followups for this item
		if (count($followups)>0) {
			if ($this->Tracker->getSendAbbreviatedEmail()) {
				$body .= "\n\nLast Follow-Up:";
				$followup = $followups[sizeof($followups)-1];
				$body .= "\n\n-------------------------------------------------------";
				$body .= "\nDate: ". $followup->getAdddate();
				$body .= "\nBy: ". $followup->getUser()->getFirstName() . " " . $followup->getUser()->getLastName();
				$body .= "\n\nComment:\n". $followup->getBody();
				// add the user who posted the followups' email to the email list				
			} else {
				$body .= "\n\nFollow-Ups:";
				foreach ($followups as $followup) {
					$body .= "\n\n-------------------------------------------------------";
					$body .= "\nDate: ". $followup->getAdddate();
					$body .= "\nBy: ". $followup->getUser()->getFirstName() . " " . $followup->getUser()->getLastName();
					$body .= "\n\nComment:\n". $followup->getBody();
					// add the user who posted the followups' email to the email list				
				}
			}
		}

		$body .=	"\n\n----------------------------------------------------------------------";
		$body .= "\n\n To respond simply reply this email or visit:\n".
			$url->getFullUrl($url->projectUrl($this->TrackerItem->getTracker()->getProject(),
				'TrackerItemEdit', 'tracker', '', array('tracker_id' => $this->TrackerItem->getTrackerId(),
				'tracker_item_id' => $this->TrackerItem->getPrimaryKey())), true)."\n\n";


		return $body;
	}

	/**
	 *	getNotificationMessage - construct a string of this message content.
	 *
	 *	@param	User
	 *	@return string
	 */
	function getNotificationMessage(User $User) {
		$url = GFUrl::getUrl();
		$body = $this->getGenericMessage();
		
				// Some users that aren't monitoring may be notified, so include the following message
		// only if the user is monitoring
		$monitors_tracker = $this->Tracker->userIsMonitoring($User);
		$monitors_trackeritem = $this->TrackerItem->userIsMonitoring($User);
		if ($monitors_tracker || $monitors_trackeritem) {
			$body .= "\n\n______________________________________________________________________".
					"\n\nYou received this email because you ".
					"\nrequested to be notified when changes were made to this tracker.".
					"\nIf you don't wish to be notified in the future, please login to \n".
					$url->getFullUrl($url->baseUrl('Login', '', 'account'), true) . " and click this link: ";
			if ($monitors_trackeritem) {
				$body .= "\n" . $url->getFullUrl($url->getMonitorURL(TrackerItem::SECTION, $this->TrackerItem->getPrimaryKey(), $url->projectUrl($this->Project, '', 'tracker'), false), true);
			} elseif ($monitors_tracker) {
				$body .= "\n" . $url->getFullUrl($url->getMonitorURL(Tracker::SECTION, $this->TrackerItem->getTracker()->getPrimaryKey(), $url->projectUrl($this->Project, '', 'tracker'), false),true);
			}
					
		}
		
		return $body."\n";

	}
	
	private function getGenericSummary() {
		$summary = $this->TrackerItem->getSummary();
		return StringUtils::unHtmlSpecialChars($summary);
	}

	/**
	 *	getSummary - construct a string of this message summary (html chars unconverted).
	 *
	 *	@param	User
	 *	@return string
	 */
	function getSummary(User $User) {
		return $this->getGenericSummary();
	}

	/**
	 *	getDetails - construct a string of this message summary (html chars unconverted).
	 *
	 *	@return string
	 */
	function getDetails() {

		$details = $this->TrackerItem->getDetails(false);
		return StringUtils::getPlainUTF8FromHtml($details);
	}	
	
	/**
	 *	getNotificationSubject - construct a string of this message subject.
	 *
	 *	@param	User
	 *	@return string
	 */
	function getNotificationSubject(User $User) {
		return $this->getGenericSubject();
	}
	
	function getGenericSubject() {
		$subject="[ " . $this->TrackerItem->getTracker()->getTrackerName() . " #". $this->TrackerItem->getPrimaryKey() .'] '.  $this->getGenericSummary();
		return $subject;
	}

	/**
	 *	getUsers - get an array of user objects
	 *	@return array
	 */
	public function getUsers() {
		return $this->TrackerItem->getMonitoringUsers(); // tracker item or tracker or project monitors
	}

	/**
	 *	getSection - retrieves the section name of the trackeritem
	 *	@return string
	 */	
	function getSection() {
		return TrackerItem::SECTION;
	}
	
	/**
	 *	getRefId - retrieves the id of the trackeritem that the notification belongs to
	 *	@return int
	 */	
	function getRefId() {
		return $this->TrackerItem->getPrimaryKey();
	}

	function getMessageId() {
		return '<'.$this->getSection().'-'.$this->getRefId().'@'.GFConf::get('system.hostName').'>';
	}

	function getInReplyTo() {
		$inReplyTo = $this->TrackerItem->getTrackerItemId();
		return '<'.$this->getSection().'-'.$inReplyTo.'@'.GFConf::get('system.hostName').'>';
	}
	
	public function notifyUpdate() {
		$addresses = StringUtils::getAddressesFromList($this->TrackerItem->getTracker()->getEmailAddress());
		
		foreach ($addresses as $address) {
			$message = $this->getGenericMessage();
			$subject = $this->getGenericSubject();
			$reply_to = 'trackeritem+'.$this->TrackerItem->getPrimaryKey();
			$from = $this->Tracker->getEmailAlias();

			SystemUtils::sendMail($subject, $message, $address, false, $from, SystemUtils::getFromAddress($reply_to));
		}
	}

	public function getReferences() {
		return $this->getInReplyTo();
	}

	
}

?>
